Package: PrInDT
Type: Package
Title: Prediction and Interpretation in Decision Trees for
        Classification and Regression
Version: 2.0.3
Authors@R: c(person("Claus", "Weihs", email = "claus.weihs@tu-dortmund.de",  role = c("aut", "cre")),
             person("Sarah", "Buschfeld", email="sarah.buschfeld@tu-dortmund.de", role="aut"),
             person("Niklas", "Nitsch", email="niklas.nitsch@tu-dortmund.de", role="ctb"))
Description: Optimization of conditional inference trees from the package 'party'
    for classification and regression.
    For optimization, the model space is searched for the best tree on the full sample by 
    means of repeated subsampling. Restrictions are allowed so that only trees are accepted 
    which do not include pre-specified uninterpretable split results (cf. Weihs & Buschfeld, 2021a).
    The function PrInDT() represents the basic resampling loop for 2-class classification (cf. Weihs 
    & Buschfeld, 2021a). The function RePrInDT() (repeated PrInDT()) allows for repeated  
    applications of PrInDT() for different percentages of the observations of the large and the 
    small classes (cf. Weihs & Buschfeld, 2021c). The function NesPrInDT() (nested PrInDT()) 
    allows for an extra layer of subsampling for a specific factor variable (cf. Weihs & Buschfeld, 
    2021b). The functions PrInDTMulev() and PrInDTMulab() deal with multilevel and multilabel 
    classification. In addition to these PrInDT() variants for classification, the function 
    PrInDTreg() has been developed for regression problems. Finally, the function PostPrInDT() 
    allows for a posterior analysis of the distribution of a specified variable in the terminal 
    nodes of a given tree.
    In version 2, additionally structured sampling is implemented in functions PrInDTCstruc() and 
    PrInDTRstruc(). In these functions, repeated measurements data can be analyzed, too. 
    Moreover, multilabel 2-stage versions of classification and regression trees are 
    implemented in functions C2SPrInDT() and R2SPrInDT() as well as interdependent 
    multilabel models in functions SimCPrInDT() and SimRPrInDT(). Finally, for mixtures of 
    classification and regression models functions Mix2SPrInDT() and SimMixPrInDT() are 
    implemented. Most of these extensions of PrInDT are described in Buschfeld & 
    Weihs (2026).
    References:
    -- Buschfeld, S., Weihs, C. (2026) "Optimizing decision trees for the analysis of World Englishes 
    and sociolinguistic data", Cambridge Elements. <doi:10.1017/9781009470346>;
    -- Weihs, C., Buschfeld, S. (2021a) "Combining Prediction and Interpretation in 
    Decision Trees (PrInDT) - a Linguistic Example" <doi:10.48550/arXiv.2103.02336>;
    -- Weihs, C., Buschfeld, S. (2021b) "NesPrInDT: Nested undersampling in PrInDT" 
    <doi:10.48550/arXiv.2103.14931>;
    -- Weihs, C., Buschfeld, S. (2021c) "Repeated undersampling in PrInDT (RePrInDT): Variation 
    in undersampling and prediction, and ranking of predictors in ensembles" <doi:10.48550/arXiv.2108.05129>.
License: GPL-2
Encoding: UTF-8
Imports: graphics, MASS, party, splitstackshape, stats, stringr, utils,
        gdata
RoxygenNote: 7.3.3
Depends: R (>= 2.10)
NeedsCompilation: no
Maintainer: Claus Weihs <claus.weihs@tu-dortmund.de>
LazyData: true
Packaged: 2026-07-20 11:16:15 UTC; Master
Author: Claus Weihs [aut, cre],
  Sarah Buschfeld [aut],
  Niklas Nitsch [ctb]
Repository: CRAN
Date/Publication: 2026-07-22 08:20:14 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,C2SPrInDT)
S3method(plot,Mix2SPrInDT)
S3method(plot,NesPrInDT)
S3method(plot,OptPrInDT)
S3method(plot,PrInDT)
S3method(plot,PrInDTAll)
S3method(plot,PrInDTAllparts)
S3method(plot,PrInDTCstruc)
S3method(plot,PrInDTMulab)
S3method(plot,PrInDTMulabAll)
S3method(plot,PrInDTMulev)
S3method(plot,PrInDTMulevAll)
S3method(plot,PrInDTRstruc)
S3method(plot,PrInDTreg)
S3method(plot,PrInDTregAll)
S3method(plot,R2SPrInDT)
S3method(plot,RePrInDT)
S3method(plot,SimCPrInDT)
S3method(plot,SimMixPrInDT)
S3method(plot,SimRPrInDT)
S3method(predict,PrInDT)
S3method(predict,PrInDTAll)
S3method(predict,PrInDTMulev)
S3method(predict,PrInDTMulevAll)
S3method(predict,PrInDTreg)
S3method(predict,PrInDTregAll)
S3method(print,C2SPrInDT)
S3method(print,Mix2SPrInDT)
S3method(print,NesPrInDT)
S3method(print,OptPrInDT)
S3method(print,PrInDT)
S3method(print,PrInDTAll)
S3method(print,PrInDTAllparts)
S3method(print,PrInDTCstruc)
S3method(print,PrInDTMulab)
S3method(print,PrInDTMulabAll)
S3method(print,PrInDTMulev)
S3method(print,PrInDTMulevAll)
S3method(print,PrInDTRstruc)
S3method(print,PrInDTreg)
S3method(print,PrInDTregAll)
S3method(print,R2SPrInDT)
S3method(print,RePrInDT)
S3method(print,SimCPrInDT)
S3method(print,SimMixPrInDT)
S3method(print,SimRPrInDT)
export(C2SPrInDT)
export(Mix2SPrInDT)
export(NesPrInDT)
export(OptPrInDT)
export(PostPrInDT)
export(PrInDT)
export(PrInDTAll)
export(PrInDTAllparts)
export(PrInDTCstruc)
export(PrInDTMulab)
export(PrInDTMulabAll)
export(PrInDTMulev)
export(PrInDTMulevAll)
export(PrInDTRstruc)
export(PrInDTreg)
export(PrInDTregAll)
export(R2SPrInDT)
export(RePrInDT)
export(SimCPrInDT)
export(SimMixPrInDT)
export(SimRPrInDT)
import(MASS)
import(party)
import(stats)
importFrom(gdata,mv)
importFrom(graphics,abline)
importFrom(graphics,barplot)
importFrom(graphics,hist)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(party,ctree)
importFrom(party,ctree_control)
importFrom(splitstackshape,stratified)
importFrom(stats,aggregate)
importFrom(stats,formula)
importFrom(stats,median)
importFrom(stats,predict)
importFrom(stats,relevel)
importFrom(stringr,str_split)
importFrom(utils,capture.output)

#' Two-stage estimation for classification
#'
#' @description The function C2SPrInDT applies two-stage estimation for finding an optimal model for relationships between the two-class factor variables
#' specified as column indices of 'datain' in the vector 'inddep' and all other factor and numerical variables in the data frame 'datain' by means of
#' 'N' repetitions of random subsampling with percentages 'percl' for the large classes and 'percs' for the small classes. One percentage of observations 
#' for each dependent variable has to be specified for the larger and the smaller class. For example, for three dependent variables, 'percl' consists of 
#' three percentages specified in the order in which the dependent variables appear in 'inddep'.\cr
#' The dependent variables have to be specified as dummies, i.e. as 0 for 'property absent' or 1 for 'property present' for a certain property the dependent variable is representing.\cr
#' The indices of the predictors relevant at 1st stage modeling can be specified in the vector 'indind'. If 'indind' is not specied, then all variables in 'datain' 
#' which are not specified in 'inddep' are used as predictors at 1st stage. At 2nd stage, all such variables are used as predictors anyway.\cr
#' The optimization citerion is the balanced accuracy on the full sample. The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'. \cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' @usage C2SPrInDT(datain,ctestv=NA,conf.level=0.95,percl,percs,N=99,indind=NA,inddep,
#'                                  minsplit=NA,minbucket=NA)
#'
#' @param datain Input data frame with class factor variables and the influential variables,\cr
#'    which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param percl list of undersampling percentages of larger class (numerical, > 0 and <= 1): one per dependent class variable in the same order as in 'inddep'
#' @param percs list of undersampling percentage of smaller class (numerical, > 0 and <= 1); one per dependent class variable in the same order as in 'inddep'
#' @param N no. of repetitions (integer > 0); default = 99
#' @param indind indices of independent variables at stage 1; default = NA (means all independent variables used)
#' @param inddep indices of dependent variables
#' @param minsplit Minimum number of elements in a node to be splitted; default = 20
#' @param minbucket Minimum number of elements in a node; default = 7
#'
#' @return
#' \describe{
#'   \item{models1}{Best trees at stage 1} 
#'   \item{models2}{Best trees at stage 2} 
#'   \item{classnames}{names of classification variables}
#'   \item{baAll}{balanced accuracies of best trees at both stages} 
#' }
#'
#' @details
#' See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.1, for further information.
#'
#' Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name)} where 'name' is the output data 
#' frame of the function.\cr
#
#' @export C2SPrInDT
#'
#' @examples
#' data <- PrInDT::data_land # load data
#' dataclean <- data[,c(1:7,23:24,11:13,22,8:10)]  # only relevant features
#' indind <- c(1:9) # original predictors
#' inddep <- c(14:16) # dependent variables
#' dataland <- na.omit(dataclean)
#' ctestv <- NA
#' perc <- c(0.45,0.05,0.25)   # percentages of observations of larger class, 
#' # 1 per dependent class variable
#' perc2 <- c(0.75,0.95,0.75)  # percentages of observations of smaller class, 
#' # 1 per dependent class variable
#' outland <- C2SPrInDT(dataland,percl=perc,percs=perc2,N=19,indind=indind,inddep=inddep)
#' outland
#' plot(outland)
#'
C2SPrInDT <- function(datain,ctestv=NA,conf.level=0.95,percl,percs,N=99,indind=NA,inddep,minsplit=NA,minbucket=NA){
## input check
if ( typeof(datain) != "list" || !(typeof(ctestv) %in% c("logical", "character"))
    || !all(0 < percl & percl <= 1) || !all(0 < percs & percs <= 1) 
    || !(typeof(N) %in% c("integer","double")) 
    || !(typeof(inddep) %in% c("integer","double")) || !(typeof(indind) %in% c("logical","integer","double"))
    || !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) 
    || !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
data <- datain
classnames <- colnames(data)[inddep]
anz <- length(inddep)
message ("1st stage")
models <- list()
equal <- matrix(TRUE,ncol=anz,nrow=dim(data)[1])
ctpreds <- matrix(0,ncol=anz,nrow=dim(data)[1])
ba <- c(1:(anz+2))*0
for (i in 1:anz){
    x <- colnames(datain)[inddep[i]]
    levels(data[,x]) <- c(levels(data[,x]),x)
    if (table(data[,x])[2] > dim(data)[1]/2){
      data[datain[,x] == 1,x] <- x
    } else {
      data[datain[,x] != 1,x] <- x
    } 
    data[,x] <- droplevels(data[,x])
#    message("\n")
    message(paste0("  ",x))
#  message(colnames(data)[inddep[i]])
  if (all(is.na(indind)) == TRUE){
    indall <- c(1:(dim(data)[2]))
    indall <- setdiff(indall,inddep)
    indall <- c(inddep[i],indall)
  } else {
    indall <- c(inddep[i],indind) # externals + ith dependent
  }
  perc <- percl[i]
  perc2 <- percs[i]
  outland <- PrInDT(data[,indall],classnames[i],ctestv=ctestv,N,perc,perc2,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)
#  print(outland)  ##
#  plot(outland$tree1st)
  ctpredsVL <- predict(outland$tree1st,newdata=data)
  equal[,i] <- data[,inddep[i]] == ctpredsVL
#  equal[,i] <- (as.integer(data[,inddep[i]])-1) == ctpredsVL
  ctpreds[,i] <- ctpredsVL
  ba[i] <- outland$ba1st[3]
  models <- c(models,outland$tree1st)
}
  acc01 <- rowSums(equal)
  ba[anz+2] <- sum(acc01) / (anz*dim(data)[1])  ## indices generalized
  acc01[acc01 < anz] <- 0
  acc01[acc01 == anz] <- 1
  ba[anz+1] <- sum(acc01)
  ba[anz+1] <- ba[anz+1] / dim(data)[1]
#print(ba)
##
## 2nd stage
#
message("2nd stage")
dataP <- data
dataP <- cbind(dataP,ctpreds-1)  ## ????????????   change of coding: 2-1 -> 1-0
dim(dataP)
baP <- ba
modelsP <- models
ctpredsP <- ctpreds
models <- list()
equal <- matrix(TRUE,ncol=anz,nrow=dim(data)[1])
ctpreds <- matrix(0,ncol=anz,nrow=dim(data)[1])
ba <- c(1:(anz+2))*0
#dataP[,dim(data)[2]+anz] <- 1 - dataP[,dim(data)[2]+anz]  ## generalized above
for (i in 1:anz){
  dataP[,dim(data)[2]+i] <- as.factor(dataP[,dim(data)[2]+i])
  names(dataP)[dim(data)[2]+i] <- paste0(classnames[i],"_2S")
}
for (i in 1:anz){
  message(paste0("  ",classnames[i]))
  indall <- c(1:(dim(dataP)[2]))
  j <- dim(data)[2]+i
  indall <- indall[-j] # externals + predictions - endogenous
  j <- inddep[-i]
  indall <- indall[-j] # externals + predictions - endogenous
  perc <- percl[i]
  perc2 <- percs[i]
  outland <- PrInDT(dataP[,indall],classnames[i],ctestv=ctestv,N,perc,perc2,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)
#  print(outland)  ##
#  plot(outland$tree1st)
  ctpredsVL <- predict(outland$tree1st,newdata=dataP)
  equal[,i] <- data[,inddep[i]] == ctpredsVL
#  equal[,i] <- (as.integer(dataP[,inddep[i]])-1) == ctpredsVL
  ctpreds[,i] <- ctpredsVL
  ba[i] <- outland$ba1st[3]
  models <- c(models,outland$tree1st)
}
  acc01 <- rowSums(equal)
  ba[anz+2] <- sum(acc01) / (anz*dim(data)[1])  ## indices generalized
  acc01[acc01 < anz] <- 0
  acc01[acc01 == anz] <- 1
  ba[anz+1] <- sum(acc01)
  ba[anz+1] <- ba[anz+1] / dim(data)[1]
baP <- rbind(baP,ba)
result <- list(models1=modelsP,models2=models,classnames=classnames,baAll=baP) 
class(result) <- "C2SPrInDT"
result
}
##
#' @export
print.C2SPrInDT <- function(x,...){
## output
cat("Accuracies of best models","\n")
colnames(x$baAll) <- c(x$classnames,"01-acc.","Hamming")
rownames(x$baAll) <- c("1st stage","2nd stage")
print(x$baAll)
cat("\n\n")
for (i in 1:length(x$classnames)){
  cat("Best model 2nd stage: ",x$classnames[i], "\n")
  print(x$models2[[i]])
  cat("\n")
}
}
#' @export
plot.C2SPrInDT <- function(x,...){
for (i in 1:length(x$classnames)){
  plot(x$models2[[i]],main=x$classnames[i])
}
}

#' Two-stage estimation for regression
#'
#' @description The function R2SPrInDT applies 'N' repetitions of subsampling for finding an optimal subsample to model
#' the relationship between the continuous variables with indices 'inddep' and all other factor and numerical variables
#' in the data frame 'datain'. \cr
#' Subsampling of observations and predictors uses the percentages in 'pobs1' and 'ppre1', respectively, at stage 1, and the percentages 'pobs2' and 'ppre2' 
#' at stage 2, accordingly.
#' The optimization criterion is the goodness of fit R2 on the full sample.  \cr
#' The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' @usage R2SPrInDT(data,ctestv=NA,inddep,N=99,pobs1=c(0.90,0.70),ppre1=c(0.90,0.70),
#'                 pobs2=pobs1,ppre2=ppre1,conf.level=0.95,minsplit=NA,minbucket=NA)
#'
#' @param data Input data frame with continuous target variable 'regname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param inddep Column indices of target variables in datain
#' @param N Number of repetitions of subsampling from predictors (integer) in versions "b" and "c";\cr
#'      default = 99
#' @param pobs1 Percentage(s) of observations for  subsampling at stage 1;\cr
#'      default=c(0.9,0.7)
#' @param ppre1 Percentage(s) of predictors for subsampling at stage 1;\cr
#'      default=c(0.9,0.7)
#' @param pobs2 Percentage(s) of observations for  subsampling at stage 2";\cr
#'      default=pobs1
#' @param ppre2 Percentage(s) of predictors for subsampling at stage 2;\cr
#'      default=ppre1
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return
#' \describe{
#'   \item{models1}{Best trees at stage 1} 
#'   \item{models2}{Best trees at stage 2} 
#'   \item{depnames}{names of dependent variables}
#'   \item{R2both}{R2s of best trees at both stages}  
#' }
#'
#' @details
#' See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.1, for further information.
#'
#' Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name} where 'name' is the output data 
#' frame of the function.\cr
#
#' @export R2SPrInDT
#'
#' @examples
#' data <- PrInDT::data_vowel
#' data <- na.omit(data)
#' CHILDvowel <- data$Nickname
#' data$Nickname <- NULL
#' syllable <- 3 - data$syllables
#' data$syllabels <- NULL
#' data$syllables <- syllable
#' data$speed <- data$word_duration / data$syllables
#' names(data)[names(data) == "target"] <- "vowel_length"
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
#' inddep <- c(13,9) 
#' out2SR <- R2SPrInDT(data,ctestv=ctestv,inddep=inddep,N=9,conf.level=0.99)
#' out2SR
#' plot(out2SR)
#'
R2SPrInDT <- function(data,ctestv=NA,inddep,N=99,pobs1=c(0.90,0.70),ppre1=c(0.90,0.70),pobs2=pobs1,ppre2=ppre1,conf.level=0.95,minsplit=NA,minbucket=NA){
# input check
if ( typeof(data) != "list" || !(typeof(ctestv) %in% c("logical", "character")) || !(typeof(inddep) %in% c("integer","double"))
    || !all(0 < pobs1 & pobs1 <= 1) || !all(0 < ppre1 & ppre1 <= 1)
    || !all(0 < pobs2 & pobs2 <= 1) || !all(0 < ppre2 & ppre2 <= 1) 
    || !(typeof(N) %in% c("integer","double")) || !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) 
    || !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
depnames <- colnames(data)[inddep]
anz <- length(inddep)
D <- dim(data)[2]
## only exogenous predictors
message("1st stage")
set.seed(7654321)
models <- list()
ctpreds <- matrix(0,ncol=anz,nrow=dim(data)[1])
R2 <- c(1:anz)*0
for (i in 1:anz){
  indpre <- c(1:(dim(data)[2]))[-inddep]
  lind <- length(indpre)
  indmid <- c(min(inddep):max(inddep))[-(inddep-min(inddep)+1)]
  indpre <- c(indpre[1:(min(inddep)-1)],indmid,inddep[i],indpre[(min(inddep)+length(indmid)):lind])
#  indpre <- c(indpre[1:9],inddep[i],indpre[10:17])
# print(indpre)
  message(depnames[i],": 1st stage")
#outreg <- PrInDTreg(datavowel,"vowel_length",ctestv,N,pobs,ppre,conf.level=cf) ## all vowel related variables
  outreg <- PrInDTreg(data[,indpre],depnames[i],ctestv,N=N,pobs=pobs1,ppre=ppre1,conf.level=conf.level,seedl=FALSE,minsplit=minsplit,minbucket=minbucket)
#  print(outreg) 
#  plot(outreg)
  models <- c(models,outreg$ctmax)
  ctpredsVL <- predict(outreg$ctmax,newdata=data)
  ctpreds[,i] <- ctpredsVL
  R2[i] <- outreg$maxR2
# print(R2[i])
}
#
## 2nd stage estimations
#
message("\n","2nd stage")
set.seed(7654321)
dataP <- data
dataP <- cbind(dataP,ctpreds)
R2P <- R2
modelsP <- models
for (i in 1:anz){
  names(dataP)[(D+i)] <- paste0(depnames[i],"_2S")
}
models <- list()
ctpreds <- matrix(0,ncol=anz,nrow=dim(dataP)[1])
R2 <- c(1:anz)*0
for (i in 1:anz){
  indpre <- c(1:(dim(dataP)[2]))[-(D+i)]
  indpre <- indpre[-inddep]
  indpre <- c(inddep[i],indpre)
#print(indpre)
  message(depnames[i],": 2nd stage")
  outreg <- PrInDTreg(dataP[,indpre],depnames[i],ctestv,N=N,pobs=pobs2,ppre=ppre2,conf.level=conf.level,seedl=FALSE,minsplit=minsplit,minbucket=minbucket)
  models <- c(models,outreg$ctmax)
  ctpredsVL <- predict(outreg$ctmax,newdata=dataP)
  ctpreds[,i] <- ctpredsVL
  R2[i] <- outreg$maxR2
}
R2P <- rbind(R2P,R2)
colnames(R2P) <- depnames
result <- list(models1=modelsP,models2=models,depnames=depnames,R2both=R2P) 
class(result) <- "R2SPrInDT"
result
}
#' @export
print.R2SPrInDT <- function(x,...){
## output
cat("Accuracies of best models","\n")
colnames(x$R2both) <- c(x$depnames)
rownames(x$R2both) <- c("1st stage","2nd stage")
print(x$R2both)
cat("\n\n")
for (i in 1:length(x$depnames)){
  cat("Best model 2nd stage: ",x$depnames[i], "\n")
  print(x$models2[[i]])
  cat("\n")
}
}
#' @export
plot.R2SPrInDT <- function(x,...){
for (i in 1:length(x$depnames)){
  plot(x$models2[[i]],main=x$depnames[i])
}
}

#' Structured subsampling for classification
#'
#' @description The function PrInDTCstruc applies structured subsampling for finding an optimal subsample to model
#' the relationship between the two-class factor variable 'classname' and all other factor and numerical variables
#' in the data frame 'datain' by means of 'N' repetitions of subsampling from a substructure and 'Ni' repetitions of subsampling from the predictors.  
#' The optimization citerion is the balanced accuracy on the validation sample 'valdat' (default is the full input sample 'datain'). 
#' Other criteria are possible (cf. parameter description of 'crit'). The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'.\cr
#' The substructure of the observations used for subsampling in modelling is specified by the list 'Struc' which consists of the factor variable 'name' representing the substructure,
#' the name 'check' of the factor variable with the information about the categories of the substructure, and the matrix 'labs' which specifies the values of 'check'
#' corresponding to two categories in its rows, i.e. in 'labs[1,]' and 'labs[2,]'. The names of the categories have to be specified by \code{rownames(labs)}.\cr
#' See parameter description of 'Struc' for its specification for 'vers="b"' and 'indrep > 0'.\cr
#' The number of predictors 'Pit' to be included in the model and the number of elements of the substructure 'Eit' have to be specified (lists allowed), and 
#' undersampling of the categories of 'classname' can be controlled by 'undersamp=TRUE/FALSE'.\cr
#' In structured subsampling, 'N' repetitions of subsampling of the variable 'name' with 'Eit' different elements of each category in 'check' are realized. If 'Eit' is a list, each entry is employed individually. If 'Eit' is larger than the maximum available number of elements with a certain value of 'check', the maximum possible number of elements is used.\cr
#' Four different versions of structured subsampling exist: \cr
#' a) just of the elements in the substructure (possibly with additional undersampling) with parameters 'N' and 'Eit',\cr
#' b) just of the predictors with parameters 'Ni' and 'Pit', \cr
#' c) of the predictors and for each subset of predictors subsampling of the elements of the substructure (possibly with additional undersampling) 
#' with parameters 'N', 'Ni', 'Eit', and 'Pit', and\cr
#' d) of the elements of the substructure (possibly with additional undersampling) and for each of these subsets subsampling of the predictors 
#' with the same parameters as version c).\cr
#' Sampling of the elements of the substructure can be influenced by using weights of the elements ('weights=TRUE') according to the number of appearances of the smaller 
#' class of 'classnames'. This way, elements with more realisations in the smaller class are preferred in modelling.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr\cr
#' The parameter 'indrep' indicates repeated measurement situations ('indrep > 0', only implemented for 'crit="ba"'); default = 0.\cr
#' Repeated measurements are multiple measurements of the same variable taken on the same subjects (or objects) either under different conditions or over two or more time periods.\cr
#' The variable with the repeatedly observed subjects (or objects) has to be specified by 'name' in 'Struc'. Only one value of 'classname' is allowed for each value of 'Struc$name'.
#' In case of 'indrep > 0' it is automatically assumed that the same number of 
#' subjects (or objects) of the two classes under study have to be used for model building. Possible such numbers can be specified by 'Eit'.\cr
#' For indrep=1, models are optimized for individual observations and the classes of the substructure elements are subsequently determined via the parameter 'thr'.\cr
#' For indrep=2, models are optimized so that the criterion "ba" is maximal for the substructure elements taking 'thr' into account.\cr
#' For indrep > 0 and 'valdat' unequal 'datain', the variable 'repvar' defines the substructure for 'valdat', length = dim(valdat)[1] necessary; default = NA'     
#'
#' @usage PrInDTCstruc(datain,classname,ctestv=NA,Struc=NA,vers="d",weight=FALSE,
#'                 Eit=NA,Pit=NA,N=99,Ni=99,undersamp=TRUE,crit="ba",ktest=0,
#'                 stest=integer(length=0),conf.level=0.95,indrep=0,
#'                 minsplit=NA,minbucket=NA,repvar=NA,valdat=datain,thr=0.5)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param Struc = list(name,check,labs), cf. description for explanations; Struc not needed for vers="b"; for indrep=1, Struc = list(name);
#' @param vers Version of structured subsampling: "a", "b", "c", "d", cf. description; \cr
#'     default = "d"
#' @param weight Weights to be used for subsampling of elements of substructure (logical, TRUE or FALSE);
#'     default = FALSE
#' @param Eit List of number of elements of substructure (integers); \cr
#'     default = c((Cl-4):Cl), Cl = maximum number of elements in both categories 
#' @param Pit List of number of predictors (integers)\cr
#'     default = c(max(1,(D-3)):D), D = maximum number of predictors
#' @param N Number of repetitions of subsampling from substructure (integer)\cr
#'      default = 0 for vers="b", = 99 otherwise; if vers="b", any input is overwritten by the default
#' @param Ni Number of repetitions of subsampling from predictors\cr
#'      default = 0 for vers="a", = 99 for vers="b", = N otherwise; if vers="a", any input is overwritten by the default
#' @param undersamp Undersampling of categories of 'classname' to be used ((logical, TRUE or FALSE)\cr
#'     default = TRUE; if indrep=1 or vers="b", undersamp = FALSE
#' @param crit Optimisation criterion: "ba" for balanced accuracy, "bat" for balanced accuracy on test sets, "ta" for test accuracy, 
#'     "tal" for test accuracy of continuing parts of length 'ktest' in substructure elements 'stest';\cr
#'     default = "ba"
#' @param ktest Length of continuing parts to be tested (for crit="tal");\cr
#'     default = 0
#' @param stest Part of substructure to be tested (for crit="tal")(integer vector);\cr
#'     default = integer vector of length 0
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param indrep Indicator for repeated measurements, i.e. more than one observation with the same class for each element;\cr 
#'     for indrep=1, Struc=list(name) only; default = 0 
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#' @param valdat validation data; default = datain
#' @param repvar Values of variable defining the substructure corresponding to valdat in the case of repeated measurements, length = dim(valdat)[1] necessary; default = NA
#' @param thr threshold for element classfication: minimum percentage of correct class entries; default = 0.5
#'
#' @return
#' \describe{
#'   \item{modbest}{Best tree} 
#'   \item{interp}{Number of interpretable trees, overall number of trees} 
#'   \item{dmax}{Number of predictors in training set for best tree} 
#'   \item{ntmax}{Size of training set for best tree} 
#'   \item{acc1}{Accuracy of best tree on large class} 
#'   \item{acc2}{Accuracy of best tree on small class }  
#'   \item{bamax}{Balanced accuracy of best tree} 
#'   \item{tamax}{Test accuracy of best tree}
#'   \item{kumin}{Number of elements with misclassified parts longer than 'ktest' for best tree}
#'   \item{elems}{Elements with long misclassified parts for best tree}
#'   \item{mindlong}{Indices of long misclassified parts for best tree}
#'   \item{ind1max}{Elements of 1st category of substructure used by best tree} 
#'   \item{ind2max}{Elements of 2nd category of substructure used by best tree} 
#'   \item{indmax}{Predictors used by best tree} 
#'   \item{bestTrain}{Training set for best tree} 
#'   \item{bestTest}{Test set for best tree} 
#'   \item{labs}{labs from Struc} 
#'   \item{lablarge}{Label of large class} 
#'   \item{labsmall}{Label of small class} 
#'   \item{vers}{Version used for structured subsampling} 
#'   \item{acc1E}{Accuracy of large class on Elements of best tree} 
#'   \item{acc2E}{Accuracy of small class on Elements of best tree}  
#'   \item{bamaxE}{Balanced accuracy of best tree on Elements} 
#'   \item{nam1}{Names of misclassified Elements of large class}
#'   \item{nam2}{Names of misclassified Elements of small class}
#'   \item{thr}{Threshold for element classification}
#' }
#'
#' @details
#' See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.1, for further information.
#'
#' Standard output can be produced by means of \code{print(name$besttree)} or just \code{name$besttree} as well as \code{plot(name$besttree)} where 'name' is the output data 
#' frame of the function.\cr
#
#' @export PrInDTCstruc
#'
#' @examples
#' data <- PrInDT::data_zero
#' data <- na.omit(data) # cleaned full data: no NAs
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
#' # substructure
#' name <- PrInDT::participant_zero
#' check <- "data$ETH"
#' labs <- matrix(1:6,nrow=2,ncol=3)
#' labs[1,] <- c("C1a","C1b","C1c")
#' labs[2,] <- c("C2a","C2b","C2c")
#' rownames(labs) <- c("Children 1","Children 2")
#' Struc <- list(name=name,check=check,labs=labs)
#' out <- PrInDTCstruc(data,"real",ctestv,Struc,vers="c",weight=TRUE,N=5,Pit=5,conf.level=0.99)
#' out
#' plot(out)
#' # indrep = 1
#' Struc <- list(name=name)
#' out <- PrInDTCstruc(data,"real",Struc=Struc,vers="c",Pit=5,Eit=5,N=5,crit="ba",indrep=1,
#'            conf.level=0.99)
#'
PrInDTCstruc <- function(datain,classname,ctestv=NA,Struc=NA,vers="d",weight=FALSE,Eit=NA,Pit=NA,N=99,Ni=99,undersamp=TRUE,
  crit="ba",ktest=0,stest=integer(length=0),conf.level=0.95,indrep=0,minsplit=NA,minbucket=NA,repvar=NA,valdat=datain,thr=0.5){
#
## input check
if ( typeof(datain) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character"))
    || !(typeof(Struc) %in% c("logical","list")) || typeof(vers) != "character" || typeof(weight) != "logical" 
    || !(typeof(Eit) %in% c("logical","integer","double")) || !(typeof(Pit) %in% c("logical","integer","double")) || !(typeof(N) %in% c("logical","integer","double")) 
    || !(typeof(Ni) %in% c("logical","integer","double")) ||  typeof(undersamp) != "logical" ||  typeof(crit) != "character" ||  !(typeof(ktest) %in% c("integer","double"))
    || !(typeof(stest) %in% c("integer","double")) || !(0 < conf.level & conf.level <= 1) || !(typeof(indrep) %in% c("integer","double"))
    || !(typeof(minsplit) %in% c("logical","double")) || !(typeof(minbucket) %in% c("logical", "double")) || typeof(valdat) != "list" || typeof(thr) != "double"){
  stop("irregular input")
}
  if (!(all(names(datain) %in% names(valdat)))){
    stop("validation data variables unequal input data variables")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
if (indrep > 0 & crit != "ba"){
  stop("irregular input: repeated measurements only implemented for ('crit = ba')")
}
if (indrep == 0 & vers != "b"){
  check <- eval(parse(text = Struc$check))
  if (length(check) == 0){
    cat("\n irregular input: check variable not in data set\n")
    return()
  }
  if (is.factor(check) != TRUE){
    cat("\n Error: 'Struc$check' has to be a factor variable\n")
    return()
  }
}
#
## re-naming
dataname <- as.character(substitute(datain))
data <- datain  
name <- Struc$name
if (is.factor(name) != TRUE){
  cat("\n Error: 'Struc$name' has to be a factor variable\n")
  return()
}
names(data)[names(data)==classname] <- "class"
names(valdat)[names(valdat)==classname] <- "class"
if (indrep > 0 & all(is.na(repvar)) == TRUE & identical(data,valdat)) {
  repv <- name
} else {
  repv <- repvar
}
if (!(identical(levels(data$class),levels(valdat$class)))){
  stop("levels of input class variable unequal levels of validation class variable")
}
# indrep = 2 indicates direct optimization of ba for substructure elements
if (indrep > 0){
  Struc$check <- "data$class"
  check <- eval(parse(text = Struc$check))
  labs <- matrix(1:2,nrow=2,ncol=1)
  labs[,1] <- levels(data$class) # as.character(unique(data$class))
  rownames(labs) <- as.character(labs[,1])
  undersamp <- FALSE  
  weight <- FALSE
}
if ( crit == "ba" & indrep > 0 ) {
  if (all(is.na(repv)) == FALSE & length(repv) != dim(valdat)[1] ){
    stop("irregular input: no. of observations different in substructure and validation set")
  }
}
if ( length(name) != dim(data)[1] | length(name) != length(check)){
  stop("irregular input: no. of observations different in substructure and data set")
}
## indrep dependencies
## Repeated measurement analysis applicable?
if (indrep > 0){
  conti <- table(Struc$name,data$class)
  if ( (sum(conti[,1] > 0) + sum(conti[,2] > 0)) > dim(conti)[1]){
    message("\n","Substructure variable has more than one class per value",
             "\n","Repeated measurement analysis not applicable","\n")
    return()
  }
}
D <- dim(data)[2] - 1
if (all(is.na(Pit)) == TRUE){
  Pit <- c(max(2,(D-3)):D)  
}
if (max(Pit) > D){stop("Too many predictors requested: Maximum is ",D)}
#
## 
## vers dependencies
if (vers == "a"){
  Ni <- 0
  Pit <- D
}
if (vers == "b"){
  N <- 0
#  Ni <- 99
  Eit <- 0
  pS <- 0
  pE <- 0
  labs <- NA
  undersamp <- FALSE
}
if (crit == "tal"){
  undersamp <- FALSE
  weight <- FALSE
}
if (vers != "b"){
  if (Ni > 1) {Ni <- N} ## ?????????
  labs <- Struc$labs
  if (anyNA(Struc[[1]]) != TRUE & indrep == 0){
    if (Struc$check == paste0(dataname,"$",names(data)[names(data)==classname])) { 
      check <- eval(parse(text = Struc$check)) 
      Struc$check <- "data$class"
      indrep <- 1
      undersamp <- FALSE
      weight <- FALSE
    }
  }
## 
  n_class1 <- table(data$class)[1] # no. of elements of class 1
  n_class2 <- table(data$class)[2] # no. of elements of class 2
  if (min(n_class1,n_class2) == 0){
    stop("irregular input: only 1 class")
  }
#print(table(data$class))
  if (n_class1 < n_class2){
    # relevel of classes if smaller class first
    valdat$class <- stats::relevel(valdat$class, levels(data$class)[2]) # larger class now first
    data$class <- stats::relevel(data$class, levels(data$class)[2]) # larger class now first
  #  lablarge <- names(table(data$class))[1] # class 1 = large
  #  labsmall <- names(table(data$class))[2] # class 2 = small
    n_class1 <- table(data$class)[1] # no. of elements of larger class 1
    n_class2 <- table(data$class)[2] # no. of elements of smaller class 2
  }
  if (indrep > 0){
    pS <- table(Struc$name,data$class)[,1] #
    pS <- pS[pS != 0]
    pS <- pS / pS
#  pS[pS != 0] <- 1
#print(length(pS))
    pE <- table(Struc$name,data$class)[,2] 
    pE <- pE[pE != 0]
    pE <- pE / pE
#print(length(pE))
  } else {
#    check <- eval(parse(text = Struc$check))  ## check not equal classname!!!
    p <- table(Struc$name,data$class)[,2] # /sum(table(Struc$name,data$class)[,2]) * 100
    pS <- p[unique(Struc$name[check %in% Struc$labs[1,]])] 
    pE <- p[unique(Struc$name[check %in% Struc$labs[2,]])]
    pS[pS == 0] <- 5e-10
    pE[pE == 0] <- 5e-10
  }
  ## weights 
  if (weight == FALSE){
    pS[pS > 5e-10] <- 1 
    pE[pE > 5e-10] <- 1
  }
  if (all(is.na(Eit)) == TRUE){
    Cl <- min(length(pS),length(pE))
    Eit <- c((Cl-4):Cl)
  }
}
if (max(Eit) > min(length(pS),length(pE))){stop("Too many substructure elements requested: Maximum is ",min(length(pS),length(pE)))}
#
## PrInDT function calls 
if ((vers == "d") | (vers == "a")){
  out <- PrInDTstruc1(data,classname,ctestv=ctestv,name=name,check=check,labs=labs,N=N,Ni=Ni,Eit=Eit,Pit=Pit,crit=crit,ktest=ktest,stest=stest,p1=pS,p2=pE,undersamp=undersamp,co=conf.level,minsplit=minsplit,minbucket=minbucket,repvar=repv,valdat=valdat,indrep=indrep,thr=thr)
  nam1 <- out$nam1
  nam2 <- out$nam2
}
if ((vers == "c") | (vers == "b")){
  out <- PrInDTstruc2(data,classname,ctestv=ctestv,name=name,check=check,labs=labs,N=N,Ni=Ni,Pit=Pit,Eit=Eit,crit=crit,ktest=ktest,stest=stest,p1=pS,p2=pE,undersamp=undersamp,co=conf.level,minsplit=minsplit,minbucket=minbucket,repvar=repv,valdat=valdat,indrep=indrep,thr=thr) 
  nam1 <- out$nam1
  nam2 <- out$nam2
}
#
## indrep = 1: 'ba' calculation for elements for models optimized for individual observations 
acc1E <- 0
acc2E <- 0
bamaxE <- 0
if (crit == "ba" & indrep == 1){
  nam1 <-  list()
  nam2 <- list()
  pred <- predict(out$modbest,newdata=valdat)  ## data???
  ch <- table(Struc[[1]],pred)
  conti <- cbind(table(Struc[[1]],valdat$class),ch)  ## data???
  no1 <- sum(conti[,1] > 0)
  no2 <- sum(conti[,2] > 0)
  for (i in 1:dim(conti)[1]){
    if (conti[i,2] > 0 & conti[i,4] <= (conti[i,2]*thr)){ 
      nam2 <- c(nam2,rownames(ch)[i])
    }
    if (conti[i,1] > 0 & conti[i,3] < (conti[i,1]*thr)){ 
      nam1 <- c(nam1,rownames(ch)[i])
    }
  }
  acc2E <- 1- length(nam2) / no2
  acc1E <- 1 - length(nam1) / no1  
  bamaxE <- (acc1E + acc2E)/2
  if (length(nam1) == 0) {nam1 <- "-"}
  if (length(nam2) == 0) {nam2 <- "-"}
}
#
## results
lablarge <- names(table(data$class))[1] # class 1 = large
labsmall <- names(table(data$class))[2] # class 2 = small 
result <- list(modbest = out$modbest, interp=out$interp, dmax=out$dmax, ntmax = out$ntmax, acc1 = out$acc1, acc2 = out$acc2, 
     bamax=out$bamax,tamax=out$tamax,kumin=out$kumin,elems=out$elems,mindlong=out$mindlong,
     ind1max=out$ind1max, ind2max=out$ind2max, indmax=out$indmax, bestTrain=out$gmaxTrain, bestTest=out$gmaxTest,labs=labs,
     lablarge=lablarge,labsmall=labsmall,crit=crit,vers=vers,indrep=indrep,acc1E = acc1E, acc2E =  acc2E, bamaxE = bamaxE, 
     nam1 = nam1, nam2 = nam2, thr=thr)
class(result) <- "PrInDTCstruc"
result
}
#' @export
print.PrInDTCstruc <- function(x,...){
  cat("Number of interpretable trees: ",x$interp[1]," of ",x$interp[2]," trees","\n","\n")
  cat("Best interpretable tree")
  print(x$modbest)
  cat("\n","Number of predictors in training set for best tree: ",x$dmax,"\n")
  cat("\n","Size of training set for best tree: ",x$ntmax,"\n")
  cat("\n","Predictors in model construction: ",x$indmax,"\n")
  if (x$vers != "b"){
  cat("\n","Elements used in model construction")
#  cat("\n","Number of observations of",rownames(x$labs)[1],"in model construction","\n")  ## labs falsch???
  if (x$indrep == 0) {
    cat("\n","Number of observations of",rownames(x$labs)[1],"in model construction: ", 
          sum(x$bestTrain$check[x$bestTrain$SubStruc %in% x$ind1max] %in% x$labs[1,]),"\n")
  } else {
    cat("\n","Number of observations of",rownames(x$labs)[1],"in model construction : ", 
          sum(x$bestTrain$check[x$bestTrain$SubStruc %in% x$ind1max] > 0),"\n")
  }
  Elements <- cbind(as.character(x$bestTrain$check[x$bestTrain$SubStruc %in% x$ind1max]),
    as.character(x$bestTrain$SubStruc[x$bestTrain$SubStruc %in% x$ind1max]))
  print(table(Elements,exclude=x$labs))
  if (x$indrep == 1){
    print(table(Elements,exclude=x$ind1max))
  } else {
    print(table(Elements,exclude=c(x$ind1max,x$labs[2,])))
  }
#  cat("Number of observations of",rownames(x$labs)[2],"in model construction","\n")
  if (x$indrep == 0) {
    cat(" Number of observations of",rownames(x$labs)[2],"in model construction: ",
           sum(x$bestTrain$check[x$bestTrain$SubStruc %in% x$ind2max] %in% x$labs[2,]),"\n")
  } else {
    cat(" Number of observations of",rownames(x$labs)[2],"in model construction: ",
           sum(x$bestTrain$check[x$bestTrain$SubStruc %in% x$ind2max] > 0),"\n")
  }
  Elements <- cbind(as.character(x$bestTrain$check[x$bestTrain$SubStruc %in% x$ind2max]),
    as.character(x$bestTrain$SubStruc[x$bestTrain$SubStruc %in% x$ind2max]))
  print(table(Elements,exclude=x$labs))
  if (x$indrep > 0){
    print(table(Elements,exclude=x$ind2max))
  } else {
    print(table(Elements,exclude=c(x$ind2max,x$labs[1,])))
  }
#  cat("\n","Test accuracy of best tree","\n")
#  class.pred <- predict(x$modbest,newdata=x$bestTest)
#  cat(" ",sum(class.pred == x$bestTest$class) / length(class.pred))
}
#  cat("Accuracy of best tree")
#  class.pred <- predict(x$outmax$treeAll,newdata=data)
# print(sum(class.pred == data$class) / length(class.pred))
if (x$crit == "ba"){
  accs <- c(round(x$acc1,digits=6),round(x$acc2,digits=6),round(x$bamax,digits=6))
  names(accs) <- c(x$lablarge,x$labsmall,"balanced")
  if (x$indrep < 2){
    cat("\n","Validation sample accuracies of best tree (observations) ","\n")
    print(accs,row.names=FALSE)
  } else {
    cat("\n","Validation sample accuracies of best tree (Elements) ","\n")
    print(accs,row.names=FALSE)
  }    
#  cat("\n","Validation sample accuracies of best tree (observations)","\n",x$lablarge,"\t",x$labsmall,"\t","balanced","\n")
} else {
  accs <- c(round(x$acc1,digits=6),round(x$acc2,digits=6),round(x$bamax,digits=6))
  names(accs) <- c(substr(x$lablarge,1,12),substr(x$labsmall,1,12),"balanced")
  cat("\n","Test sample accuracies of best tree (observations) ","\n")
  print(accs,row.names=FALSE)
#  cat("\n","Test accuracies of best tree","\n"," ",substr(x$lablarge,1,12),"\t",substr(x$labsmall,1,12),"\t"," balanced","\n")
}
#  cat(" ",c(round(x$acc1,4),"\t",round(x$acc2,4),"\t",round(x$bamax,4)),"\n")
if (x$crit == "ta" | x$crit == "tal"){
  cat("\n","Overall test accuracy of best tree: ",x$tamax,"\n")
}
if (x$crit == "tal"){
  cat("\n","Optimal number of long misclassified parts: ",x$kumin)
  cat("\n","Elements with long misclassified parts:")
  cat("\n",x$elems,"\n")
}
if (x$indrep == 1 & x$crit == "ba"){
  cat("\n","Repeated measurements: Element classification: Threshold = ",x$thr)
  accs <- c(round(x$acc1E,digits=6),round(x$acc2E,digits=6),round(x$bamaxE,digits=6))
  names(accs) <- c(x$lablarge,x$labsmall,"balanced")
  cat("\n","Input sample accuracies of best tree (Elements)","\n")
  print(accs,row.names=FALSE)
#  cat("\n","Input sample accuracies of best tree (Elements)","\n",x$lablarge,"\t",x$labsmall,"\t","balanced","\n") 
#  cat(" ",c(round(x$acc1E,4),"\t","  ",round(x$acc2E,4),"\t","  ",round(x$bamaxE,4)),"\n\n")
  cat("Wrongly predicted Elements:","\n")
  cat(x$lablarge,"\n")
  cat(unlist(x$nam1),"\n")
  cat(x$labsmall,"\n")
  cat(unlist(x$nam2),"\n")
}
if (x$indrep == 2 & x$crit == "ba"){
    cat("Wrongly predicted Elements:","\n")
  cat(x$lablarge,"\n")
  cat(unlist(x$nam1),"\n")
  cat(x$labsmall,"\n")
  cat(unlist(x$nam2),"\n")
}
}
#' @export
plot.PrInDTCstruc <- function(x,...){
  plot(x$modbest,main="Best interpretable tree")
}

#' Multiple label classification based on all observations
#'
#' @description Multiple label classification based on all observations. We consider two ways of modeling (Binary relevance modeling, 
#' dependent binary modeling) and three ways of model evaluation: single 
#' assessment, joint assessment, and true prediction (see the Value section for more information).\cr
#' Interpretability is checked (see ctestv).\cr
#' Variables should be arranged in 'datain' according to indices specified in 'indind', 'indaddind', and 'inddep'.\cr
#' Please note that the dependent variables have to be specified as dummies, i.e. as 'absent' (value 0) or 'present' (value 1).\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' \strong{Reference}\cr Probst, P., Au, Q., Casalicchio, G., Stachl, C., and Bischl, B. 2017. Multilabel Classification with 
#' R Package mlr. arXiv:1703.08991v2
#'
#' @usage PrInDTMulabAll(datain,classnames=NA,ctestv=NA,conf.level=0.95,indind=NA,
#'                indaddind=NA,inddep,minsplit=NA,minbucket=NA)
#' 
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classnames names of class variables (character vector)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'   default = 0.95
#' @param indind indices of independent variables
#' @param indaddind indices of additional predictors used in the case of dependent binary relevance modeling
#' @param inddep indices of dependent variables
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return
#' \describe{ 
#' \item{accabr}{model errors for Binary Relevance (single assessment) - only independent predictors are used for modeling one label at a time,  
#' the other labels are not used as predictors. The classification rules are trained on all observations. 
#' As the performance measure for the resulting classification rules, the balanced accuracy of the models for each individual label is employed.}
#' \item{errabin}{combined error for Binary Relevance (joint assessment) - the best prediction models for the different labels are combined to assess the 
#' combined prediction. The 01-accuracy counts a label combination as correct only if all labels are correctly predicted. 
#' The hamming accuracy corresponds to the proportion of labels whose value is correctly predicted.}
#' \item{accadbr}{model errors in Dependent Binary Relevance (Extended Model) (single assessment) - each label is trained by means of an extended model which 
#' not only includes the independent predictors but also the other labels. For these labels the truly observed values are used for estimation 
#' and prediction. In the extended model, further labels, which are not treated as dependent variables, can be used as additional predictors.}
#' \item{erraext}{combined errors for Dependent Binary Relevance (Extended Model) (joint assessment) }
#' \item{erratrue}{combined errors for Dependent Binary Relevance (True Prediction) - in the prediction phase, the values 
#' of all modeled labels are first predicted by the independent predictors only (see Binary Relevance) and then the predicted labels are used in  
#' the estimated extended model in a 2nd step to ultimately predict the labels.}
#' \item{coldata}{column names of input data}
#' \item{inddep}{indices of dependent variables (labels to be modeled)}
#' \item{treeabr}{list of trees from Binary Relevance modeling, one tree for each label; refer to an individual tree as \code{treeabr[[i]]}, 
#'                 i = 1, ..., no. of labels}
#' \item{treeadbr}{list of trees from Dependent Binary Relevance modeling, one for each label; refer to an individual tree as \code{treeadbr[[i]]}, 
#'                 i = 1, ..., no. of labels}
#' }
#'
#' @details
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)}  where 'name' is the output data 
#' frame of the function.\cr
#' The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
#' \code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
#'
#' @export PrInDTMulabAll
#'
#' @examples
#' data <- PrInDT::data_land # load data
#' dataclean <- data[,c(1:7,23:24,11:13,22,8:10)]  # only relevant features
#' indind <- c(1:9) # original predictors
#' indaddind <- c(10:13) # additional predictors
#' inddep <- c(14:16) # dependent variables
#' dataclean <- na.omit(dataclean)
#' ctestv <- NA
#' ##
#' # Call PrInDTAll: language by language
#' ##
#' outmultAll <- PrInDTMulabAll(dataclean,colnames(dataclean)[inddep],ctestv,conf.level=0.95,
#'                      indind,indaddind,inddep)
#' outmultAll
#' plot(outmultAll)
#'
#' @import party
#' @import stats
#'
PrInDTMulabAll <- function(datain,classnames=NA,ctestv=NA,conf.level=0.95,indind=NA,indaddind=NA,inddep,minsplit=NA,minbucket=NA){
  ## input check
  if ( typeof(datain) != "list" || !(typeof(classnames) %in% c("character","logical"))  || !(typeof(ctestv) %in% c("logical", "character")) || 
      !(0 < conf.level & conf.level <= 1) || !(typeof(indind) %in% c("integer", "double","logical")) || 
      !(typeof(indaddind) %in% c("integer", "double","logical")) || !(typeof(inddep) %in% c("integer", "double")) || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
####
####
## Modeling with no additional independent and no dependent variable as predictors
####
  data <- datain
  if (all(is.na(classnames)) == TRUE){
    classnames <- colnames(data)[inddep]
  }
  if (all(is.na(indind)) == TRUE & all(is.na(indaddind)) == TRUE){
    indind <- c(1:dim(data)[2])[-inddep]
  }  
  if (all(is.na(indind)) == TRUE & all(is.na(indaddind)) != TRUE){
    indind <- c(1:dim(data)[2])[-c(inddep,indaddind)]
  }  
 print(indind)
#  message("\n")
  message("*** Binary relevance ***")
#  message("\n")
  message("Modeling the labels:")
  accabr <- matrix(0,nrow=length(inddep),ncol=1)
  rownames(accabr) <- vector(mode="character",length=length(inddep))
  colnames(accabr) <- "balanced"
  ctpreds <- matrix(0,ncol=length(inddep),nrow=dim(datain)[1])
  unequal <- matrix(TRUE,ncol=length(inddep),nrow=dim(datain)[1])
  hsum <- 0
  for (i in 1:length(inddep)){
    x <- colnames(datain)[inddep[i]]
    levels(data[,x]) <- c(levels(data[,x]),x)
    data[datain[,x] == 1,x] <- x
    data[,x] <- droplevels(data[,x])
#    message("\n")
    message(paste0("  ",colnames(datain)[inddep[i]]))
    out <- PrInDTAll(data[,c(indind,inddep[i])],x,ctestv,conf.level,minsplit=minsplit,minbucket=minbucket)
    rownames(accabr)[i] <- x
    if (i == 1) {
      treeabr <- list(out$treeAll)
    } else {
      treeabr <- c(treeabr,out$treeAll)
    }
    accabr[i] <- out$baAll
    ctpreds[,i] <- stats::predict( treeabr[[i]],newdata=data[,c(indind,inddep[i])] )
    unequal[,i] <- as.integer(data[,inddep[i]]) != ctpreds[,i]
    if (table(data[,inddep[i]])[1] < table(data[,inddep[i]])[2] ){
      unequal[,i] <- as.integer(data[,inddep[i]]) == ctpreds[,i]
    }
    hsum <- hsum + sum(unequal[,i])
  }
# Joint assessment: all
  errabin <- c(0,0)
  # 01 loss
  err01 <- rowSums(unequal)
  err01[err01 > 1] <- 1
  errabin[1] <- sum(err01)
  errabin[1] <- errabin[1] / dim(data)[1]
  # hamming loss
  errabin[2] <- hsum / (3*dim(data)[1])
  names(errabin) <- c("01-accuracy", "hamming-accuracy")
####
## Modeling with additional independent and dependent variable as predictors
####
#  message("\n")
  message("\n","*** Dependent binary relevance: Extended model ***")
#  message("\n")
  message("Modeling the labels:")
  accadbr <- matrix(0,nrow=length(inddep),ncol=1)
  rownames(accadbr) <- vector(mode="character",length=length(inddep))
  colnames(accadbr) <- "balanced"
  hsum <- 0
  for (i in 1:length(inddep)) {
    x <- colnames(datain)[inddep[i]]
#   message("\n")
    message(paste0("  ",x))
    out <- PrInDTAll(data,x,ctestv,conf.level,minsplit=minsplit,minbucket=minbucket)
    rownames(accadbr)[i] <- x
    if (i == 1) {
      treeadbr <- list(out$treeAll)
    } else {
      treeadbr <- c(treeadbr,out$treeAll)
    }
    accadbr[i] <- out$baAll
    ctpreds[,i] <- stats::predict( treeadbr[[i]],newdata=data )
#print(ctpreds[1:20,i])
#print(as.integer(data[1:20,inddep[i]]))
#print(table(data[,inddep[i]]))
    unequal[,i] <- as.integer(data[,inddep[i]]) != ctpreds[,i]
    if (table(data[,inddep[i]])[1] < table(data[,inddep[i]])[2] ){
      unequal[,i] <- as.integer(data[,inddep[i]]) == ctpreds[,i]
    }
#print(sum(unequal[,i]))
    hsum <- hsum + sum(unequal[,i])
  }
#  message("\n")
# Joint assessment: all
  erraext <- c(0,0)
  # 01 loss
  err01 <- rowSums(unequal)
  err01[err01 > 1] <- 1
  erraext[1] <- sum(err01)
  erraext[1] <- erraext[1] / dim(data)[1]
  # hamming loss
  erraext[2] <- hsum / (3*dim(data)[1])
  names(erraext) <- c("01-accuracy", "hamming-accuracy")
  ##
  ## nested
  ##
# Dependent binary relevance: True prediction
  datain2 <- data
  ctpreds2 <- matrix(0,ncol=length(inddep),nrow=dim(data)[1])
  unequal2 <- matrix(TRUE,ncol=length(inddep),nrow=dim(data)[1])
  accabrt <- matrix(0,nrow=length(inddep),ncol=1)
  rownames(accabrt) <- classnames
  hsum2 <- 0
  for (i in 1:length(inddep)){
    #  p <- ctpreds[,i] - 1
    datain2[,inddep[i]] <- stats::predict(treeabr[[i]],newdata=data) # as.factor(p)
    if (table(data[,inddep[i]])[1] < table(data[,inddep[i]])[2] ){
      datain2[,inddep[i]] <- relevel(datain2[,inddep[i]],"0")
    }
  }
  for (i in 1:length(inddep)){
#    rownames(accabrt)[i] <- colnames(data)[inddep[i]] ### ????
    ctpreds2[,i] <- stats::predict(treeadbr[[i]],newdata=datain2)
    tabdbrt <- table(data[,inddep[i]],ctpreds2[,i])
#print(tabdbrt)
    if (length(colnames(tabdbrt)) == 2){
      if (table(data[,inddep[i]])[1] >= table(data[,inddep[i]])[2] ){
        unequal2[,i] <- as.integer(data[,inddep[i]]) != ctpreds2[,i]
        accabrt[i] <- ( tabdbrt[1,1] / sum(tabdbrt[1,]) + tabdbrt[2,2] / sum(tabdbrt[2,]) ) / 2
      } else {
        unequal2[,i] <- as.integer(data[,inddep[i]]) == ctpreds2[,i]
        accabrt[i] <- (tabdbrt[1,2] / sum(tabdbrt[1,]) + tabdbrt[2,1] / sum(tabdbrt[2,]) ) / 2
      }
    } else {
    if (table(data[,inddep[i]])[1] >= table(data[,inddep[i]])[2] ){
        unequal2[,i] <- as.integer(data[,inddep[i]]) != ctpreds2[,i]
        accabrt[i] <- 0.5  ## tabdbrt[2,1] / sum(tabdbrt[,1])    ##  ???
      } else {
        unequal2[,i] <- as.integer(data[,inddep[i]]) == ctpreds2[,i]
        accabrt[i] <- 0.5  ## tabdbrt[1,1] / sum(tabdbrt[,1])    ## ???
      }
    } 
#    if (table(data[,inddep[i]])[1] >= table(data[,inddep[i]])[2] ){
#      unequal2[,i] <- as.integer(data[,inddep[i]]) != ctpreds2[,i]
#      accabrt[i] <- ( tabdbrt[1,1] / sum(tabdbrt[1,]) + tabdbrt[2,2] / sum(tabdbrt[2,]) ) / 2
#    } else {
#      unequal2[,i] <- as.integer(data[,inddep[i]]) == ctpreds2[,i]
#      accabrt[i] <- (tabdbrt[1,2] / sum(tabdbrt[1,]) + tabdbrt[2,1] / sum(tabdbrt[2,]) ) / 2
#    }
    hsum2 <- hsum2 + sum(unequal2[,i])
  }
  erratrue <- c(0,0)
  # 01 loss
  err01 <- rowSums(unequal2)
  err01[err01 > 1] <- 1
  erratrue[1] <- sum(err01)
  erratrue[1] <- erratrue[1] / dim(data)[1]
  # hamming loss
  erratrue[2] <- hsum2 / (3*dim(data)[1])
  names(erratrue) <- c("01-accuracy", "hamming-accuracy")
  result <- list(accabr = accabr, errabin = errabin, accadbr = accadbr, erraext = erraext, erratrue = erratrue, 
     coldata = colnames(datain), inddep = inddep, treeabr = treeabr, treeadbr = treeadbr,accabrt=accabrt)
  class(result) <- "PrInDTMulabAll"
  result
}
#' @export
print.PrInDTMulabAll <- function(x,...){
  cat("\n","Multi-label classification on full sample","\n")
  cat("\n")
  cat("*** Binary relevance ***")
  cat("\n")
  L <- length(x$inddep)
  for (i in 1:L){
    cat("\n")
    cat( "Binary Relevance: Best tree on subsamples for: ",x$coldata[x$inddep[i]] )
    print( x$treeabr[[i]] )
  }
  ### single assessment: all
  cat("\n\n")
  cat("Single assessment: all","\n")
  print(x$accabr)
  ### joint assessment all
  cat("\n")
  cat("Joint assessment: all")
  cat("\n","Assessment of models on full sample (labels are NOT predictors)","\n")
  print(1-x$errabin)
  cat("\n\n")
  cat("*** Dependent binary relevance: Extended Model ***")
  cat("\n")
  for (i in 1:L){
    cat("\n")
    cat( "Dependent Binary Relevance: Best tree on full sample for: ",x$coldata[x$inddep[i]] ) 
    print( x$treeadbr[[i]] )
  }
  ### single assessment: all
  cat("\n\n")
  cat("Single assessment: all","\n")
  print(x$accadbr)
  ### joint assessment all
  cat("\n")
  cat("Joint assessment: all")
  cat("\n","Assessment of models on full sample (labels are predictors)","\n")
  print(1-x$erraext)
  ##
  ## nested
  ##
  cat("\n\n")
  cat("*** Dependent binary relevance: True prediction ***")
  colnames(x$accabrt) <- "balanced"
  cat("\n\n")
  cat("Single assessment: all","\n")
  print(x$accabrt)
  cat("\n")
  cat("\n","Joint assessment all","\n")
  print(1-x$erratrue)
}
#' @export
plot.PrInDTMulabAll <- function(x,...){
L <- length(x$inddep)
  for (i in 1:L){
    plot( x$treeabr[[i]],main=paste0("Binary Relevance on full sample: Tree on all data for: ",x$coldata[x$inddep[i]]) )
  }
  for (i in 1:L){
    plot( x$treeadbr[[i]],main=paste0("Dependent Binary Relevance on full sample: Tree on all data for: ",x$coldata[x$inddep[i]]) )
  }
}

#' Check for forbidden split results in trees
#'
#' Check whether one of the character strings in the vector 'ctestv' appears as a split result in the conditional inference tree 'ct';
#' ctestv is a vector of character strings of forbidden split results.\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in 'variable 1' in a tree.
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     For an application, please refer to, e.g., the functions \code{\link{PrInDT}} and \code{\link{PrInDTreg}}.\cr
#'     If no restrictions exist, the default = NA is used.
#'
#' @usage FindSubstr(ct, ctestv)
#'
#' @param ct Tree to be checked
#' @param ctestv Vector with character strings of excluded split results
#'
#' @return
#' \describe{ 
#' \item{testt}{TRUE if any of the split results in 'ctestv' appears in 'ct'; FALSE otherwise}
#' }
#'
#' @importFrom utils capture.output
#'
#' @noRd
#'
FindSubstr <- function(ct,ctestv){
  ## input check
  if (typeof(ct) != "S4" || typeof(ctestv) != "character"){
    stop("irregular input")
  }
  ##
  J <- length(ctestv)
  ctvec <- utils::capture.output(print(ct))
  K <- length(ctvec)
  testc <- array(rep("FALSE",J*K),c(J,K))
  for(j in 1:J){             # loop over strings to be tested
     nj <- nchar(ctestv[j])
     for(k in 1:K){     # loop over parts of tree description
         nc <- nchar(ctvec[k]) - nj     # nc = length of part of tree - length of tested string
         if (nc >= 0) {
           if (grepl( "<=",ctestv[j],fixed=TRUE ) == FALSE){
             testc[j,k] <- grepl( ctestv[j],ctvec[k],fixed=TRUE )
             }
           else {
            if (grepl( ") ",ctvec[k],fixed=TRUE) == TRUE){
             if ( strsplit(strsplit(ctvec[k],"<=")[[1]][1],") ")[[1]][2] == strsplit(ctestv[j],"<=")[[1]][1] &
               as.numeric(gsub(".*?([0-9]+).*", "\\1",ctvec[k])) <= as.numeric(gsub(".*?([0-9]+).*", "\\1",ctestv[j])) ){
               testc[j,k] <- "TRUE"
             }
            }
           }
         }
     }
  }
  testt <- any(as.logical(testc))
  return(testt)
}

#' Structured subsampling for regression
#'
#' @description The function \code{\link{PrInDTRstruc}} applies structured subsampling for finding an optimal subsample to model
#' the relationship between the continuous variable 'regname' and all other factor and numerical variables
#' in the data frame 'datain' by means of 'M' repetitions of subsampling from a substructure and 'N' repetitions of subsampling from the predictors. 
#' The optimization citerion is the goodness of fit R2 on the validation sample 'valdat' (default = 'datain').  
#' The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'.\cr
#' The substructure of the observations used for subsampling is specified by the list 'Struc' which consists of the factor variable 'name' representing the substructure,
#' the name 'check' of the factor variable with the information about the categories of the substructure, and the matrix 'labs' which specifies the values of 'check'
#' corresponding to two categories in its rows, i.e. in 'labs[1,]' and 'labs[2,]'. The names of the categories have to be specified by \code{rownames(labs)}.\cr
#' In structured subsampling, 'M' repetitions of subsampling of the variable 'name' with 'Mit' different elements of each category in 'check' are realized. If 'Mit' is a list, each entry is employed individually. If 'Mit' is larger than the maximum available number of elements with a certain value of 'check', the maximum possible number of elements is used.\cr
#' The number of predictors 'Pit' to be included in the model and the number of elements of the substructure 'Mit' have to be specified (lists allowed).\cr
#' The percentages of involved observations and predictors can be controlled by the parameters 'pobs' and 'ppre', respectively.\cr
#' The parameter 'Struc' is needed for all versions of subsampling except "b". Four different versions of structured subsampling exist:\cr 
#' a) just of the elements in the substructure with parameters 'M' and 'Mit',\cr
#' b) just of the predictors with parameters 'N' and 'Pit',\cr
#' c) of the predictors and for each subset of predictors subsampling of the elements of the substructure with parameters 'M', 'N', 'Mit', 'Pit', 'pobs', and 'ppre', and\cr
#' d) of the elements of the substructure and for each of these subsets subsampling of the predictors with the same parameters as version c).\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#' Besides the maximal R2, the minimal MAE (Mean Absolute Error) is reported. \cr\cr
#' Repeated measurements can also be handled by this function (indrep=1). They are multiple measurements of the same variable taken on the same subjects (or objects) either under different conditions 
#' or over two or more time periods.\cr
#' The variable with the repeatedly observed subjects (or objects) is assumed to be 'name' in 'Struc'.\cr
#' The measure MAE is split according to the values of 'name'. 
#'
#' @usage PrInDTRstruc(datain,regname,ctestv=NA,Struc=NA,vers="d",M=NA,Mit=NA,N=99,Pit=NA,
#'                pobs=c(0.9,0.7),ppre=c(0.9,0.7),conf.level=0.95,minsplit=NA,minbucket=NA,
#'                valdat=datain,indrep=0)
#'
#' @param datain Input data frame with continuous target variable 'regname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param regname Name of target variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param Struc = list(name,check,labs), cf. description for explanations; Struc not needed for vers="b"
#' @param vers Version of structured subsampling: "a", "b", "c", "d", cf. description; \cr
#'     default = "d"
#' @param Mit List of number of elements of substructure (integers); \cr
#'     default = c((Cl-4):Cl), Cl = maximum number elements in both categories 
#' @param Pit List of number of predictors (integers)\cr
#'     default = c(max(1,(D-3)):D), D = maximum number of predictors
#' @param M Number of repetitions of subsampling from substructure (integer) in versions "a" and "d";\cr
#'      default = 99
#' @param N Number of repetitions of subsampling from predictors (integer) in versions "b" and "c";\cr
#'      default = 99
#' @param pobs Percentage(s) of observations for subsampling in versions "c" and "d";\cr
#'      default=c(0.9,0.7)
#' @param ppre Percentage(s) of predictors for subsampling in versions "c" and "d";\cr
#'      default=c(0.9,0.7)
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#' @param valdat Validation data; default = datain
#' @param indrep Indicator for repeated measurements, i.e. more than one observation with the same class for each element;\cr 
#'     indrep=1: Struc=list(name) only; default = 0 
#'
#' @return
#' \describe{
#'   \item{outmax}{Best tree} 
#'   \item{interp}{Number of interpretable trees, overall number of trees} 
#'   \item{ntmax}{Size of training set for best tree} 
#'   \item{R2max}{R squared of best tree}
#'   \item{R2sub}{Mean R squared of objects in substructure}
#'   \item{MAEmax}{MAE (Mean Absolute Error) of best tree}
#'   \item{MAEsub}{Mean MAE of objects in substructure}
#'   \item{ind1max}{Elements of 1st category of substructure used by best tree} 
#'   \item{ind2max}{Elements of 2nd category of substructure used by best tree} 
#'   \item{indmax}{Predictors used by best tree} 
#'   \item{gmaxTrain}{Training set for best tree} 
#'   \item{labs}{labs from Struc} 
#'   \item{vers}{Version used for structured subsampling} 
#'   \item{lMit}{Number of different numbers of substructure elements}
#'   \item{lPit}{Number of different numbers of predictors}
#'   \item{M}{Number of repetitions of selection of substructure elements} 
#'   \item{N}{Number of repetitions of selection of predictors}
#'   \item{indrep}{Indicator of repeated measurements: indrep=1} 
#' }
#'
#' @details
#' See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.4, for further information.
#'
#' Standard output can be produced by means of \code{print(name$besttree)} or just \code{name$besttree} as well as \code{plot(name$besttree)} where 'name' is the output data 
#' frame of the function.\cr
#
#' @export PrInDTRstruc
#'
#' @examples
#' data <- PrInDT::data_vowel
#' data <- na.omit(data)
#' CHILDvowel <- data$Nickname
#' data$Nickname <- NULL
#' data$syllables <- 3 - data$syllables
#' data$speed <- data$word_duration / data$syllables  ## NEW NEW
#' names(data)[names(data) == "target"] <- "vowel_length"
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
#' name <- CHILDvowel
#' check <- "data$ETH"
#' labs <- matrix(1:6,nrow=2,ncol=3)
#' labs[1,] <- c("C1a","C1b","C1c")
#' labs[2,] <- c("C2a","C2b","C2c")
#' rownames(labs) <- c("children 1","children 2")
#' Struc <- list(name=name,check=check,labs=labs)
#' outstruc <- PrInDTRstruc(data,"vowel_length",ctestv=ctestv,Struc=Struc,vers="d",
#'                   M=3,Mit=21,N=9,pobs=c(0.95,0.7),ppre=c(1,0.7),conf.level=0.99)
#' outstruc
#' plot(outstruc)
#'
PrInDTRstruc <- function(datain,regname,ctestv=NA,Struc=NA,vers="d",M=NA,Mit=NA,N=99,Pit=NA,pobs=c(0.9,0.7),ppre=c(0.9,0.7),conf.level=0.95,
                                         minsplit=NA,minbucket=NA,valdat=datain,indrep=0){
## input check
if ( typeof(datain) != "list" || typeof(regname) != "character" || !(typeof(ctestv) %in% c("logical", "character"))
    || !(typeof(Struc) %in% c("logical","list")) || typeof(vers) != "character" || !all(0 < pobs & pobs <= 1) || !all(0 < ppre & ppre <= 1) 
    || !(typeof(Mit) %in% c("logical","integer","double")) || !(typeof(Pit) %in% c("logical","integer","double")) || !(typeof(N) %in% c("logical","integer","double")) 
    || !(typeof(M) %in% c("logical","integer","double")) || !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) 
    || !(typeof(minbucket) %in% c("logical", "double")) || typeof(valdat) != "list" || !(typeof(indrep) %in% c("integer","double")) ) {
    stop("irregular input")
  }
  if (!(all(names(datain) %in% names(valdat)))){
    stop("validation data variables unequal input data variables")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
data <- datain
if ((vers == "a") | (vers == "c") | (vers == "d")){
  name <- Struc$name
  if (is.factor(name) != TRUE){
    cat("\n Error: 'Struc$name' has to be a factor variable\n")
    return()
  }
#check <- Struc$check
  check <- eval(parse(text = Struc$check))
  if (is.factor(check) != TRUE){
    cat("\n Error: 'Struc$check' has to be a factor variable\n")
    return()
  }
  if (length(check) == 0){
    cat("\n irregular input: check variable not in data set\n")
    return()
  }
  labs <- Struc$labs
  if (all(is.na(Mit)) == TRUE){stop("No number of elements in substructure specified")}
  lMit <- length(Mit)
} else {
  lMit <- NA
  gmaxTrain <- NA
}
##
if ((vers == "a") | (vers == "d")){
  R2max <- 0
  interp <- 0
  indmax <- NA
  set.seed(7654321)
  for (Mi in Mit){
#    set.seed(7654321)
    message("Number of substructure elements per category: ",Mi)
    for (i in 1:M){
      ind1 <- sample(unique(name[check %in% labs[1,]]))[1:Mi]
#print(ind1)
      ind2 <- sample(unique(name[check %in% labs[2,]]))[1:Mi]
#print(ind2)
      gTrain <- data[(name %in% ind1 | name %in% ind2),]
      gTest <- data[!(name %in% ind1 | name %in% ind2),]
      gcheck <- check[(name %in% ind1 | name %in% ind2)]
      gname <- name[(name %in% ind1 | name %in% ind2)]
      nt <- nrow(gTrain)
      if (vers == "d"){
#print(dim(gTrain))
        suppressMessages(outAll <- PrInDTreg(gTrain,regname,ctestv=ctestv,N,pobs=pobs,ppre=ppre,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket) )
        if (outAll$interpmax == FALSE){
          interp <- interp + 1
        }
        ctpreds <- predict(outAll$ctmax,newdata=valdat)
        R2 <- 1 - sum((valdat[,regname] - ctpreds)^2) / sum((valdat[,regname]-mean(valdat[,regname]))^2)
        MAE <- sum(abs(valdat[,regname] - ctpreds))/ dim(valdat)[1]
#print(R2)
        if (R2 > R2max & outAll$interpmax == FALSE){
          R2max <- R2
          MAEmax <- MAE
#print(c(i,R2max))
          ntmax <- nt 
          ind1max <- ind1
          ind2max <- ind2
          outmax <- outAll
          gmaxTrain <- cbind(gTrain,as.vector(gname),as.vector(gcheck))
          colnames(gmaxTrain) <- c(colnames(gTrain),"SubStruc","check")
          gmaxTest <- gTest
        }
      }
      if (vers == "a"){
        suppressMessages(outAll <- PrInDTregAll(gTrain,regname,ctestv=ctestv,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket) )
        if (outAll$interpAll == FALSE){
          interp <- interp + 1
        }
        ctpreds <- predict(outAll$treeAll,newdata=data)
        R2 <- 1 - sum((data[,regname] - ctpreds)^2) / sum((data[,regname]-mean(data[,regname]))^2)
        MAE <- sum(abs(data[,regname] - ctpreds))/ dim(data)[1]
        if (R2 > R2max & outAll$interpAll == FALSE){
          R2max <- R2
          MAEmax <- MAE
#print(c(i,R2max))
          ntmax <- nt 
          ind1max <- ind1
          ind2max <- ind2
          outmax <- outAll
          gmaxTrain <- cbind(gTrain,as.vector(gname),as.vector(gcheck))
          colnames(gmaxTrain) <- c(colnames(gTrain),"SubStruc","check")
          gmaxTest <- gTest
        }
      }
    }
message("current optimal R2 value: ",round(R2max,digits=4))
  }
}
##
D <- dim(data)[2] - 1 
if (all(is.na(Pit)) == TRUE){
  Pit <- c(D-3,D-2,D-1)
}
lPit <- length(Pit)
if ((vers == "b") | (vers == "c")){
  interp <- 0
  R2max <- 0
  set.seed(7654321)
  for (J in Pit) {
    message("number of predictors: ",J)
    if (vers == "b"){
      for (n in 1:N){
        datat <- data[,names(data)!=regname] 
        ind <- sample(1:D)[1:J] 
#print(ind)
        datat <- datat[,ind]
        datat <- cbind(datat,data[,regname])
        colnames(datat)[J+1] <- regname
        suppressMessages(outAll <- PrInDTregAll(datat,regname,ctestv,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket))
        if (outAll$interpAll == FALSE){
          interp <- interp + 1
        }
        ctpreds <- predict(outAll$treeAll,newdata=data)
        R2 <- 1 - sum((data[,regname] - ctpreds)^2) / sum((data[,regname]-mean(data[,regname]))^2)
        MAE <- sum(abs(data[,regname] - ctpreds))/ dim(data)[1]
        if (R2 > R2max & outAll$interpAll == FALSE){
          outmax <- outAll
          indmax <- ind
          R2max <- R2
          MAEmax <- MAE
#print(c(J,n,R2max))
        }
      }
      ntmax <- nrow(datat)
      ind1max <- NA
      ind2max <- NA
    }
    if (vers == "c"){
       for (i in Mit){
#         for (m in 1:N){
           datat <- data[,names(data)!=regname] 
           ind <- sample(1:D)[1:J] 
           datat <- datat[,ind]
           datat <- cbind(datat,data[,regname])
           colnames(datat)[J+1] <- regname
           ind1 <- sample(unique(name[check %in% labs[1,]]))[1:i] # select observations
           ind2 <- sample(unique(name[check %in% labs[2,]]))[1:i]
           gTrain <- datat[(name %in% ind1 | name %in% ind2),]
           gname <- name[(name %in% ind1 | name %in% ind2)]
           gcheck <- check[(name %in% ind1 | name %in% ind2)]
           nt <- nrow(gTrain)
          suppressMessages(outAll <- PrInDTreg(gTrain,regname,ctestv=ctestv,N,pobs=pobs,ppre=ppre,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket) )
          if (outAll$interpmax == FALSE){
            interp <- interp + 1
          }
          ctpreds <- predict(outAll$ctmax,newdata=valdat)
          R2 <- 1 - sum((valdat[,regname] - ctpreds)^2) / sum((valdat[,regname]-mean(valdat[,regname]))^2)
          MAE <- sum(abs(valdat[,regname] - ctpreds))/ dim(valdat)[1]
          if (R2 > R2max & outAll$interpmax == FALSE){
            outmax <- outAll
            indmax <- ind
            R2max <- R2
            MAEmax <- MAE
            ntmax <- nt 
            ind1max <- ind1
            ind2max <- ind2
            gmaxTrain <- cbind(gTrain,as.vector(gname),as.vector(gcheck))
            colnames(gmaxTrain) <- c(colnames(gTrain),"SubStruc","check")
#print(c(J,i,R2max))
          }
#        }
      }
    }
message("current optimal R2 value: ",round(R2max,digits=4))
  }
}
R2sub <- 0
MAEsub <- 0
if ((vers == "a") | (vers == "c") | (vers == "d")){ 
  pred <- predict(outmax,newdata=data)
  diff <- abs(pred - data[,regname])
  agg <- aggregate(diff,list(name=name),mean)
  MAEsub <- mean(agg[,2])
##
  diff <- (pred - data[,regname])^2
  agg <- aggregate(diff,list(name=name),sum)
  diff <- (data[,regname]-mean(data[,regname]))^2
  agg <- 1 - agg[,2] / aggregate(diff,list(name=name),sum)[,2]
  R2sub <- mean(agg)
#  meansub <- vector()
#  aggsub <- vector()
#  diffsub <- matrix(0,nrow=dim(valdat)[1],ncol=length(unique(name)))
#  for (l in 1:length(unique(name))){
#    meansub[l] <- mean(valdat[as.integer(name)==l,regname])
#    diffsub[1:length(valdat[as.integer(name)==l,regname]),l] <- (valdat[as.integer(name)==l,regname] - meansub[l])^2
#    aggsub[l] <- sum(diffsub[,l])
#  }
#print(agg[,2])
#print(aggsub)  ## much smaller: WHY??
#  agg <- agg[,2] / aggsub
#  R2sub <- 1 - mean(agg)
}
## 
## results
result <- list(outmax=outmax,interp=interp,ind1max=ind1max,ind2max=ind2max,indmax=indmax,vers=vers,R2max=R2max,MAEmax=MAEmax,
     ntmax=ntmax,lMit=lMit,lPit=lPit,M=M,N=N,gmaxTrain=gmaxTrain,labs=labs,MAEsub=MAEsub,R2sub=R2sub,indrep=indrep)
class(result) <- "PrInDTRstruc"
result
}
#' @export
print.PrInDTRstruc <- function(x,...){
  if ((x$vers == "d") | (x$vers == "a")){
    cat("Number of interpretable best trees: ",x$interp," of ",(x$lMit*x$M)," trees","\n","\n")
#    cat(x$interp," of ",(x$lMit*x$M)," trees","\n","\n")
    cat("Best interpretable tree from subsampling","\n")
    if (x$vers == "d") {print(x$outmax$ctmax)}
    if (x$vers == "a") {print(x$outmax$treeAll)}
##plot(outmax)
    cat("\n","Best R2 on validation data: ",x$R2max)
    cat("\n","Best MAE on validation data: ",x$MAEmax,"\n")
    cat("\n","Size of full training set for best tree: ",x$ntmax,"\n")
    cat("\n","Number of observations of ",rownames(x$labs)[1]," in model construction: ",sum(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind1max)] %in% x$labs[1,]),"\n")
#    cat(sum(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind1max)] %in% x$labs[1,]),"\n")
    Elements <- cbind(as.character(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind1max)]),
      as.character(x$gmaxTrain$SubStruc[x$gmaxTrain$SubStruc %in% droplevels(x$ind1max)]))
    print(table(Elements,exclude=x$labs[1,]))
    print(table(Elements,exclude=x$ind1max))
    cat("\n","Number of observations of ",rownames(x$labs)[2]," in model construction: ",sum(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind2max)] %in% x$labs[2,]),"\n")
#    cat(sum(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind2max)] %in% x$labs[2,]),"\n")
    Elements <- cbind(as.character(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind2max)]),
      as.character(x$gmaxTrain$SubStruc[x$gmaxTrain$SubStruc %in% droplevels(x$ind2max)]))
    print(table(Elements,exclude=x$labs[2,]))
    print(table(Elements,exclude=x$ind2max))
#cat("\n","Test accuracy of best tree","\n")
#ctpreds <- predict(outmax$treeAll,newdata=gmaxTest)
#print(1 - sum((data[,regname] - ctpreds)^2) / sum((data[,regname]-mean(data[,regname]))^2))
#cat("Accuracy of best tree")
#ctpreds <- predict(outmax$treeAll,newdata=data)
#print(1 - sum((data[,regname] - ctpreds)^2) / sum((data[,regname]-mean(data[,regname]))^2))
#cat("Accuracies of best tree","\n"," ",lablarge," ",labsmall," balanced","\n")
#cat(" ",c(acc1,acc2,bamax))
  }
  if ((x$vers == "b")){
    cat("Number of interpretable best trees: ",x$interp," of ",(x$N*x$lPit)," trees","\n","\n")
#    cat(x$interp," of ",(x$N*x$lPit)," trees","\n","\n") 
    cat("Best interpretable tree","\n")
    print(x$outmax$treeAll)
##    plot(outmax)
    cat("\n","Best R2 on validation data: ",x$R2max)
    cat("\n","Best MAE on validation data: ",x$MAEmax,"\n")
  }
  if ((x$vers == "c")){
    cat("Number of interpretable best trees: ",x$interp," of ",(x$N*x$lPit*x$lMit)," trees","\n","\n")
#    cat(x$interp," of ",(x$N*x$Pit*x$lMit)," trees","\n","\n")
    cat("Best interpretable tree","\n")
    print(x$outmax$ctmax)
##    plot(outmax)
    cat("\n","Best R2 on validation data: ",x$R2max)
    cat("\n","Best MAE on validation data: ",x$MAEmax,"\n")
    cat("\n","Size of training set for best tree: ",x$ntmax,"\n")
#    cat(x$ntmax,"\n")
    cat("\n","Tokens of ",rownames(x$labs)[1]," in model construction: ",sum(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind1max)] %in% x$labs[1,]),"\n")
#    cat(sum(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind1max)] %in% x$labs[1,]),"\n")
    Elements <- cbind(as.character(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind1max)]),
      as.character(x$gmaxTrain$SubStruc[x$gmaxTrain$SubStruc %in% droplevels(x$ind1max)]))
    print(table(Elements,exclude=x$labs[1,]))
    print(table(Elements,exclude=x$ind1max))
    cat("\n","Tokens of ",rownames(x$labs)[2]," in model construction: ",sum(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind2max)] %in% x$labs[2,]),"\n")
#    cat(sum(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind2max)] %in% x$labs[2,]),"\n")
    Elements <- cbind(as.character(x$gmaxTrain$check[x$gmaxTrain$SubStruc %in% droplevels(x$ind2max)]),
      as.character(x$gmaxTrain$SubStruc[x$gmaxTrain$SubStruc %in% droplevels(x$ind2max)]))
    print(table(Elements,exclude=x$labs[2,]))
    print(table(Elements,exclude=x$ind2max))
  }
  if (x$indrep == 1){
    if ((x$vers == "a") | (x$vers == "c") | (x$vers == "d")){
      cat("\n","Repeated measurements","\n") 
      cat("Measures for objects in substructure","\n")
#    cat("Mean R2 of objects in substructure: ",x$R2sub,"\n")
      cat("Mean MAE on full original data: ",x$MAEsub,"\n\n")
    }
  }
}
#' @export
plot.PrInDTRstruc <- function(x,...){
  plot(x$outmax,main="Best interpretable tree")
}

#' Interdependent estimation for classification-regression mixtures
#'
#' @description The function SimMixPrInDT applies structured subsampling for finding an optimal subsample to model
#' the relationship between the dependent variables specified in the sublist 'targets' of the list 'datalist' and all other factor and numerical variables
#' in the corresponding individual data frame specified in the sublist 'datanames' of the list 'datalist' in the same order as 'targets'. The data frames have to be different of each other! \cr
#' The function is prepared to handle classification tasks with 2 or more classes and regression tasks. At first stage, the targets are estimated based on
#' the full sample of all exogenous variables and on summaries of the observed endogenous variables.\cr
#' For generating summaries, the variables representing the substructure have to be specified in the sublist 'datastruc' of 'datalist'.\cr
#' The sublist 'summ' of 'datalist' includes for each discrete target the classes for which you want to calculate the summary percentages and for each 
#' continuous target just NA (in the same order as in the sublist 'targets'). For a discrete target in this list, you can provide a sublist of classes to be combined 
#' (for an example see the below example).\cr
#' At second stage, structured subsampling is used for improving the models from first stage.
#' The substructure of the observations used for structured subsampling is specified by the list 'Struc' which here only consists of 
#' the name 'check' of the variable with the information about the categories of the substructure (without specification of the dataset names already specified 
#' in 'datanames', see example below), and the matrix 'labs' which specifies the values of 'check' corresponding to two categories in its rows, i.e. in 'labs[1,]' and 'labs[2,]'. 
#' The names of the categories have to be specified by  \code{rownames(labs)}.\cr
#' In structured subsampling, first 'M' repetitions of subsampling of the variable 'name' with 'nsub' different elements of each category in 'check' are realized. If 'nsub' is a list, each entry is employed individually. If 'nsub' is larger than the maximum available number of elements with a certain value of 'check', the maximum possible number of elements is used. Then,
#' for each of the subsamples 'N' repetitions of subsampling in classification or regression with the specified percentages of classes, observations, and predictors are carried out.\cr 
#' These percentages are specified in the matrix 'percent', one row per estimation task. For binary classification tasks, 
#' percentages 'percl' and 'percs' for the larger and the smaller class have to be specified. For multilevel classification tasks, NA is specified (see the below example) 
#' since the percentages are generated automatically. For regression tasks, 'pobs' and 'ppre' have to be specified for observations and predictors, respectively.\cr
#' The optimization citerion is balanced accuracy for classification and goodness of fit R2 for regression on the full sample, respectively.  
#' At stage 2, the models are optimized individually. At stage 3, the models are optimized on the maximum of joint elements in their substructures.\cr
#' The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' @usage SimMixPrInDT(datalist,ctestv=NA,Struc,M=12,N=99,nsub,percent=NA,conf.level=0.95,
#'                                       minsplit=NA,minbucket=NA)
#'
#' @param datalist list(datanames,targets,datastruc,summ) Input data: For specification see the above description
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param Struc list(name,check,labs) Paprametes for structured subsampling, as explained in the desciption above.
#' @param M Number of repetitions of subsampling of elements of substructure; default = 12
#' @param N Number of repetitions of subsampling for predictors (integer); default = 99
#' @param nsub (List of) numbers of different elements of substructure per subsample
#' @param percent matrix of percent spefications: For specification see the above description; default: 'percent = NA' meaning default values for percentages.
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1); default = 0.95
#' @param minsplit Minimum number of elements in a node to be splitted; default = 20
#' @param minbucket Minimum number of elements in a node; default = 7
#'
#' @return
#' \describe{
#'   \item{modelsF}{Best trees at stage 1 (Full sample)} 
#'   \item{modelsI}{Best trees at stage 2 (Individual optimization)} 
#'   \item{modelsJ}{Best trees at stage 3 (Joint optimization)} 
#'   \item{depnames}{names of dependent variables}
#'   \item{nmod}{number of models of tasks}
#'   \item{nlev}{levels of tasks}
#'   \item{accAll}{accuracies of best trees at both stages}  
#' }
#'
#' @details
#' See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.2, for further information.
#'
#' Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name} where 'name' is the output data 
#' frame of the function.\cr
#
#' @export SimMixPrInDT
#' @importFrom stats aggregate relevel
#' @importFrom stringr str_split
#' @importFrom gdata mv
#'
#' @examples
#' # zero data
#' datazero <- PrInDT::data_zero
#' datazero <- na.omit(datazero) # cleaned full data: no NAs
#' names(datazero)[names(datazero)=="real"] <- "zero"
#' CHILDzero <- PrInDT::participant_zero
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
#' ##
#' # multi-level data
#' datastrat <- PrInDT::data_zero
#' datamult <- na.omit(datastrat)
#' # ctestv <- NA
#' datamult$mult[datamult$ETH %in% c("C1a","C1b","C1c") & datamult$real == "zero"] <- "zero1"
#' datamult$mult[datamult$ETH %in% c("C2a","C2b","C2c") & datamult$real == "zero"] <- "zero2"
#' datamult$mult[datamult$real == "realized"] <- "real"
#' datamult$mult <- as.factor(datamult$mult) # mult is new class variable
#' datamult$real <- NULL # remove old class variable
#' CHILDmult <- CHILDzero
#' ##
#' # vowel data
#' data <- PrInDT::data_vowel
#' data <- na.omit(data)
#' CHILDvowel <- data$Nickname
#' CHILDvowel <- as.factor(gsub("Nick","P",CHILDvowel))
#' data$Nickname <- NULL
#' syllable <- 3 - data$syllables
#' data$syllabels <- NULL
#' data$syllables <- syllable
#' data$speed <- data$word_duration / data$syllables
#' names(data)[names(data) == "target"] <- "vowel"
#' datavowel <- data
#' ##
#' # function preparation and call
#' # datalist and percent
#' datanames <- list("datazero","datamult","datavowel")
#' targets <- c("zero","mult","vowel")
#' datastruc <- list(CHILDzero,CHILDmult,CHILDvowel)
#' summult <- paste("zero1","zero2",sep=",")  
#' summ <- c("zero",summult,NA)
#' datalist <- list(datanames=datanames,targets=targets,datastruc=datastruc,summ=summ)
#' percent <- matrix(NA,nrow=3,ncol=2)
#' percent[1,] <- c("percl=0.075","percs=0.9") # percentages for datazero
#' # no percentages needed for datapast
#' percent[3,] <- c("pobs=0.9","ppre=c(0.9,0.8)") # percentages for datavowel
#' # substructures
#' labs <- matrix(1:6,nrow=2,ncol=3)
#' labs[1,] <- c("C1a","C1b","C1c")
#' labs[2,] <- c("C2a","C2b","C2c")
#' rownames(labs) <- c("children 1","children 2")
#' Struc <- list(check="ETH",labs=labs)
#' outSimMix <- SimMixPrInDT(datalist,ctestv=ctestv,Struc=Struc,M=2,N=9,nsub=c(19,20),
#'                        percent=percent,conf.level=0.99)
#' outSimMix
#' plot(outSimMix)
#'
SimMixPrInDT <- function(datalist,ctestv=NA,Struc,M=12,N=99,nsub,percent=NA,conf.level=0.95,minsplit=NA,minbucket=NA){
## input check
if ( typeof(datalist) != "list" || !(typeof(ctestv) %in% c("logical", "character")) 
    || !(typeof(percent) %in% c("logical","character")) || typeof(Struc) != "list" || !(typeof(M) %in% c("integer","double"))
    || !(typeof(N) %in% c("integer","double")) || !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) 
    || !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
######################
#### interdependent system
#
  anz <- length(datalist$datanames)
  if (all(is.na(percent) == TRUE)){
    percent <- matrix(NA,nrow=anz,ncol=2)
  }
  nmod <- c(1:anz)
  nlev <- list()
  datat <- datalist[[1]]
  summ <- datalist[[4]]
  inds <- list()
  extdata <- list()
#####
## elements of substructure in all datasets
#####
nl <- list()
for (K in 1:anz){
  nc <- unique(datalist$datastruc[K])
  nl <- c(nl,nc)
}
ninter <- Reduce(intersect,nl)
ninter <- as.factor(ninter)
#####
## estimate with observables only
models <- list()
modmax <- list()
  for (K in 1:anz){
    data <- as.data.frame(eval(parse(text = datat[K])))   
    dataname <- datat[K]
    target <- datalist$targets[K]
    if (length(levels(data[,target])) == 2){
      nmod[K] <- 1
      nlev <- c(nlev,"class")
      sub <- unlist(datalist$datastruc[K])
      if (is.factor(sub) != TRUE){
        cat("\n Error: 'datastruc' has to contain factor variables only\n")
        return()
      }
      sub <- droplevels(sub)
      s <- unlist(str_split(summ[K],","))
      si <- which(levels(data[,target])==s)
      snotreal <- table(sub,data[,target])[,si]
      sc <- as.character(sub)
      z1 <- aggregate(data[,target] ~ sc,data=cbind(sc,data[,target]),FUN="length")
      ind <- snotreal / z1[,2]
      inds[[K]] <- ind
#      print(length(inds[[K]]))
      extdata[[K]] <- paste0(dataname,"i")
    } 
    if (length(levels(data[,target])) > 2){
      nmod[K] <- length(levels(data[,target]))
      nlev <- c(nlev,levels(data[,target]))
      sub <- unlist(datalist$datastruc[K])
      if (is.factor(sub) != TRUE){
        cat("\n Error: 'datastruc' has to contain factor variables only\n")
        return()
      }
      sub <- droplevels(sub)
      s <- unlist(str_split(summ[K],","))
      si <- which(levels(data[,target]) %in% s)
      snotmarked <- table(sub,data[,target])[,si[1]]
      if (length(s) > 1){
        for (i in 2:length(s)){
          snotmarked <- snotmarked + table(sub,data[,target])[,si[i]]
        }
      }
      sc <- as.character(sub)
      p1 <- aggregate(data[,target] ~ sc,data=cbind(sc,data[,target]),FUN="length")
      ind <- snotmarked / p1[,2]
      inds[[K]] <- ind
#      print(length(inds[[K]]))
      extdata[[K]] <- paste0(dataname,"i")
     }
     if (length(levels(data[,target])) == 0){
      nmod[K] <- 1
      nlev <- c(nlev,"real")
      sub <- unlist(datalist$datastruc[K])
      if (is.factor(sub) != TRUE){
        cat("\n Error: 'datastruc' has to contain factor variables only\n")
        return()
      }
      sc <- as.character(sub)
      v1 <- aggregate(data[,target] ~ sc,data=cbind(sc,data[,target]),FUN="mean")
      ind <- v1[,2]
      names(ind) <- v1[,1]
      inds[[K]] <- ind
#      print(length(inds[[K]]))
      extdata[[K]] <- paste0(dataname,"i")
    }
  }
  for (K in 1:anz){
    datai <- as.data.frame(eval(parse(text = datat[K])))    
    mv("datai",extdata[[K]])
  }
## name definitions
for (K in 1:anz){
  datai <- eval(parse(text = extdata[K]))
  for (i in 1:anz){
    if (i != K){
      ind <- inds[[i]]
      D <- dim(datai)[2]
      datai$indi <- mean(ind) ## NA
      for (j in 1:length(ind)) {
        datai$indi[unlist(datalist$datastruc[K]) == names(ind)[j]] <- ind[j]
      }
    names(datai)[(D+1)] <- paste0("p",datalist$targets[i])
    }
  }
  mv("datai",extdata[[K]])
#print(str(eval(parse(text = extdata[K]))))
}
  modelsP <- list()
  acc <- c(1:anz)*0
## first stage
  for (K in 1:anz){
    data <- eval(parse(text = extdata[K]))
    l <- colnames(data)
    if (length(unique(l)) != length(l)){
      cat("\n Error: Duplicated variable name generated: Make sure that the data frames in 'datanames' are different!\n")
      return()
    }
    dataname <- datat[K]
    target <- datalist$targets[K]
#print(levels(as.factor(data[,target])))
message(dataname,": 1st stage")
# 2 classes: call of PrInDT
    if (length(levels(data[,target])) == 2){
      out2S <- PrInDTAll(data,target,ctestv,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)
      acc[K] <- out2S$baAll
      modelsP <- c(modelsP,out2S$treeAll)
    } 
# more than 2 classes: call of PrInDTMulev
    if (length(levels(data[,target])) > 2){
      out2S <- PrInDTMulevAll(data,target,ctestv,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket) # percentages chosen automatically
      acc[K] <- out2S$baAll
      modelsP <- c(modelsP,out2S$treeAll)
     }
# regression: call of PrInDTreg
     if (length(levels(data[,target])) == 0){
#       x <- str_split(percent[K,1],"=")
#       assign(x[[1]][1],eval(parse(text=percent[K,1])))
#       x <- str_split(percent[K,2],"=")
#      assign(x[[1]][1],eval(parse(text=percent[K,2])))
      out2S <- PrInDTregAll(data,target,ctestv,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)
      acc[K] <- out2S$R2All
      modelsP <- c(modelsP,out2S$treeAll)
    }
 }
accP <- acc
#print(accP)
accF <- acc
modelsF <- modelsP
#print(table(datazeroi$ETH[(!(datazeroi$MLU == 1)) & datazeroi$SEX == "male" & 
#  datazeroi$pvowel <= 168.441 & datazeroi$PRN %in% c("he","I","she","they")]))
#print(table(datazeroi$ETH[(!(datazeroi$MLU == 1)) & datazeroi$SEX == "male" & 
#  datazeroi$pvowel <= 168.441 & datazeroi$PRN %in% c("he","I","she","they") & datazeroi$zero == "zero"]))
#print(table(datazeroi$ETH[(!(datazeroi$MLU == 1)) & datazeroi$SEX == "male" & 
#  datazeroi$pvowel <= 168.441 & datazeroi$PRN %in% c("he","I","she","they") & datazeroi$zero == "real"]))
#print(table(datazeroi$ETH[(!(datazeroi$MLU == 1)) & datazeroi$SEX == "male" & 
#  datazeroi$PRN %in% c("he","I","she","they")]))
#print(table(datazeroi$ETH[(!(datazeroi$MLU == 1)) & datazeroi$SEX == "male" & 
#  datazeroi$PRN %in% c("he","I","she","they") & datazeroi$zero == "zero"]))
####################
## 2nd stage
ssize <- nsub
modmax <- modelsP
#set.seed(7654321)
 acc <- c(1:anz)*0
 NM <- N  ## NEW version
# NM <- N / M * 2 # 333
 labs <- Struc$labs
 check <- Struc$check
message("\n","Second stage")
# call of PrInDT functions
for (i in ssize){
  message("Number of elements: ",i)
  for (n in 1:M){
    message("replication ",n)
    for (K in 1:anz){
      dataname <- datat[K]
message(dataname,": 2nd stage")
      set.seed(54321*n + K*n)                                        
      datai <- eval(parse(text = extdata[K]))
      if (!(check %in% colnames(datai))){
        cat("\n Error: 'check' is not in data set\n")
        return()
      }
      z <- datai[,check]
      if (is.factor(z) != TRUE){
        cat("\n Error: 'check' has to be factor variable\n")
        return()
      }
      sub <- unlist(datalist$datastruc[K])
      sub <- droplevels(sub)
      target <- datalist$targets[K]
## 2 classes
      if (length(levels(datai[,target])) == 2){
        ind1 <- sample(unique(sub[datai[,check ]%in% labs[1,]]))[1:i] 
        ind2 <- sample(unique(sub[datai[,check] %in% labs[2,]]))[1:i] 
        datas <- datai[(as.factor(sub) %in% ind1 | as.factor(sub) %in% ind2) ,]
# assigning percl, percs
        percl <- NA
        percs <- 1
        x <- str_split(percent[K,1],"=")
        assign(x[[1]][1],eval(parse(text=percent[K,1])))
        x <- str_split(percent[K,2],"=")
        assign(x[[1]][1],eval(parse(text=percent[K,2])))
        if (!(all(0 < percl & percl <= 1) || is.na(percl)==TRUE) || !(all(0 < percs & percs <= 1))){
          stop("irregular input for percl or percs")
        } 
# outzero <- suppressMessages(PrInDT(datazeroi,"real",ctestv, N,percl=0.075,percs=0.9,conf.level,seedl=FALSE)) # best percentages found
        outSim <- suppressMessages(PrInDT(datas,target,ctestv, NM,percl=percl,percs=percs,conf.level,seedl=FALSE,minsplit=minsplit,minbucket=minbucket))
        class.pred <- predict(outSim$tree1st,newdata=datai)
        class.pred <- relevel(class.pred,ref=levels(datai[,target])[1])
        conti <- table(class.pred,datai[,target])
        if (dim(conti)[1] == 1){                                                ## NEWNEW
          conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
        }
        ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
        ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
        acc[K] <- (ba1 + ba2)/2
        models[[K]] <- outSim$tree1st
        if (acc[K] > accP[K]){ 
         accP[K] <- acc[K]
#         print(c(n,accP))
         modmax[[K]] <- models[[K]]
#         accPmean <- (accP[1] + accP[2] + sqrt(accP[3]))/3  ## NOT GENERAL
       }}
## more than 2 classes
       if (length(levels(datai[,target])) > 2){
#sample(unique(name[check %in% labs[1,]]),prob=p1)[1:M] 
         ind1 <- sample(unique(sub[datai[,check ]%in% labs[1,]]))[1:i] 
         ind2 <- sample(unique(sub[datai[,check] %in% labs[2,]]))[1:i] 
         datas <- datai[(as.factor(sub) %in% ind1 | as.factor(sub) %in% ind2) ,]
         outSim <- suppressMessages(PrInDTMulev(datas,target,ctestv,NM,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)) # percentages chosen automatically
         models[[K]] <- outSim$trees
         predclass <- predict(outSim,target,newdata=datai)
         conf <- table(predclass,datai[,target]) # confusion matrix
         if (dim(conf)[1] == 1){                                                ## NEWNEW
           conf <- matrix(c(conf,0,0),nrow=2,byrow=TRUE)
         }
#           accx <- 0
#           for (i in 1:length(levels(datas[,target]))){
#             accx <- accx + conf[i,i] / colSums(conf)[i]
#           }
#           acc[K] <- accx / length(levels(datas[,target])
           accx <- rep(0,length(levels(datas[,target])))
           for (j in 1:length(rownames(conf))){
            for (l in 1:length(colnames(conf))){
              if (rownames(conf)[j] == colnames(conf)[l]){
                accx[l] <- conf[j,l] / colSums(conf)[l]
              }
           }
         }
          acc[K] <- sum(accx) / length(levels(datas[,target]))
          if (acc[K] > accP[K]) { 
            accP[K] <- acc[K]
#            print(accP)
            modmax[[K]] <- models[[K]]
#            accPmean <- (accP[1] + accP[2] + sqrt(accP[3]))/3  ## not general
          }
       }
## regression
       if (length(levels(datai[,target])) == 0){
#sample(unique(name[check %in% labs[1,]]),prob=p1)[1:M] 
         ind1 <- sample(unique(sub[datai[,check ]%in% labs[1,]]))[1:i] 
         ind2 <- sample(unique(sub[datai[,check] %in% labs[2,]]))[1:i] 
         datas <- datai[(as.factor(sub) %in% ind1 | as.factor(sub) %in% ind2) ,]
# assigning pobs, ppre
         pobs <- c(0.9,0.7)
         ppre <- c(0.9,0.7)
         x <- str_split(percent[K,1],"=")
         assign(x[[1]][1],eval(parse(text=percent[K,1])))
         x <- str_split(percent[K,2],"=")
         assign(x[[1]][1],eval(parse(text=percent[K,2]))) 
         if (!(all(0 < pobs & pobs <= 1)) || !(all(0 < ppre & ppre <= 1))){
          stop("irregular input for pobs or ppre")
        } 
         outSim <- suppressMessages(PrInDTreg(datas,target,ctestv,NM,pobs,ppre,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket))
         models[[K]] <- outSim$ctmax
         ctpreds <- predict(outSim$ctmax,newdata=datai)
         acc[K] <- 1 - sum((datai[,target] - ctpreds)^2) / sum((datai[,target]-mean(datai[,target]))^2) ## R2 !!!!!
         if (acc[K] > accP[K]) { 
           accP[K] <- acc[K]
#           print(accP)
           modmax[[K]] <- models[[K]]
#           accPmean <- (accP[1] + accP[2] + sqrt(accP[3]))/3       ## not general
         }
       }          
}}}
#print(accP)
accI <- accP
accF <- cbind(accF,accI)
modelsI <- modmax
#plot(unlist(modmax[[1]]))
#plot(unlist(modmax[[2]][[1]]))
#plot(unlist(modmax[[2]][[2]]))
#plot(unlist(modmax[[2]][[3]]))
#plot(unlist(modmax[[3]]))
##############################################
# call on subsamples: joined version
##
#ssize <- length(ninter)/2
set.seed(7654321)
acc <- c(1:anz)*0
#accPmean <- (accP[1] + accP[2] + sqrt(accP[3]))/3  ## not general
n <- 1
message("\n","Third stage: Joint version")
message(length(ninter)," elements in substructure")
#for (i in ssize){
#  print(i)
#  for (n in 1:M){
#    print(n)
    for (K in 1:anz){  
      dataname <- datat[K]
      set.seed(54321*n + K*n)                                        
      datai <- eval(parse(text = extdata[K]))
      sub <- unlist(datalist$datastruc[K])
      sub <- droplevels(sub)
      if (K == 1){
        ind1 <- sample(unique(sub[sub %in% ninter & datai[,check ] %in% labs[1,]]))  ## [1:i] 
        ind2 <- sample(unique(sub[sub %in% ninter & datai[,check ] %in% labs[2,]]))  ## [1:i] 
        ind1 <- droplevels(ind1)
        ind2 <- droplevels(ind2)
        lind1 <- length(ind1)
        lind2 <- length(ind2)
message("number of elements in categories: ",lind1," , ",lind2)
     }
      target <- datalist$targets[K]
message(dataname,": joint version")
## 2 classes
      if (length(levels(datai[,target])) == 2){
#        ind1 <- sample(unique(sub[sub %in% ninter & datai[,check ] %in% labs[1,]]))  ## [1:i] 
#        ind2 <- sample(unique(sub[sub %in% ninter & datai[,check] %in% labs[2,]]))   ## [1:i] 
        datas <- datai[(as.factor(sub) %in% ind1) | (as.factor(sub) %in% ind2) ,]
#if (K == 1){
#print(table(datas$ETH[(!(datas$MLU == 1)) & datas$SEX == "male" & 
#  datas$pvowel <= 168.441 & datas$PRN %in% c("he","I","she","they") & datas$zero == "zero"]))
#}
# assigning percl, percs
        percl <- NA
        percs <- 1
        x <- str_split(percent[K,1],"=")
        assign(x[[1]][1],eval(parse(text=percent[K,1])))
        x <- str_split(percent[K,2],"=")
        assign(x[[1]][1],eval(parse(text=percent[K,2])))
        outSim <- suppressMessages(PrInDT(datas,target,ctestv, N,percl=percl,percs=percs,conf.level,seedl=FALSE,minsplit=minsplit,minbucket=minbucket)) 
        class.pred <- predict(outSim$tree1st,newdata=datai)
        class.pred <- relevel(class.pred,ref=levels(datai[,target])[1])
        conti <- table(class.pred,datai[,target])
        if (dim(conti)[1] == 1){                                                ## NEWNEW
          conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
        }
        ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
        ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
        acc[K] <- (ba1 + ba2)/2
        models[[K]] <- outSim$tree1st
#        if (acc[K] > accP[K]){ 
#         accP[K] <- acc[K]
#         print(c(n,accP))
#         modmax[[K]] <- models[[K]]
#         accPmean <- (accP[1] + accP[2] + sqrt(accP[3]))/3  ## NOT GENERAL
#       }
      }
## more than 2 classes
       if (length(levels(datai[,target])) > 2){
#         ind1 <- sample(unique(sub[sub %in% ninter & datai[,check ]%in% labs[1,]]))  ## [1:i] 
#         ind2 <- sample(unique(sub[sub %in% ninter & datai[,check] %in% labs[2,]]))  ## [1:i] 
         datas <- datai[(as.factor(sub) %in% ind1) | (as.factor(sub) %in% ind2) ,]
         outSim <- suppressMessages(PrInDTMulev(datas,target,ctestv,N,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)) # percentages chosen automatically
         models[[K]] <- outSim$trees
         predclass <- predict(outSim,target,newdata=datai)
         conf <- table(predclass,datai[,target]) # confusion matrix
         if (dim(conf)[1] == 1){                                                ## NEWNEW
           conf <- matrix(c(conf,0,0),nrow=2,byrow=TRUE)
         }
#print(conf)
#           accx <- 0
#           for (i in 1:length(levels(datas[,target]))){
#             accx <- accx + conf[i,i] / colSums(conf)[i]
#           }
#           acc[K] <- accx / length(levels(datas[,target])
           accx <- rep(0,length(levels(datas[,target])))
           for (j in 1:length(rownames(conf))){
            for (l in 1:length(colnames(conf))){
              if (rownames(conf)[j] == colnames(conf)[l]){
                accx[l] <- conf[j,l] / colSums(conf)[l]
              }
           }
         }
          acc[K] <- sum(accx) / length(levels(datas[,target]))
#          if (acc[K] > accP[K]) { 
#            accP[K] <- acc[K]
#            print(accP)
#            modmax[[K]] <- models[[K]]
#            accPmean <- (accP[1] + accP[2] + sqrt(accP[3]))/3  ## not general
#          }
       }
## regression
       if (length(levels(datai[,target])) == 0){
#         ind1 <- sample(unique(sub[sub %in% ninter & datai[,check ] %in% labs[1,]]))  ## [1:i] 
#         ind2 <- sample(unique(sub[sub %in% ninter & datai[,check] %in% labs[2,]]))  ## [1:i] 
         datas <- datai[(as.factor(sub) %in% ind1) | (as.factor(sub) %in% ind2) ,]
# assigning pobs, ppre
         pobs <- c(0.9,0.7)
         ppre <- c(0.9,0.7)
         x <- str_split(percent[K,1],"=")
         assign(x[[1]][1],eval(parse(text=percent[K,1])))
         x <- str_split(percent[K,2],"=")
         assign(x[[1]][1],eval(parse(text=percent[K,2]))) 
         NM <- N*M/2  ## NEW version
         outSim <- suppressMessages(PrInDTreg(datas,target,ctestv,NM,pobs,ppre,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket))
         models[[K]] <- outSim$ctmax
         ctpreds <- predict(outSim$ctmax,newdata=datai)
         acc[K] <- 1 - sum((datai[,target] - ctpreds)^2) / sum((datai[,target]-mean(datai[,target]))^2) ## R2 !!!!!
#         if (acc[K] > accP[K]) { 
#           accP[K] <- acc[K]
#           print(accP)
#           modmax[[K]] <- models[[K]]
#           accPmean <- (accP[1] + accP[2] + sqrt(accP[3]))/3       ## not general
#         }
       }          
}#}
#accmean <- (acc[1] + acc[2] + sqrt(acc[3]))/3  ## not general
#  if (accmean > accPmean) { 
#    accP[1:anz] <- acc[1:anz]
#    print(accP)
#    for (i in 1:anz){
#      modmax[[i]] <- models[[i]]
#    }
#   accPmean <- accmean
#  }
#print(acc)
#print(accP)
#print(accPmean)
accJ <- acc
accF <- cbind(accF,acc)
modelsJ <- models
#plot(modelsJ[[1]])
#plot(modelsJ[[2]][[1]])
#plot(modelsJ[[2]][[2]])
#plot(modelsJ[[2]][[3]])
#plot(modelsJ[[3]])
result <- list(modelsF=modelsF,modelsI=modelsI,modelsJ=modelsJ,depnames=unlist(datalist$target),nmod=nmod,nlev=nlev,accAll=accF)
class(result) <- "SimMixPrInDT"
result
}
#' @export
print.SimMixPrInDT <- function(x,...){
## output
cat("Accuracies of best models","\n")
rownames(x$accAll) <- c(x$depnames)
colnames(x$accAll) <- c("1st stage","2nd stage","3rd stage")
print(x$accAll)
cat("\n\n")
J <- 0
for (i in 1:length(x$depnames)){
  if (x$nmod[i] == 1){
    J <- J + 1
    cat("Best model 2nd stage: ",x$depnames[i], "\n") 
    print(x$modelsI[[i]])
    cat("\n\n")
  } else {
    K <- x$nmod[i]
    for (k in 1:K){
       J <- J + 1
       cat("Best model 2nd stage: ",x$depnames[i],":",x$nlev[[J]],"\n") 
       print(x$modelsI[[i]][[k]])
       cat("\n\n")
    }
  }
}}
#' @export
plot.SimMixPrInDT <- function(x,...){
J <- 0
for (i in 1:length(x$depnames)){
  if (x$nmod[i] == 1){
    J <- J + 1
    plot(x$modelsI[[i]],main=paste0("2nd stage: ",x$depnames[i]))
  } else {
    K <- x$nmod[i]
    for (k in 1:K){
       J <- J + 1
       plot(x$modelsI[[i]][[k]],main=paste0("2nd stage: ",x$depnames[i],":",x$nlev[[J]]))
    }
  }
}}

#' Interdependent estimation for regression
#'
#' @description The function \code{\link{SimRPrInDT}} applies structured subsampling for finding an optimal subsample to model
#' the relationship between the continuous variables with indices 'inddep' and all other factor and numerical variables
#' in the data frame 'datain'. \cr
#' The substructure of the observations used for subsampling is specified by the list 'Struc' which consists of the factor variable 'name' representing the substructure,
#' the name 'check' of the factor variable with the information about the categories of the substructure, and the matrix 'labs' which specifies the values of 'check'
#' corresponding to two categories in its rows, i.e. in 'labs[1,]' and 'labs[2,]'. The names of the categories have to be specified by \code{rownames(labs)}.\cr
#' In structured subsampling, first 'M' repetitions of subsampling of the variable 'name' with 'nsub' different elements of each category in 'check' are realized. If 'nsub' is a list, each entry is employed individually. If 'nsub' is larger than the maximum available number of elements with a certain value of 'check', the maximum possible number of elements is used. Then,
#' for each of the subsamples 'N' repetitions of subsampling of 'ppre' percentages of the predictors are carried out.\cr 
#' Subsampling of observations can additionally be restricted to 'pobs' percentages.\cr
#' The optimization citerion is the goodness of fit R2 on the full sample. At stage 2, the models are optimized individually. 
#' At stage 3, the mean of accuracies is optimized over all models.\cr
#' Struc=NA causes random subsampling of observations instead of structured subsampling.\cr
#' The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' @usage SimRPrInDT(data,ctestv=NA,Struc=NA,inddep,N=99,pobs=0.9,ppre=c(0.9,0.7),
#'                 M=1,nsub=1,conf.level=0.95,minsplit=NA,minbucket=NA)
#'
#' @param data Input data frame with continuous target variables with column indices 'inddep' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param Struc = list(name,check,labs), cf. description for explanations
#' @param inddep Column indices of target variables in datain
#' @param N Number of repetitions of subsampling (integer) of predictors; default = 99
#' @param pobs Percentage(s) of observations for subsampling; default=0.9
#' @param ppre Percentage(s) of predictors for subsampling; default=c(0.9,0.7)
#' @param M Number of repetitions of subsampling of elements of substructure
#' @param nsub (List of) numbers of different elements of substructure per subsample
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return
#' \describe{
#'   \item{modelsF}{Best trees at stage 1} 
#'   \item{modelsI}{Best trees for the different values of 'nsub' at stage 2}
#'   \item{modelsJ}{Best trees for the different values of 'nsub' after mean optimization}
#'   \item{Struc}{Used structure}
#'   \item{msub}{Best numbers of elements in substructure: 2nd stage, 3rd stage}
#'   \item{depnames}{names of dependent variables}
#'   \item{R2All}{R2s of best trees at stages 1, 2, mean max.}  
#' }
#'
#' @details
#' See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.2, for further information.
#'
#' Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name} where 'name' is the output data 
#' frame of the function.\cr
#
#' @export SimRPrInDT
#'
#' @examples
#' data <- PrInDT::data_vowel
#' data <- na.omit(data)
#' CHILDvowel <- data$Nickname
#' data$Nickname <- NULL
#' syllable <- 3 - data$syllables
#' data$syllabels <- NULL
#' data$syllables <- syllable
#' data$speed <- data$word_duration / data$syllables
#' names(data)[names(data) == "target"] <- "vowel_length"
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
#' # structure definition
#' name <- CHILDvowel
#' check <- "data$ETH"
#' labs <- matrix(1:6,nrow=2,ncol=3)
#' labs[1,] <- c("C1a","C1b","C1c")
#' labs[2,] <- c("C2a","C2b","C2c")
#' rownames(labs) <- c("children 1","children 2")
#' Struc <- list(name=name,check=check,labs=labs)
#' # column indices of dependent variables
#' inddep <- c(13,9) 
#' outSimR <- SimRPrInDT(data,ctestv=ctestv,Struc=Struc,inddep=inddep,N=3,M=2,
#'                    nsub=c(19,20),conf.level=0.99)
#' outSimR
#' plot(outSimR)
#'
SimRPrInDT <- function(data,ctestv=NA,Struc=NA,inddep,N=99,pobs=0.9,ppre=c(0.9,0.7),M=1,nsub=1,conf.level=0.95,minsplit=NA,minbucket=NA){
## input check
if ( typeof(data) != "list" || !(typeof(ctestv) %in% c("logical", "character"))
    || !(typeof(Struc) %in% c("logical","list")) || !all(0 < pobs & pobs <= 1) || !all(0 < ppre & ppre <= 1) 
    || !(typeof(N) %in% c("integer","double")) || !(typeof(M) %in% c("integer","double")) 
    || !(typeof(inddep) %in% c("integer","double")) || !(typeof(nsub) %in% c("integer","double"))
    || !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
depnames <- colnames(data)[inddep]
anz <- length(inddep)
D <- dim(data)[2]
if (any(is.na(Struc) != TRUE)){
  name <- Struc$name
  if (is.factor(name) != TRUE){
    cat("\n Error: 'Struc$name' has to be a factor variable\n")
    return()
  }
#check <- Struc$check
  check <- eval(parse(text = Struc$check))
  if (length(check) == 0){
    cat("\n irregular input: check variable not in data set\n")
    return()
  }
  if (is.factor(check) != TRUE){
    cat("\n Error: 'Struc$check' has to be a factor variable\n")
    return()
  }
  labs <- Struc$labs
} else {
  M <- 1
  nsub <- 1
}
#message ("Simultaneous modeling")
models <- list()
modmax <- as.list(1:(anz*(1+length(nsub)))) 
dim(modmax) <- c((1+length(nsub)),anz) 
ctpreds <- matrix(0,ncol=anz,nrow=dim(data)[1])
R2 <- c(1:anz)*0
max01 <- matrix(0,nrow=(length(nsub)+1),ncol=(anz+1)) # nrow = full, all sizes
##
message("1st stage: full sample")
for (i in 1:anz){
#  message(depnames[i])
  #outreg <- PrInDTreg(datavowel,"vowel_length",ctestv,N,pobs,ppre,conf.level=cf) ## all vowel related variables
  outreg <- PrInDTregAll(data,depnames[i],ctestv,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)
  models[[i]] <- outreg$treeAll
  ctpredsVL <- predict(outreg$treeAll,newdata=data)
  ctpreds[,i] <- ctpredsVL
  R2[i] <- outreg$R2All
  modmax[[1,i]] <- models[[i]]
}
max01[1,] <- c(R2,mean(R2))
colnames(max01) <- c(depnames,"mean")
#print(max01[1,])
R2P <- max01[1,]
modelsP <- models
##
message("\n","2nd stage: individual optimization")
set.seed(7654321)
k <- 1
for (i in nsub){
  if (any(is.na(Struc) != TRUE)){
    message("Number of elements: ",i)
  }
  k <- k + 1
  max01[k,] <- max01[1,]
  for (n in 1:M){
    if (any(is.na(Struc) != TRUE)){
      message("repetition ",n)
      ind1 <- sample(unique(name[check %in% labs[1,]]))[1:i] 
      ind2 <- sample(unique(name[check %in% labs[2,]]))[1:i] 
      datas <- data[(as.factor(name) %in% ind1) | (as.factor(name) %in% ind2) ,]
##  datas <- cbind(dataMend[,sample(colnames(dataMend))[1:psize]],datavowel[,c(8,9,10,13)])
    } else {
      datas <- data
    }
    for (j in 1:anz){
      outland <- suppressMessages(PrInDTreg(datas,depnames[j],ctestv=ctestv,N,pobs,ppre,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket))
      ctpredsVL <- predict(outland$ctmax,newdata=data)
      R2[j] <- 1 - sum((data[,inddep[j]] - ctpredsVL)^2) / sum((data[,inddep[j]]-mean(data[,inddep[j]]))^2) ## R2 !!!!!
      models[[j]] <- outland$ctmax
      if (R2[j] > max01[k,j]) { 
        max01[k,j] <- R2[j]
        max01[k,(anz+1)] <- mean(max01[k,1:anz])
#        print(max01[k,])
        modmax[[k,j]] <- models[[j]]
      }
   }  
 } 
}
colnames(max01) <- c(depnames,"mean")
#R2I <- max01
#modelsI <- modmax
k2 <- which.max(max01[2:k,(anz+1)])
R2I <- max01[(k2+1),]
modelsI <- modelsP
for (i in 1:anz){
  modelsI[[i]] <- modmax[[(k2+1),i]]
}
#print(k)
#for (i in 1:anz){
#  plot(unlist(modmax[[k,i]]))
#}
###  optimization of mean
message("\n","3rd stage: joint optimization")
k  <- 1
for (i in nsub){
  if (any(is.na(Struc) != TRUE)){
    message("Number of elements: ",i)
  }
  k <- k + 1
  max01[k,] <- max01[1,]
  R21m <- sum(max01[1,1:anz]) / anz
  for (n in 1:M) {
    set.seed(4321 * n + n)
    if (any(is.na(Struc) != TRUE)){
      message("repetition ",n)
      ind1 <- sample(unique(name[check %in% labs[1,]]))[1:i] 
      ind2 <- sample(unique(name[check %in% labs[2,]]))[1:i] 
      datas <- data[(as.factor(name) %in% ind1) | (as.factor(name) %in% ind2) ,]
##  datas <- cbind(dataMend[,sample(colnames(dataMend))[1:psize]],datavowel[,c(8,9,10,13)])
    } else {
      datas <- data
    }
    for (j in 1:anz){
      outland <- suppressMessages(PrInDTreg(datas,depnames[j],ctestv=ctestv,N,pobs,ppre,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket))
      ctpredsVL <- predict(outland$ctmax,newdata=data)
      R2[j] <- 1 - sum((data[,inddep[j]] - ctpredsVL)^2) / sum((data[,inddep[j]]-mean(data[,inddep[j]]))^2) ## R2 !!!!!
      models[[j]] <- outland$ctmax
    } 
    R2m <- sum(R2[1:anz]) / anz
    if (R2m > R21m) { 
      max01[k,1:anz] <- R2
      max01[k,(anz+1)] <- mean(max01[k,1:anz])
#      print(max01[k,])
      for (j in 1:anz){
        modmax[[k,j]] <- models[[j]]
      }
      R21m <- R2m
    }
  }
}
colnames(max01) <- c(depnames,"mean")
k3 <- which.max(max01[2:k,(anz+1)])
R2J <- max01[(k3+1),]
modelsJ <- modelsP
for (i in 1:anz){
  modelsJ[[i]] <- modmax[[(k3+1),i]]
}
#for (i in 1:anz){
#  plot(unlist(modmax[[k,i]]))
#}
R2All <- rbind(R2P,R2I,R2J)
result <- list(modelsF=modelsP,modelsI=modelsI,modelsJ=modelsJ,Struc=Struc,msub=c(nsub[k2],nsub[k3]),depnames=depnames,R2All=R2All)
class(result) <- "SimRPrInDT"
result
}
#' @export
print.SimRPrInDT <- function(x,...){
## output
#anz <- length(x$depnames)
if (any(is.na(x$Struc) != TRUE)){
  cat("Best numbers of elements in substructure","\n")
  cat("2nd stage: ",x$msub[1],"\n")
  cat("3rd stage: ",x$msub[2],"\n","\n")
}
cat("Accuracies of best models","\n")
colnames(x$R2All) <- c(x$depnames,"mean")
rownames(x$R2All) <- c("full sample","indiv. opt.","joint opt.")
print(x$R2All)
cat("\n\n")
#k <- which.max(modelsI[,(anz+1)])
#print(k)
for (i in 1:length(x$depnames)){
  cat("Best model 2nd stage: ",x$depnames[i], "\n")
  print(x$modelsI[[i]])
  cat("\n")
}
cat("\n")
#k <- which.max(modelsJ[,(anz+1)])
#print(k)
for (i in 1:length(x$depnames)){
  cat("Best model after joint opt.: ",x$depnames[i], "\n")
  print(x$modelsJ[[i]])
  cat("\n")
}
}
#' @export
plot.SimRPrInDT <- function(x,...){
#anz <- length(x$depnames)
#k <- which.max(R2All[,(anz+1)])
#print(k)
for (i in 1:length(x$depnames)){
  plot(x$modelsI[[i]],main=paste0("2nd stage: ",x$depnames[i]))
}
for (i in 1:length(x$depnames)){
  plot(x$modelsJ[[i]],main=paste0("mean max.: ",x$depnames[i]))
}
}

#' Optimisation of undersampling percentages for classification
#'
#' @description The function \code{\link{OptPrInDT}} applies an iterative technique for finding optimal undersampling percentages 
#'  'percl' for the larger class and 'percs' for the smaller class by a nested grid search for the use of the function \code{\link{PrInDT}} for
#' the relationship between the two-class factor variable 'classname' and all other factor and numerical variables
#' in the data frame 'data' by means of 'N' repetitions of undersampling. The optimization citerion is the balanced accuracy 
#' on the validation sample 'valdat' (default = full sample 'data'). The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'.\cr
#' The inputs plmax and psmax determine the maximal values of the percentages and the inputs distl and dists the
#' the distances to the next smaller percentage to be tried.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#' The parameter 'steps' controls, how many of the 3 possible optimization steps should be carried out; default=3.\cr
#'
#' @usage OptPrInDT(data,classname,ctestv=NA,N=99,plmax=0.09,psmax=0.9,
#'                distl=0.01,dists=0.1,conf.level=0.95,minsplit=NA,minbucket=NA,
#'                valdat=data,steps=3)
#'
#' @param data Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param N Number (> 7) of repetitions (integer)
#' @param plmax Maximal undersampling percentage of larger class (numerical, > 0 and <= 1);\cr
#'     default = 0.09
#' @param psmax Maximal undersampling percentage of smaller class (numerical, > 0 and <= 1);\cr
#'     default = 0.9
#' @param distl Distance to the next lower undersampling percentage of larger class (numerical, > 0 and < 1);\cr
#'     default = 0.01
#' @param dists Distance to the next lower undersampling percentage of smaller class (numerical, > 0 and < 1);\cr
#'     default = 0.1
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#' @param valdat validation data; default = data
#' @param steps number of optimization steps = 1, 2, 3; default = 3
#'
#' @return
#' \describe{
#' \item{besttree}{best tree on full sample}
#' \item{bestba}{balanced accuracy of best tree on full sample}
#' \item{percl}{undersampling percentage of large class of best tree on full sample}
#' \item{percs}{undersampling percentage of small class of best tree on full sample}
#' }
#'
#' @details
#' See help("RePrInDT") and help("PrInDT") for further information.
#'
#' Standard output can be produced by means of \code{print(name$besttree)} or just \code{name$besttree} as well as \code{plot(name$besttree)} where 'name' is the output data 
#' frame of the function.\cr
#
#' @export OptPrInDT
#'
#' @examples
#' datastrat <- PrInDT::data_zero
#' data <- na.omit(datastrat) # cleaned full data: no NAs
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
#' # call of OptPrInDT
#' out <- OptPrInDT(data,"real",ctestv,N=24,conf.level=0.99,steps=1) # unstratified
#' out # print best model and ensembles as well as all observations
#' plot(out)
#'
OptPrInDT <- function(data,classname,ctestv=NA,N=99,plmax=0.09,psmax=0.9,distl=0.01,dists=0.1,conf.level=0.95,minsplit=NA,minbucket=NA,valdat=data,steps=3){
  ## input check
  if (typeof(data) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) || N <= 0 ||
      !(0 < plmax & plmax <= 1) || !(0 < psmax & psmax <= 1) || !(0 < distl & distl < 1) || !(0 < dists & dists < 1) ||
      !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double")) || typeof(valdat) != "list" || !(1 <= steps & steps <= 3) ) {
    stop("irregular input")
  }
  if (!(all(names(data) %in% names(valdat)))){
    stop("validation data variables unequal input data variables")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
  if (N < 24){
      stop("Number of repetitions too low (must be > 23)")
    }
  names(data)[names(data)==classname] <- "class"
  names(valdat)[names(valdat)==classname] <- "class"
  if (!(identical(levels(data$class),levels(valdat$class)))){
    stop("levels of input class variable unequal levels of validation class variable")
  }
## parameter setting
incl <- c(distl,distl/2,distl/4) ## grid radius in 1st, 2nd, 3rd grid for large class
incs <- c(dists,dists/2,dists/4)    ## grid radius in 1st, 2nd, 3rd grid for small class
n <- c(N/8,N/4,N/2,N)  ## repetitions at the different stages (smaller version)
# 1st grid
N <- n[1]  # no. of repetitions 
psmall <- c(psmax,max(0.01,(psmax-dists)),max(0.01,(psmax-2*dists)))  ## check against 0
plarge <- c(plmax,max(0.01,(plmax-distl)),max(0.01,(plmax-2*distl)))
message("\n","1st grid")
out <- RePrInDT(data,"class",ctestv,N,plarge=plarge,psmall=psmall,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket,valdat=valdat)
#cat("Accuracies of best trees for different combinations of resampling percentages","\n")
#apply(out$acc1st,3,print)
im <- which.max(out$acc1st[,,5])
acc <- out$acc1st
dim(acc) <- c(dim(acc)[1] * dim(acc)[2],dim(acc)[3])
accbest <- acc[im,]
percl <- acc[im,1]  # undersampling percentage of the larger class
percs <- acc[im,2] # undersampling percentage of the smaller class
#message("Accuracies of best tree of resampling percentages","\n")
#message(acc[im,])
# 2nd grid
if (steps > 1){
N <- n[2]
psmall <- c(max(0.01,acc[im,2]-incs[2]),acc[im,2],min(1,acc[im,2]+incs[2])) ## check against 0 and 1
plarge <- c(max(0.01,acc[im,1]-incl[2]),acc[im,1],min(1,acc[im,1]+incl[2]))
message("\n","2nd grid")
out <- RePrInDT(data,"class",ctestv,N,plarge=plarge,psmall=psmall,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket,valdat=valdat)
#cat("Accuracies of best trees for different combinations of resampling percentages","\n")
#apply(out$acc1st,3,print)
im <- which.max(out$acc1st[,,5])
acc <- out$acc1st
dim(acc) <- c(dim(acc)[1] * dim(acc)[2],dim(acc)[3])
accbest <- acc[im,]
percl <- acc[im,1]  # undersampling percentage of the larger class
percs <- acc[im,2] # undersampling percentage of the smaller class
#message("Accuracies of best tree of resampling percentages","\n")
#message(acc[im,])
}
# 3rd grid
if (steps > 2){
N <- n[3]
psmall <- c(max(0.01,acc[im,2]-incs[3]),acc[im,2],min(1,acc[im,2]+incs[3]))  ## check against 0 and 1
plarge <- c(max(0.01,acc[im,1]-incl[3]),acc[im,1],min(1,acc[im,1]+incl[3]))
message("\n","3rd grid")
out <- RePrInDT(data,"class",ctestv,N,plarge=plarge,psmall=psmall,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket,valdat=valdat)
#cat("Accuracies of best trees for different combinations of resampling percentages","\n")
#apply(out$acc1st,3,print)
im <- which.max(out$acc1st[,,5])
acc <- out$acc1st
dim(acc) <- c(dim(acc)[1] * dim(acc)[2],dim(acc)[3])
accbest <- acc[im,]
percl <- acc[im,1]  # undersampling percentage of the larger class
percs <- acc[im,2] # undersampling percentage of the smaller class
#message("Accuracies of best tree of resampling percentages","\n")
#message(acc[im,])
}
##########################################
## 3. Best model
############################
N <- n[4]  # no. of repetitions 
## call of PrInDT
message("\n","Final evaluation")
out <- PrInDT(data,"class",ctestv,N,percl,percs,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket,valdat=valdat)
accbest <- out$ba1st
result <- list(tree1st=out$tree1st,bestba=accbest,percl=percl,percs=percs)
class(result) <- "OptPrInDT"
result
}
#' @export
print.OptPrInDT <- function(x,...){
#  NextMethod()
  print(x$tree1st)
  cat("\n")
  cat("Best balanced accuracy on full sample: ",x$bestba[3],"\n")
  cat("Best undersampling percentages: large class:",x$percl,", small class:",x$percs,"\n\n")
}
#' @export
plot.OptPrInDT <- function(x,...){
#  NextMethod()
  plot(x$tree1st,main="Best interpretable classification tree")
  invisible(x)
}

#' Nested \code{\link{PrInDT}} with additional undersampling of a factor with two unbalanced levels 
#'
#' @description Function for additional undersampling of the factor 'nesvar' with two unbalanced levels to avoid dominance of the level with higher frequency.
#' The factor 'nesvar' is allowed not be part of the input data frame 'datain'. The data of this factor is given in the vector 'nesunder'. 
#' The observations in 'nesunder' have to represent the same cases as in 'datain' in the same ordering.\cr
#' \code{\link{PrInDT}} is called 'repin' times with subsamples of the original data so that the level with the larger frequency in the vector 'nesunder' has 
#' approximately the same number of values as the level with the smaller frequency.\cr
#' Only the arguments 'nesvar', 'nesunder', and 'repin' relate to the additional undersampling, all the other arguments relate to the standard 
#' \code{\link{PrInDT}} procedure. \cr As in \code{\link{PrInDT}}, the aim is to optimally model the relationship between the two-class factor variable 'classname' and all other factor and  
#' numerical variables in the data frame 'datain' by means of 'N' repetitions of undersampling. The trees generated by \code{\link{PrInDT}} can be
#' restricted by excluding unacceptable trees which include split results specified in the character strings of the vector 'ctestv'.\cr
#' The probability threshold 'thres' for the prediction of the smaller class may be specified (default = 0.5).\cr
#' Undersampling may be stratified in two ways by the feature 'strat'.\cr
#' The results are evaluated on the full sample and on the subsamples of 'nesunder'.
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' \strong{Reference} \cr Weihs, C., Buschfeld, S. 2021b. NesPrInDT: Nested undersampling in PrInDT. 
#'	arXiv:2103.14931
#'
#' @usage NesPrInDT(datain,classname,ctestv=NA,N,plarge,psmall=1.0,conf.level=0.95,
#'        thres=0.5,stratvers=0,strat=NA,seedl=TRUE,nesvar,nesunder,repin,
#'        minsplit=NA,minbucket=NA)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param N Number of repetitions (integer > 0)
#' @param plarge Undersampling percentage of larger class (numerical, > 0 and <= 1)
#' @param psmall Undersampling percentage of smaller class (numerical, > 0 and <= 1);\cr
#'     default = 1
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param thres Probability threshold for prediction of smaller class; default = 0.5
#' @param stratvers Version of stratification;\cr
#'     = 0: none (default),\cr
#'     = 1: stratification according to the percentages of the values of the factor variable 'strat',\cr
#'     > 1: stratification with minimum number 'stratvers' of observations per value of 'strat'
#' @param strat Name of one (!) stratification variable for undersampling (character);\cr
#'     default = NA (no stratification)
#' @param seedl Should the seed for random numbers be set (TRUE / FALSE)?\cr
#'     default = TRUE
#' @param nesvar Name of factor to be undersampled (character)
#' @param nesunder Data of factor to be undersampled (integer)
#' @param repin Number of repetitions (integer) for undersampling of 'nesvar'
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return
#' \describe{  
#' \item{undba}{balanced accuracies on undersamples}
#' \item{imax}{indices of best trees on undersamples}
#' \item{undba3en}{balanced accuracies of ensembles of 3 best trees on undersamples}
#' \item{accF}{balanced accuracies on full sample}
#' \item{accE}{balanced accuracy on full sample of best ensemble of 3 trees from undersampling}
#' \item{maxt}{indices of best trees on full sample}
#' \item{treesb}{3 best trees of all undersamples of 'nesunder'; refer to an individual tree as \code{treesb[[k]]}, k = 1, ..., 3*repin}
#' }
#' 
#' @details
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)}  where 'name' is the output data 
#' frame of the function.\cr
#' The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
#' \code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
#'
#' @export NesPrInDT
#' @import MASS
#' @import party
#'
#' @examples
#' # data input and preparation --> data frame with 
#' #   class variable, factors, and numericals (no character variables)!!
#' data <- PrInDT::data_speaker
#' data <- na.omit(data)
#' nesvar <- "SPEAKER"
#' N <- 49  # no. of repetitions in inner loop
#' plarge <- 0.06 # sampling percentage for larger class in nesunder-subsample
#' psmall <- 1 # sampling percentage for smaller class in nesunder-subsample
#' nesunder <- data$SPEAKER
#' data[,nesvar] <- list(NULL)
#' outNes <- NesPrInDT(data,"class",ctestv=NA,N,plarge,psmall,conf.level=0.95,nesvar=nesvar,
#'   nesunder=nesunder,repin=5)
#' outNes
#' plot(outNes)
#' hist(outNes$undba,main=" ",xlab = "balanced accuracies of 3 best trees of all undersamples")
#'
NesPrInDT <- function(datain,classname,ctestv=NA,N,plarge=NA,psmall=1.0,conf.level=0.95,thres=0.5,stratvers=0,strat=NA,seedl=TRUE,nesvar,nesunder,repin,
                                     minsplit=NA,minbucket=NA){
  ## input check
  if (typeof(datain) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) || N <= 0 ||
      !((0 < plarge & plarge <= 1) | typeof(plarge)=="logical") || !(0 < psmall & psmall <= 1) || !(0 < conf.level & conf.level <= 1) || !(0 <= thres & thres <= 1) ||
      !(0 <= stratvers) || !(typeof(strat) %in% c("logical", "character")) || typeof(seedl) != "logical" || typeof(nesvar) != "character" || 
      typeof(nesunder) != "integer" || typeof(repin) != "double" || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
  ##
  if (seedl == TRUE){
    set.seed(87654321)  # seed of random numbers
  }
  ###
  data <- datain
  names(data)[names(data)==classname] <- "class"
  n_class1 <- table(data$class)[1] # no. of elements of class 1
  n_class2 <- table(data$class)[2] # no. of elements of class 2
  if (min(n_class1,n_class2) == 0){
   stop("irregular input: only 1 class")
  }
  if (n_class1 < n_class2){
    # relevel of classes if smaller class first
    data$class <- stats::relevel(data$class, levels(data$class)[2]) # larger class now first
    n_class1 <- table(data$class)[1] # no. of elements of larger class 1
    n_class2 <- table(data$class)[2] # no. of elements of smaller class 2
  }
  n <- n_class1 + n_class2 # overall no. of observations in data
  ##
  n_classind1 <- table(nesunder)[1] # no. of elements of class 1 in nesunder
  n_classind2 <- table(nesunder)[2] # no. of elements of class 2 in nesunder
  if (n_classind1 < n_classind2){
    # relevel of classes if smaller class first
    nesunder <- stats::relevel(nesunder, levels(nesunder)[2]) # larger class now first
    n_classind1 <- table(nesunder)[1] # no. of elements of larger class 1 in nesunder
    n_classind2 <- table(nesunder)[2] # no. of elements of smaller class 2 in nesunder
  }
  n_ind <- n_classind1 + n_classind2 # no. of elements in nesunder
  if (n_ind != n){
    stop("Independent and dependent variables do not have the same length")
  }
  ### initialisation
  undba <- matrix(0,ncol=3,nrow=repin)
  undba3en <- rep(0,repin)
  index <- matrix(0,ncol=table(nesunder)[2],nrow=repin)
  ##
  # Start of predictor resampling loop
  for (j in 1:repin){
    message("repetition:", j)
    order_class2 <- order(as.numeric(nesunder),decreasing=TRUE)
    x <- data[order_class2,] # data now reordered: class2 first in nesunder
    index[j,] <- sample.int(n_classind1,n_classind2) + n_classind2 # class2 fully included
    x <- x[index[j,],]
    y <- rbind(data[as.integer(nesunder) == 2,],x) # data is reordered: class2 first (with indund resampling)
    y <- as.data.frame(y)
    ## call of PrInDT (without re-setting seed for random numbers)
    outde <- PrInDT(y,classname,ctestv,N,percl=plarge,percs=psmall,conf.level,seedl=FALSE,minsplit=minsplit,minbucket=minbucket)
    ##
    undba[j,] <- c(outde$ba1st[3],outde$ba2nd[3],outde$ba3rd[3])
    undba3en[j] <- outde$baen[2,3]
    if (j == 1){
      treesb <- c(outde$tree1st,outde$tree2nd,outde$tree3rd)
    } else {
      treesb <- c(treesb,outde$tree1st,outde$tree2nd,outde$tree3rd)
    }
  }
  ###
  ## evaluation
  ###  indices of trees with maximum balanced accuracy on undersamples
  imax <- which(t(undba) == max(undba)) # output
  ## balanced accuracies on full sample
  accF <- rep(0,length(treesb))
  for (i in 1:length(treesb)){
    ct <- treesb[[i]]
    ctpreds <- as.integer(predict(ct,newdata=data))
    preds1 <- ctpreds[as.integer(data$class) == 1]
    preds2 <- ctpreds[as.integer(data$class) == 2]
    accF1 <- sum (preds1 == 1) / n_class1
    accF2 <- sum (preds2 == 2) / n_class2
    accF[i] <- (accF1 + accF2) / 2
  }
  mbaF <- max(accF) # output
  maxt <- which(accF == mbaF) # output: indices of trees with maximum balanced accuracy on full sample
  ##
  ### ensemble of the 3 best trees on undersamples
  modus <- function(x){which.max(tabulate(x))} # definition of Mode function
  three <- sort(undba,decreasing=TRUE,index.return=TRUE)$ix[1:3]
  treesE <- treesb[three]
  preds_orderedE <- matrix(0,nrow=dim(data)[1],ncol=length(treesE))
  for (i in 1:length(treesE)){
    preds_orderedE[,i] <- predict(treesE[[i]],newdata=data)
  }
  classE <- as.integer(data$class)
  n_class1 <- table(data$class)[1] # no. of elements of larger class 1
  n_class2 <- table(data$class)[2] # no. of elements of smaller class 2
  preds1 <- preds_orderedE[classE == 1,]
  preds2 <- preds_orderedE[classE == 2,]
  accE1 <- sum (apply(preds1,1,modus) == 1) / n_class1
  accE2 <- sum (apply(preds2,1,modus) == 2) / n_class2
  accE <- (accE1 + accE2) / 2
### results
  message("\n")
  result <- list( undba = undba, imax = imax, undba3en = undba3en, accF = accF, accE = accE, maxt = maxt, treesb = treesb) 
  class(result) <- "NesPrInDT"
  result
}
#' @export
print.NesPrInDT <- function(x, ...){
cat("\n","Evaluation of nested PrInDT","\n\n","************************","\n")
  ## balanced accuracies on undersamples of nesunder
  cat(" results for undersamples","\n","************************","\n")
  mba <- max(x$undba) # output
  cat("best balanced accuracy of undersampled trees","\n")
  cat(unname(mba))
  cat("\n","table of balanced accuracies (with no. of occurences) of 3 best trees on all undersamples","\n")
  print(table(unname(x$undba)))
  cat("\n","Best trees on undersamples","\n")
  for (k in 1:length(x$imax)){
    print(x$treesb[[x$imax[k]]])
  }
  ## balanced accuracies on full sample
  cat("\n","***************************","\n")
  cat(" results for the full sample","\n","***************************","\n")
  cat("balanced accuracy of best undersampled tree on full sample","\n")
  cat(unname(x$accF[round(x$imax[1],digits=6)]))
  cat("\n","table of balanced accuracies (with no. of occurences) of 3 best trees of all replications on full sample")
  print(table(unname(round(x$accF,digits=6))))
  cat("\n","Best trees on full sample","\n")
  for (k in 1:length(x$maxt)){
    print(x$treesb[[x$maxt[k]]])
  }
  ##
  cat("\n","*************************************","\n","results for ensembles of 3 best trees",
      "\n","*************************************","\n")
  ### ensemble of the 3 best trees on undersamples
  cat("balanced accuracy on full sample of ensemble of best 3 trees from undersampling","\n")
  cat(unname(x$accE))
  ### best ensemble of 3 trees from undersampling
  cat("\n")
  cat("best balanced accuracy on undersample of ensemble of best 3 trees from undersampling","\n")
  cat(unname(max(x$undba3en)),"\n")
}
#' @export
plot.NesPrInDT <- function(x, ...){
## Best trees on undersamples
if (length(x$imax) > 1){
  for (k in 1:length(x$imax)){
    plot(x$treesb[[x$imax[k]]],main=paste0("One of the ",length(x$imax)," best trees on undersamples"))
  }
}
else {
  plot(x$treesb[[x$imax[1]]],main="The best tree on undersamples")
}  
## Best trees on full sample
if (length(x$maxt) > 1){
  for (k in 1:length(x$maxt)){
    plot(x$treesb[[x$maxt[k]]],main=paste0("One of the ",length(x$maxt)," best trees on full sample"))
  }
}
else {
  plot(x$treesb[[x$maxt[1]]],main="The best tree on full sample")
} 
}

#' Two-stage estimation for classification-regression mixtures
#'
#' @description The function Mix2SPrInDT applies 'N' repetitions of subsampling for finding an optimal subsample to model
#' the relationship between the dependent variables specified in the sublist 'targets' of the list 'datalist' and all other factor and numerical variables
#' in the corresponding data frame specified in the sublist 'datanames' of the list 'datalist' in the same order as 'targets'. \cr
#' The function is prepared to handle classification tasks with 2 or more classes and regression tasks. At first stage, the targets are estimated only on
#' the basis of the exogenous predictors. At second stage, summaries of the predictions of the endogenous features from the first stage models are added as new predictors.\cr
#' Subsampling of observations and predictors uses the percentages specified in the matrix 'percent', one row per estimation taks. For classification tasks, 
#' percentages 'percl' and 'percs' for the larger and the smaller class have to be specified. For regression tasks, 'pobs' and 'ppre' have to specified 
#' for observations and predictors, respectively.\cr
#' For generating summaries, the variables representing the substructure have to be specified in the sublist 'datastruc' of 'datalist'.\cr
#' The sublist 'summ' of 'datalist' includes for each discrete target the classes for which you want to calculate the summary percentages and for each 
#' continuous target just NA (in the same order as in the sublist 'targets'). For a discrete target in this list, you can provide a sublist of classes to be combined 
#' (for an example see the below example).\cr
#' The optimization citerion is the balanced accuracy for classification and R2 for regression, both evaluated on the full sample.  \cr
#' The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' @usage Mix2SPrInDT(datalist,ctestv=NA,N=99,percent=NA,conf.level=0.95,minsplit=NA,minbucket=NA)
#'
#' @param datalist list(datanames,targets,datastruc,summ) Input data: For specification see the above description
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param N Number of repetitions of subsampling for predictors (integer); default = 99
#' @param percent matrix of percent spefications: For specification see the above description; default: 'percent = NA' meaning default values for percentages.
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1); default = 0.95
#' @param minsplit Minimum number of elements in a node to be splitted; default = 20
#' @param minbucket Minimum number of elements in a node; default = 7
#'
#' @return
#' \describe{
#'   \item{models1}{Best trees at stage 1} 
#'   \item{models2}{Best trees at stage 2} 
#'   \item{depnames}{names of dependent variables}
#'   \item{nmod}{number of models of tasks}
#'   \item{nlev}{levels of tasks}
#'   \item{accAll}{accuracies of best trees at both stages}  
#' }
#'
#' @details
#' See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.1, for further information.
#'
#' Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name} where 'name' is the output data 
#' frame of the function.\cr
#'
#' @export Mix2SPrInDT
#' @importFrom stats aggregate
#' @importFrom stringr str_split
#' @importFrom gdata mv
#'
#' @examples
#' # zero data
#' datazero <- PrInDT::data_zero
#' datazero <- na.omit(datazero) # cleaned full data: no NAs
#' names(datazero)[names(datazero)=="real"] <- "zero"
#' CHILDzero <- PrInDT::participant_zero
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
#' ##
#' # multi-level data
#' datastrat <- PrInDT::data_zero
#' datamult <- na.omit(datastrat)
#' # ctestv <- NA
#' datamult$mult[datamult$ETH %in% c("C1a","C1b","C1c") & datamult$real == "zero"] <- "zero1"
#' datamult$mult[datamult$ETH %in% c("C2a","C2b","C2c") & datamult$real == "zero"] <- "zero2"
#' datamult$mult[datamult$real == "realized"] <- "real"
#' datamult$mult <- as.factor(datamult$mult) # mult is new class variable
#' datamult$real <- NULL # remove old class variable
#' CHILDmult <- CHILDzero
#' ##
#' # vowel data
#' data <- PrInDT::data_vowel
#' data <- na.omit(data)
#' CHILDvowel <- data$Nickname
#' data$Nickname <- NULL
#' syllable <- 3 - data$syllables
#' data$syllabels <- NULL
#' data$syllables <- syllable
#' data$speed <- data$word_duration / data$syllables
#' names(data)[names(data) == "target"] <- "vowel"
#' datavowel <- data
#' ##
#' # function preparation and call
#' datanames <- list("datazero","datamult","datavowel")
#' targets <- c("zero","mult","vowel")
#' datastruc <- list(CHILDzero,CHILDmult,CHILDvowel)
#' summult <- paste("zero1","zero2",sep=",")  
#' summ <- c("zero",summult,NA)
#' datalist <- list(datanames=datanames,targets=targets,datastruc=datastruc,summ=summ)
#' percent <- matrix(NA,nrow=3,ncol=2)
#' percent[1,] <- c("percl=0.075","percs=0.9") # percentages for datazero
#' # no percentages needed for datapast
#' percent[3,] <- c("pobs=0.9","ppre=c(0.9,0.8)") # percentages for datavowel
#' out2SMix <- Mix2SPrInDT(datalist,ctestv=ctestv,N=19,percent=percent,conf.level=0.99)
#' out2SMix
#' plot(out2SMix)
#'
Mix2SPrInDT <- function(datalist,ctestv=NA,N=99,percent=NA,conf.level=0.95,minsplit=NA,minbucket=NA){
## input check
if ( typeof(datalist) != "list" || !(typeof(ctestv) %in% c("logical", "character")) 
    || !(typeof(percent) %in% c("logical","character"))
    || !(typeof(N) %in% c("integer","double")) || !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) 
    || !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
######################
#### interdependent system of classification and regression models
## estimate with observables only
#
  anz <- length(datalist$datanames)
  if (all(is.na(percent) == TRUE)){
    percent <- matrix(NA,nrow=anz,ncol=2)
  }
  nmod <- c(1:anz)
  nlev <- list()
  acc <- c(1:anz)*0
  datat <- datalist[[1]]
  summ <- datalist[[4]]
  inds <- list()
  extdata <- list()
  modelsP <- list()
  for (K in 1:anz){
    data <- as.data.frame(eval(parse(text = datat[K])))    
    dataname <- datat[K]
    target <- datalist$targets[K]
    ind <- c(1:dim(data)[1])
#print(levels(as.factor(data[,target])))
message(dataname,": 1st stage")
# 2 classes: call of PrInDT
    if (length(levels(data[,target])) == 2){
      nmod[K] <- 1
      nlev <- c(nlev,"class")
# assigning percl, percs
      percl <- NA
      percs <- 1
      x <- str_split(percent[K,1],"=")
      assign(x[[1]][1],eval(parse(text=percent[K,1])))
      x <- str_split(percent[K,2],"=")
      assign(x[[1]][1],eval(parse(text=percent[K,2])))
      if (!(all(0 < percl & percl <= 1) || is.na(percl)==TRUE) || !(all(0 < percs & percs <= 1))){
        stop("irregular input for percl or percs")
      } 
      out2S <- PrInDT(data,target,ctestv,N,percl=percl,percs=percs,conf.level,minsplit=minsplit,minbucket=minbucket)
      sub <- unlist(datalist$datastruc[K])
      sub <- droplevels(sub)
      s <- unlist(str_split(summ[K],","))
#      si <- which(levels(data[,target])==s)
#      snotreal <- table(sub,data[,target])[,si]
#      sc <- as.character(sub)
#      z1 <- aggregate(data[,target] ~ sc,data=cbind(sc,data[,target]),FUN="length")
      ctpreds <- predict(out2S$tree1st,newdata=data)  ## !!
      si <- which(levels(ctpreds)==s)
      snotreal <- table(sub,ctpreds)[,si]
      sc <- as.character(sub)
      z1 <- aggregate(ctpreds ~ sc,data=cbind(sc,ctpreds),FUN="length")
      ind <- snotreal / z1[,2]
      inds[[K]] <- ind
#      print(length(inds[[K]]))
      extdata[[K]] <- paste0(dataname,"i")
      acc[K] <- out2S$ba1st[3]
      modelsP <- c(modelsP,out2S$tree1st)
    } 
# more than 2 classes: call of PrInDTMulev
    if (length(levels(data[,target])) > 2){
      nmod[K] <- length(levels(data[,target]))
      nlev <- c(nlev,levels(data[,target]))
      out2S <- PrInDTMulev(data,target,ctestv,N,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket) # percentages chosen automatically
      sub <- unlist(datalist$datastruc[K])
      sub <- droplevels(sub)
      s <- unlist(str_split(summ[K],","))
#      si <- which(levels(data[,target]) %in% s)
#      snotmarked <- table(sub,data[,target])[,si[1]] + table(sub,data[,target])[,si[2]]
#      sc <- as.character(sub)
#      p1 <- aggregate(data[,target] ~ sc,data=cbind(sc,data[,target]),FUN="length")
      ctpreds <- predict(out2S,target,newdata=data)
#print(table(ctpreds,data[,target]))
      si <- which(levels(ctpreds) %in% s)
      snotmarked <- table(sub,ctpreds)[,si[1]]
      if (length(s) > 1){
        for (i in 2:length(s)){
          snotmarked <- snotmarked + table(sub,ctpreds)[,si[i]]
        }
      }
      sc <- as.character(sub)
      ctpreds <- as.character(ctpreds)
      p1 <- aggregate(ctpreds ~ sc,data=cbind(sc,ctpreds),FUN="length")
      ind <- snotmarked / p1[,2]
      inds[[K]] <- ind
#     print(length(inds[[K]]))
      extdata[[K]] <- paste0(dataname,"i")
      acc[K] <- out2S$ba
      modelsP <- c(modelsP,out2S$trees)
     }
# regression: call of PrInDTreg
     if (length(levels(data[,target])) == 0){
       nmod[K] <- 1
       nlev <- c(nlev,"real")
# assigning pobs, ppre
       pobs <- c(0.9,0.7)
       ppre <- c(0.9,0.7) 
       x <- str_split(percent[K,1],"=")
       assign(x[[1]][1],eval(parse(text=percent[K,1])))
       x <- str_split(percent[K,2],"=")
      assign(x[[1]][1],eval(parse(text=percent[K,2])))
      if (!(all(0 < pobs & pobs <= 1)) || !(all(0 < ppre & ppre <= 1))){
          stop("irregular input for pobs or ppre")
      } 
      out2S <- PrInDTreg(data,target,ctestv,N,pobs=pobs,ppre=ppre,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)
      sub <- unlist(datalist$datastruc[K])
      sc <- as.character(sub)
#      v1 <- aggregate(data[,target] ~ sc,data=cbind(sc,data[,target]),FUN="mean")
      ctpreds <- predict(out2S$ctmax,newdata=data)
      v1 <- aggregate(ctpreds ~ sc,data=cbind(sc,ctpreds),FUN="mean")
      ind <- v1[,2]
      names(ind) <- v1[,1]
      inds[[K]] <- ind
#      print(length(inds[[K]]))
      extdata[[K]] <- paste0(dataname,"i")
      acc[K] <- out2S$maxR2
      modelsP <- c(modelsP,out2S$ctmax)
    }
#   cat("Best model for",target,"\n")
#   print(out2S)
#   plot(out2S)
 }
accP <- acc
#print(accP)
#print(extdata)
#################
## 2nd stage
#################
## extended datasets
for (K in 1:anz){
  datai <- as.data.frame(eval(parse(text = datat[K])))    
  mv("datai",extdata[[K]])
}
## name definitions
for (K in 1:anz){
  datai <- eval(parse(text = extdata[K]))
  for (i in 1:anz){
    if (i != K){
      ind <- inds[[i]]
      D <- dim(datai)[2]
      datai$indi <- mean(ind) ## NA
      for (j in 1:length(ind)) {
        datai$indi[unlist(datalist$datastruc[K]) == names(ind)[j]] <- ind[j]
      }
   names(datai)[(D+1)] <- paste0("p",datalist$targets[i])
   }
 }
mv("datai",extdata[[K]])
#print(str(eval(parse(text = extdata[K]))))
} 
#
#### estimation
#
acc <- c(1:anz)*0
models <- list()
message("\n","Second stage")
for (K in 1:anz){
   datai <- eval(parse(text = extdata[K]))
   target <- datalist$targets[K]
message(datat[K],": 2nd stage")
## 2 classes
   if (length(levels(datai[,target])) == 2){
     x <- str_split(percent[K,1],"=")
     assign(x[[1]][1],eval(parse(text=percent[K,1])))
     x <- str_split(percent[K,2],"=")
     assign(x[[1]][1],eval(parse(text=percent[K,2])))
     out2S <- PrInDT(datai,target,ctestv,N,percl=percl,percs=percs,conf.level,minsplit=minsplit,minbucket=minbucket) # best percentages found
     acc[K] <- out2S$ba1st[3]
     models <- c(models,out2S$tree1st)
  }
## more than 2 classes
  if (length(levels(datai[,target])) > 2){
      datai[,target] - as.factor(datai[,target])
      out2S <- PrInDTMulev(datai,target,ctestv,N,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket) # percentages chosen automatically
      message("balanced accuracies of class models: ",paste(round(out2S$acc,3),collapse=", "))  ## individual balanced accuracies
      acc[K] <- out2S$ba
      models <- c(models,out2S$trees)
  }
## regression
  if (length(levels(datai[,target])) == 0){
    x <- str_split(percent[K,1],"=")
    assign(x[[1]][1],eval(parse(text=percent[K,1])))
    x <- str_split(percent[K,2],"=")
    assign(x[[1]][1],eval(parse(text=percent[K,2])))
    out2S <- PrInDTreg(datai,target,ctestv,N,pobs=pobs,ppre=ppre,conf.level=conf.level,minsplit=minsplit,minbucket=minbucket)
    acc[K] <- out2S$maxR2
    models <- c(models,out2S$ctmax)
  }
#   cat("Best model for",target,"\n")
#   print(out2S)
#   plot(out2S)
}
accP <- cbind(accP,acc)
#rownames(accP) <- datalist$targets
#colnames(accP) <- c("acc","acc2S")
result <- list(models1=modelsP,models2=models,depnames=unlist(datalist$target),nmod=nmod,nlev=nlev,accAll=accP)
class(result) <- "Mix2SPrInDT"
result
}
#' @export
print.Mix2SPrInDT <- function(x,...){
## output
cat("Accuracies of best models","\n")
rownames(x$accAll) <- c(x$depnames)
colnames(x$accAll) <- c("1st stage","2nd stage")
print(x$accAll)
cat("\n\n")
J <- 0
for (i in 1:length(x$depnames)){
  if (x$nmod[i] == 1){
    J <- J + 1
    cat("Best model 2nd stage: ",x$depnames[i], "\n") 
    print(x$models2[[i]])
    cat("\n\n")
  } else {
    K <- x$nmod[i]
    for (k in 1:K){
       J <- J + 1
       cat("Best model 2nd stage: ",x$depnames[i],":",x$nlev[[J]],"\n") 
       print(x$models2[[i]])
       cat("\n\n")
    }
  }
}}
#' @export
plot.Mix2SPrInDT <- function(x,...){
J <- 0
for (i in 1:length(x$depnames)){
  if (x$nmod[i] == 1){
    J <- J + 1
    plot(x$models2[[i]],main=paste0("2nd stage: ",x$depnames[i]))
  } else {
    K <- x$nmod[i]
    for (k in 1:K){
       J <- J + 1
       plot(x$models2[[i]],main=paste0("2nd stage: ",x$depnames[i],":",x$nlev[[J]]))
    }
  }
}}

#' Subject pronouns 
#'
#' @description
#' Usage of subject pronouns and its predictors.
#' 
#' @format
#' A data frame with 1024 observations and 7 columns
#' \describe{
#'   \item{real}{subject pronoun realized? Factor (2 levels "zero","realized")}
#'   \item{AGE}{age: Numerical (in months)}
#'   \item{LiBa}{linguistic background: Factor (3 levels "mono","multi", NA)}
#'   \item{ETH}{ethnic group: Factor (6 levels "C1a","C1b","C1c","C2a","C2b","C2c") (anonymized)}
#'   \item{SEX}{gender: Factor (2 levels "female","male")}
#'   \item{MLU}{mean length of utterance: Factor (4 levels "1","2","3","OL")}
#'   \item{PRN}{pronoun: Factor (7 levels "I","you_s","he","she","it","we","they")}
#' }
#' @source
#' Sarah Buschfeld, TU Dortmund
"data_zero"
#' Participants of subject pronoun study
#'
#' @description
#' Participants (random excerpt according to data_zero)
#'
#' @format
#' A dataset with 1024 observations and 1 column
#' \describe{
#'   \item{participant}{participant of study (P1 - P51)}
#' }
#' @source
#' Sarah Buschfeld, TU Dortmund
"participant_zero"
#' Subject pronouns and a predictor with one very frequent level
#'
#' @description
#' Usage of subject pronouns and its predictors; speaker level "adult" very frequent.
#' 
#' @format
#' A data frame with 3370 observations and 6 columns
#' \describe{
#'   \item{class}{subject pronoun realized? Factor (2 levels "zero","realized")}
#'   \item{AGE}{age: Numerical (in months)}
#'   \item{ETHN_GROUP}{ethnic group: Factor (3 levels "C","I","n_a") (anonymized)}
#'   \item{MLU}{mean length of utterance: Factor (5 levels "1","2","3","adult","OL")}
#'   \item{PRN_TYPE}{pronoun type: Factor (5 levels "dem","it_con","it_ex","it_ref","refer")}
#'   \item{SPEAKER}{speaker: Factor (2 levels "adult","child")}
#' }
#' @source
#' Sarah Buschfeld, TU Dortmund
"data_speaker"
#' Vowel length
#'
#' @description
#' Vowel length and categorization criteria.
#' 
#' @format
#' A data frame with 82 observations and 22 columns
#' \describe{
#'   \item{Nickname}{nickname: Factor (43 levels "Nick1","Nick2",...,"Nick43") (anonymized)}
#'   \item{LiBa}{linguistic background: Factor (2 levels "mono","multi")}
#'   \item{MLU}{mean length of utterance: Factor (3 levels "1","2","3")}
#'   \item{phone_label}{phone label: Factor (2 levels "fleece","kit")}
#'   \item{lexeme}{lexeme: Factor (14 levels "bee","cheek","cheese","chicken", ...)}
#'   \item{phone_left_1_duration}{duration of phone to the left of the vowel: Numerical (in msec)}
#'   \item{phone_right_1_duration}{duration of phone to the right of the vowel: Numerical (in msec)}
#'   \item{word_duration}{duration of word: Numerical (in msec)}
#'   \item{vowel_minimum_pitch}{minimum pitch of vowel: Numerical (in Hertz)}
#'   \item{vowel_maximum_pitch}{maximum pitch of vowel: Numerical (in Hertz)}
#'   \item{vowel_intensity_mean}{mean intensity of vowel: Numerical (in decibel)}
#'   \item{f1_fifty}{first formant F1 at midpoint of vowel (50\%): Numerical (in Hertz)}
#'   \item{f2_fifty}{second formant F2 at midpoint of vowel (50\%): numercial (in Hertz)}
#'   \item{target}{vowel length: Numerical (in msec)}
#'   \item{cons_class_l}{class of consonant to the left of the vowel: Factor (6 levels "l","r","tsh",...)}
#'   \item{cons_class_r}{class of consonant to the right of the vowel: Factor (7 levels "?"(glottal stop),"empty","nas",...)}
#'   \item{ETH}{ethnic group: Factor (6 levels "C1a","C1b","C1c","C2a","C2b","C2c") (anonymized)}
#'   \item{SEX}{gender: Factor (2 levels "female","male)}
#'   \item{AGE}{age: Numerical (in months)}
#'   \item{syllables}{number of syllables in lexeme: integer (1,2)}
#'   \item{speed}{speed of speech: Numerical (word duration / syllables; in msec)}
#'   \item{country}{country: Factor (2 levels "E","S")  (anonymized)}
#' } 
#' @source
#' Sarah Buschfeld, TU Dortmund
"data_vowel"
#' Landscape analysis
#'
#' @description
#' The use of language(s) on public signs and categorization criteria.
#' 
#' @format
#' A data frame with 149 observations and 28 columns
#' \describe{
#'   \item{coder}{who coded the sign: Factor (3 levels "C","E","S")  (anonymized)}
#'   \item{researcher}{who took a photograph of the sign: Factor (2 levels "L","S")  (anonymized)}
#'   \item{sign}{where the sign was found: Factor (11 levels "digi","door","graf",...)}
#'   \item{type.of.sign}{kind of sign: Factor (5 levels "com","commem","infra","reg","trans")}
#'   \item{permanent}{was the sign permanent? Factor (2 levels "no","yes")}
#'   \item{proper.noun}{kind of proper noun on the sign: Factor (9 levels "bn","bn+","cn",...)}
#'   \item{no.languages}{number of languages on sign: Factor (4 levels "1","2","3","4+")}
#'   \item{French}{French on sign? Factor (2 levels "0","1")}
#'   \item{Dutch}{Dutch on sign? Factor (2 levels "0","1")}
#'   \item{English}{English on sign? Factor (2 levels "0","1")}
#'   \item{Italian}{Italian on sign? Factor (2 levels "0","1")}
#'   \item{Spanish}{Spanish on sign? Factor (2 levels "0","1")}
#'   \item{German}{German on sign? Factor (2 levels "0","1")}
#'   \item{Indian,Mandarin,Greek,Indonesian,Portuguese,Libanese,Japanese,Russian,Danish,
#'             Norwegian,Hebrew,Catalan}{12 infrequent languages: Factor (2 levels "0","1")}
#'   \item{bn.unclear}{brand name unclear: Factor (2 levels "0","1")}
#'   \item{multilingual.type}{type of multilingualism on sign: Factor (6 levels "0","1","2","3","4","uc")}
#'   \item{location}{location of sign: Factor (2 levels "M","P")  (anonymized)}
#' }
#' @source
#' Sarah Buschfeld, TU Dortmund
"data_land"

#' Conditional inference trees (ctrees) based on consecutive parts of the full sample
#'
#' @description ctrees based on the full sample of the smaller class and consecutive parts of the larger class of the nesting variable 'nesvar'. 
#' The variable 'nesvar' has to be part of the data frame 'datain'.\cr   
#' Interpretability is checked (see 'ctestv'); probability threshold can be specified.
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' \strong{Reference}\cr Weihs, C., Buschfeld, S. 2021b. NesPrInDT: Nested undersampling in PrInDT. 
#'	arXiv:2103.14931
#'
#' @usage PrInDTAllparts(datain, classname, ctestv=NA, conf.level=0.95, thres=0.5,
#'        nesvar, divt,minsplit=NA,minbucket=NA)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1); default = 0.95
#' @param thres Probability threshold for prediction of smaller class (numerical, >= 0 and < 1); default = 0.5
#' @param nesvar Name of nesting variable (character)
#' @param divt Number of parts of nesting variable nesvar for which models should be determined individually
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return
#' \describe{
#' \item{baAll}{balanced accuracy of tree on full sample}
#' \item{nesvar}{name of nesting variable}
#' \item{divt}{number of consecutive parts of the sample}
#' \item{badiv}{balanced accuracy of trees on 'divt' consecutive parts of the sample}
#' }
#'
#' @details
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } where 'name' is the output data 
#' frame of the function.
#'
#' @export PrInDTAllparts
#'
#' @examples
#' data <- PrInDT::data_speaker
#' data <- na.omit(data)
#' nesvar <- "SPEAKER"
#' outNesAll <- PrInDTAllparts(data,"class",ctestv=NA,conf.level=0.95,thres=0.5,nesvar,divt=8)
#' outNesAll
#'
PrInDTAllparts <- function(datain,classname,ctestv=NA,conf.level=0.95,thres=0.5,nesvar,divt,minsplit=NA,minbucket=NA){
  ## input check
  if (typeof(datain) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) || 
      !(0 < conf.level & conf.level <= 1) || !(0 <= thres & thres < 1) ||
      typeof(nesvar) != "character" || divt <= 0 || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
  ##
  data <- datain
  names(data)[names(data)==classname] <- "class"
  n_class1 <- table(data$class)[1] # no. of elements of larger class 1
  n_class2 <- table(data$class)[2] # no. of elements of smaller class 2
  if (min(n_class1,n_class2) == 0){
   stop("irregular input: only 1 class")
  }
  if (n_class1 < n_class2){
    # relevel of classes if smaller class first
    data$class <- stats::relevel(data$class, levels(data$class)[2]) # larger class now first level
    n_class1 <- table(data$class)[1] # no. of elements of larger class
    n_class2 <- table(data$class)[2] # no. of elements of smaller class
  }
  ## full sample
  out <- PrInDTAll(data,"class",minsplit=minsplit,minbucket=minbucket)
  baAll <- out$baAll
  ## analyses of parts
  names(data)[names(data)==nesvar] <- "NEST"
  n_classind1 <- table(data$NEST)[1] # no. of elements of class 1 in nesvar
  n_classind2 <- table(data$NEST)[2] # no. of elements of class 2 in nesvar
  if (n_classind1 < n_classind2){
    # relevel of classes if smaller class first
    data$NEST <- stats::relevel(data$NEST, levels(data$NEST)[2]) # larger class now first level
    n_classind1 <- table(data$NEST)[1] # no. of elements of larger class
    n_classind2 <- table(data$NEST)[2] # no. of elements of smaller class
  }
  ## reordering of classes: smaller class first
  if (n_classind1 > n_classind2){
    order_class2 <- order(as.numeric(data$NEST),decreasing=TRUE)
  } else {
    order_class2 <- order(as.numeric(data$NEST))
  }
  x <- data[order_class2,] # data now reordered: smaller class of nesvar first
  badiv <- vector("numeric",length=divt)
  for (i in 0:(divt-1)){
    datadiv <- rbind (x[1:n_classind2,], x[(n_classind2 + (i*n_classind1/divt)+1):(n_classind2 + (i+1)*n_classind1/divt),])
    outdiv <- PrInDTAll(datadiv,"class",minsplit=minsplit,minbucket=minbucket)
    badiv[i+1] <- outdiv$baAll
  }
  result <- list(baAll = baAll, nesvar = nesvar, divt = divt, badiv = badiv)
  class(result) <- "PrInDTAllparts"
  result 
}
#' @export
print.PrInDTAllparts <- function(x,...){
  cat("\n","Balanced accuracy of tree on full sample: ",unname(x$baAll),"\n")
  cat("\n","Consecutive parts of variable ",x$nesvar,"\n")
  cat("Balanced accuracy of trees on", x$divt, "consecutive parts of the sample","\n")
  for (i in 1:x$divt){
    ba <- x$badiv[i]
    cat("\n","**** part",i," ",unname(ba))
  }
  cat("\n\n")
}
#' @export
plot.PrInDTAllparts <- function(x,...){
  cat("\n","No plots available","\n")
}

#' Regression tree based on all observations
#'
#' @description Regression tree based on the full sample; interpretability is checked (see 'ctestv').\cr
#' The relationship between the target variable 'regname' and all other factor and numerical variables
#' in the data frame 'datain' is modeled based on all observations. \cr 
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#' Besides the maximal R2, the minimal MAE (Mean Absolute Error) is reported.
#'
#' @usage PrInDTregAll(datain,regname,ctestv=NA,conf.level=0.95,minsplit=NA,minbucket=NA)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param regname name of regressand variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return
#' \describe{
#' \item{treeall}{tree based on all observations}
#' \item{R2All}{goodness of fit of 'treeall' based on all observations}
#' \item{MAEAll}{MAE of 'treeall' based on all observations}
#' \item{interpAll}{criterion of interpretability of 'treeall' (TRUE / FALSE)}
#' }
#'
#' @details
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
#' frame of the function.
#'
#' @export PrInDTregAll 
#'
#' @examples
#' data <- PrInDT::data_vowel
#' data <- na.omit(data)
#' ctestv <- 'vowel_maximum_pitch <= 320'
#' outreg <- PrInDTregAll(data,"target",ctestv)
#' outreg
#' plot(outreg)
#'
#' @importFrom party ctree ctree_control
#' @importFrom stats formula
#' @importFrom graphics plot
#'
PrInDTregAll <- function(datain,regname,ctestv=NA,conf.level=0.95,minsplit=NA,minbucket=NA){
  ## input check
  if (typeof(datain) != "list" || typeof(regname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) ||
      !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) || !(typeof(minbucket) %in% c("logical", "double")) ){
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
####
  data <- datain
  ct <- party::ctree(formula(paste(regname, "~ .")), data = data,control = party::ctree_control(mincriterion=conf.level,minsplit=minsplit,minbucket=minbucket))
  ctpreds <- predict(ct)
  R2 <- 1 - sum((data[,regname] - ctpreds)^2) / sum((data[,regname]-mean(data[,regname]))^2)
  MAE <- sum(abs(data[,regname] - ctpreds))/ dim(data)[1]
  crit1 <- "FALSE"
  if (all(is.na(ctestv)) == FALSE) {
    crit1 <- FindSubstr(ct,ctestv) # call of the above function for overall tree
  }
  result <- list(treeAll = ct, R2All = R2, interpAll = crit1, MAEAll = MAE)
  class(result) <- "PrInDTregAll"
  result
}
#' @export
print.PrInDTregAll <- function(x,...){
  cat("\n","Regression tree on full data set","\n")
  print(x$treeAll)
  cat("\n","Interpretability:",unname(!as.logical(x$interpAll)))
  cat("\n","Goodness of Fit R2:",x$R2All)
  cat("\n","MAE:",x$MAEAll,"\n\n")
}
#' @export
plot.PrInDTregAll <- function(x,...){
  plot(x$treeAll,main="Regression tree on full data set")
}
#' @export
predict.PrInDTregAll <- function(object,...){
  predict(object$treeAll,...)
}

#' Conditional inference tree (ctree) based on all observations
#'
#' @description ctree based on all observations in 'datain'. Interpretability is checked (see 'ctestv'); probability threshold can be specified.
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.
#'
#' \strong{Reference} \cr
#' Weihs, C., Buschfeld, S. 2021a. Combining Prediction and Interpretation in  Decision Trees (PrInDT) - 
#' a Linguistic Example. arXiv:2103.02336
#'
#' In the case of repeated measurements ('indrep=1'), the values of the substructure variable have to be given in 'repvar'.
#' Only one value of 'classname' is allowed for each value of 'repvar'.
#' If for a value of 'repvar' the percentage 'thr' of the observed occurence of a value of 'classname' is not reached by the number of predictions of the value of 'classname', a misclassification is detected.
#'
#' @usage PrInDTAll(datain, classname, ctestv=NA, conf.level=0.95, thres=0.5,
#'                  minsplit=NA,minbucket=NA,repvar=NA,indrep=0,thr=0.5)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1); default = 0.95
#' @param thres Probability threshold for prediction of smaller class (numerical, >= 0 and < 1); default = 0.5
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#' @param repvar Values of variable defining the substructure in the case of repeated measurements, length = dim(datain)[1] necessary; default=NA
#' @param indrep Indicator of repeated measurements ('indrep=1'); default = 0
#' @param thr threshold for element classification: minimum percentage of correct class entries; default = 0.5
#'
#' @return
#' \describe{
#' \item{treeall}{ctree based on all observations}
#' \item{baAll}{balanced accuracy of 'treeall'}
#' \item{interpAll}{criterion of interpretability of 'treeall' (TRUE / FALSE)}
#' \item{confAll}{confusion matrix of 'treeall'}
#' \item{acc1AE}{Accuracy of full sample tree on Elements of large class} 
#' \item{acc2AE}{Accuracy of full sample tree on Elements of small class}  
#' \item{bamaxAE}{balanced accuracy of full sample tree on Elements} 
#' \item{namA1}{Names of misclassified Elements by full sample tree of large class}
#' \item{namA2}{Names of misclassified Elements by full sample tree of small class}
#' \item{lablarge}{Label of large class} 
#' \item{labsmall}{Label of small class}
#' \item{thr}{Threshold for repeated measurements} 
#' }
#'
#' @details
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
#' frame of the function.\cr
#'
#' @export PrInDTAll
#'
#' @examples
#' datastrat <- PrInDT::data_zero
#' data <- na.omit(datastrat)
#' ctestv <- rbind('ETH == {C2a,C1a}','MLU == {1, 3}')
#' conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
#' outAll <- PrInDTAll(data,"real",ctestv,conf.level) 
#' print(outAll) # print model based on all observations
#' plot(outAll) # plot model based on all observations
#'
#' @importFrom stats relevel predict
#' @importFrom party ctree ctree_control
#'
#'
PrInDTAll <- function(datain,classname,ctestv=NA,conf.level=0.95,thres=0.5,minsplit=NA,minbucket=NA,repvar=NA,indrep=0,thr=0.5){
  ## input check
  if ( typeof(datain) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character"))
      || !(0 < conf.level & conf.level <= 1) || !(0 <= thres & thres < 1) || !(typeof(minsplit) %in% c("logical","double")) 
      || !(typeof(minbucket) %in% c("logical", "double")) || !(typeof(repvar) %in% c("logical","integer","character"))
      || !(typeof(indrep) %in% c("integer","double")) || typeof(thr) != "double" ){
    stop("irregular input")
  }  
  if (indrep == 1 & all(is.na(repvar)) == TRUE){
    message("\n","Note: repeated measurement analysis disabled: no repeated measurement variable specified","\n")
    indrep <- 0
  }  
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
  ##
  data <- datain
  repv <- repvar
  names(data)[names(data)==classname] <- "class"
  n_class1 <- table(data$class)[1] # no\item{of elements of larger class 1
  n_class2 <- table(data$class)[2] # no\item{of elements of smaller class 2
  if (min(n_class1,n_class2) == 0){
   stop("irregular input: only 1 class")
  }
  n <- n_class1 + n_class2
  if (n_class1 < n_class2){
    # relevel of classes if smaller class first
    data$class <- stats::relevel(data$class, levels(data$class)[2]) # larger class now first
    n_class1 <- table(data$class)[1] # no of elements of larger class
    n_class2 <- table(data$class)[2] # no of elements of smaller class
  }
  ## reordering of classes: smaller class first
  if (n_class1 > n_class2){
    order_class <- order(as.numeric(data$class),decreasing=TRUE)
  } else {
    order_class <- order(as.numeric(data$class))
  }
  data <- data[order_class,] # data now reordered: smaller class first
  if (indrep == 1) {repv <- repv[order_class]}
  if (indrep == 1 & length(repv) != dim(data)[1]){
    stop("irregular input: no. of observations different in 'repvar' and validation set")
  }
  ######
  ## model with all observations
  ######
  ct <- party::ctree(class ~ ., data = data,control = party::ctree_control(mincriterion=conf.level,minsplit=minsplit,minbucket=minbucket)) 
  if (thres != 0.5){
    predsprob <- stats::predict(ct,type="prob")
    ctpreds <- as.factor(sapply( 1:dim(data)[1],
                                 function(x)
                                   ifelse(predsprob[[x]][2] > thres,
                                          levels(data$class)[2], levels(data$class)[1] )))
  } else {
    ctpreds <- stats::predict(ct) # predictions for all observations
}
  conf <- table(ctpreds, data$class)  # confusion matrix calculation
  ## balanced accuracy
  acc1 <- conf[1,1] / n_class1
  acc2 <- conf[2,2] / n_class2
  crit2 <- (acc1 + acc2)/2
  crit1 <- "FALSE"
  if (is.na(ctestv[1]) == FALSE) {
    crit1 <- FindSubstr(ct,ctestv) # call of the above function for overall tree
  }
lablarge <- names(table(data$class))[1] # class 1 = large
labsmall <- names(table(data$class))[2] # class 2 = small 
acc1AE <- 0
acc2AE <- 0
bamaxAE <- 0
namA1 <- list()
namA2 <- list()
if (indrep == 1){
  pred <- predict(ct,newdata=data)
  ch <- table(repv,pred)
  conti <- cbind(table(repv,data$class),ch)
#print(conti)
  no1 <- sum(conti[,1] > 0)
  no2 <- sum(conti[,2] > 0)
  if ((no1 + no2) > dim(conti)[1]){
#    cat("repeated measurement variable has more than one class per value")
    acc2AE <- -1
    acc1AE <- -1 
  } else {
    for (i in 1:dim(conti)[1]){
      if (conti[i,2] > 0 & conti[i,4] <= (conti[i,2]*thr)){ 
        namA2 <- c(namA2,rownames(ch)[i])
      }
      if (conti[i,1] > 0 & conti[i,3] < (conti[i,1]*thr)){ 
        namA1 <- c(namA1,rownames(ch)[i])
      }
    }
    acc2AE <- 1 - length(namA2) / no2
    acc1AE <- 1 - length(namA1) / no1  
    bamaxAE <- (acc1AE + acc2AE)/2
    if (length(namA1) == 0) {namA1 <- "-"}
    if (length(namA2) == 0) {namA2 <- "-"} 
  }
}
###
  result <- list(treeAll = ct, baAll=crit2, interpAll=crit1, confAll = conf,acc1AE=acc1AE,acc2AE=acc2AE,bamaxAE=bamaxAE,
                 namA1=namA1,namA2=namA2,indrep=indrep,lablarge=lablarge,labsmall=labsmall,thr=thr)
  class(result) <- "PrInDTAll"
  result
}

#' PrInDT analysis for a classification problem with multiple classes.

#' @description PrInDT analysis for a classification problem with more than 2 classes. For each combination of one class vs. 
#' the other classes a 2-class \code{\link{PrInDT}} analysis is carried out.\cr
#' The percentages for undersampling of the larger class ('percl' in \code{\link{PrInDT}}) are chosen so that the resulting sizes
#' are comparable with the size of the smaller classes for which all their observations are used in undersampling ('percs' = 1 in \code{\link{PrInDT}}).\cr
#' The class with the highest probability in the K (= number of classes) analyses is chosen for prediction.\cr
#' Interpretability is checked (see 'ctestv').
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' @usage PrInDTMulev(datain,classname,ctestv=NA,N,conf.level=0.95,seedl=FALSE,
#'                                     minsplit=NA,minbucket=NA)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param N Number of repetitions (integer > 0)
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1)\cr
#'     (default = 0.95)
#' @param seedl Should the seed for random numbers be set (TRUE / FALSE)?\cr
#'     default = FALSE
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return 
#' \describe{
#' \item{class}{levels of class variable} 
#' \item{trees}{trees for the levels of the class variable; refer to an individual tree as \code{trees[[k]]}, k = 1, ..., no. of levels}
#' \item{ba}{balanced accuracy of combined predictions} 
#' \item{conf}{confusion matrix of combined predictions} 
#' \item{ninterp}{no. of non-interpretable trees} 
#' \item{acc}{balanced accuracies of best models for individual classes}
#' }
#'
#' @details
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)}  where 'name' is the output data 
#' frame of the function.\cr
#' The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
#' \code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
#'
#' @export PrInDTMulev
#'
#' @examples
#' datastrat <- PrInDT::data_zero
#' data <- na.omit(datastrat)
#' ctestv <- NA
#' data$rel[data$ETH %in% c("C1a","C1b","C1c") & data$real == "zero"] <- "zero1"
#' data$rel[data$ETH %in% c("C2a","C2b","C2c") & data$real == "zero"] <- "zero2"
#' data$rel[data$real == "realized"] <- "real"
#' data$rel <- as.factor(data$rel) # rel is new class variable
#' data$real <- NULL # remove old class variable
#' N <- 51
#' conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
#' out <- PrInDTMulev(data,"rel",ctestv,N,conf.level) 
#' out # print best models based on subsamples
#' plot(out) # corresponding plots
#'
#' @importFrom stats relevel predict
#' @importFrom party ctree ctree_control
#'
PrInDTMulev <- function(datain,classname,ctestv=NA,N,conf.level=0.95,seedl=FALSE,minsplit=NA,minbucket=NA){
  ## input check
  if (typeof(datain) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) || N <= 0 ||
      !(0 < conf.level & conf.level <= 1) || typeof(seedl) != "logical" || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
  if (seedl == TRUE){
    set.seed(7654321)  # set seed of random numbers
  }
  data <- datain
  names(data)[names(data)==classname] <- "class"
  K <- nlevels(data$class)
  n <- dim(data)[1]
  datastud <- data
  datastud$class <- NULL
  ctpredsk <- rep(0,n)
  ctpreds <- matrix(0,nrow=n,ncol=K)
  crit1 <- rep(FALSE,K)
  acci <- rep(0,K)
##
## loop over the classes (one vs. rest)
##
message("loop over all individual classes vs. rest")
 for (k in 1:K){  # loop over the classes
   message("individual class = ",levels(data$class)[k])
   n_class1 <- table(data$class)[k] # no. of elements of studied class
   n_class2 <- n - n_class1 # no. of elements in all other classes
   if (min(n_class1,n_class2) == 0){
     stop("irregular situation: only 1 class left")
   }
   class1 <- as.character(levels(data$class)[k])
   datastud$classstud[data$class == levels(data$class)[k]] <- class1
   datastud$classstud[!(data$class == levels(data$class)[k])] <- "rest"
   datastud$classstud <- as.factor(datastud$classstud)
   names(datastud$classstud) <- "classstud"
  ## determining the percentage for larger class
   if (n_class1 > n_class2){
     perclin <- n_class2 / (n - n_class2) # * 0.8
   } else {
     perclin <- n_class1 / (n - n_class1) # * 0.8
   }
  ######
  ## call of PrInDT model for one class vs. rest 
  ######
   outin <- PrInDT(datastud,"classstud",ctestv,N,percl=perclin,percs=1,conf.level,minsplit=minsplit,minbucket=minbucket)
#   outin <- PrInDT(datastud,"classstud",ctestv,N,percl=perclin,percs=0.8,conf.level)
   ct <- outin$tree1st
   if (k == 1){
     trees <- ct
   }
   else {
     trees <- c(trees,ct)
   }
   acci[k] <- outin$ba1st[3]
   if (is.na(ctestv[1]) == FALSE) {
     crit1[k] <- FindSubstr(ct,ctestv) # call of the above function for overall tree
   }
   x <- t(array(unlist(stats::predict(ct,type="prob",newdata=datastud)),dim=c(2,n)))
   if (n_class1 < n_class2){
     ctpreds[,k] <- x[,2]
   }
   else {
     ctpreds[,k] <- x[,1]
   }
    datastud$classstud <- NULL
 }
## deployment 
## comparison of probabilities of predicted classes: identify maximum
##
  predclassno <- rep(0,n)
  for (i in 1:n){
    predclassno[i] <- which.max(ctpreds[i,1:K])
  }
  predclass <- names(table(data$class))[predclassno]
  conf <- table(predclass,data$class) # confusion matrix
  ## balanced accuracy
  acc <- rep(0,K)
  for (i in 1:length(rownames(conf))){
    for (j in 1:length(colnames(conf))){
      if (rownames(conf)[i] == colnames(conf)[j]){
        acc[j] <- conf[i,j] / colSums(conf)[j]
      }
    }
  }
  crit2 <- sum(acc) / K
  crit1sum <- sum(crit1)
## results
  result <- list(class = levels(data$class), trees = trees, ba=crit2, conf = conf, ninterp=crit1sum, acc=acci)
  class(result) <- "PrInDTMulev"
  result
}
#' @export
print.PrInDTMulev <- function(x, ...){
  # output of models
  K <- length(x$class)
  for (k in 1:K){
    cat("\n","tree for class",x$class[k],"\n")
    cat("balanced accuracy ",x$acc[k],"\n")
    print(x$trees[[k]])
  }
  # output for model characteristics
  cat("\n")
  cat("****************","\n")
  cat("Confusion matrix of combined predictions","\n")
  cat("****************")
  print(x$conf)
  cat("***********************************************","\n")
  cat("Criteria: balanced accuracy","\n")
  cat("***********************************************","\n")
  cat("             ",unname(x$ba),"\n")
  cat("***********************************************","\n")
  cat("Criteria: no. of non-interpretable trees","\n")
  cat("***********************************************","\n")
  cat("             ",unname(x$ninterp),"\n")
}
#' @export
plot.PrInDTMulev <- function(x, ...){
  # plot of models
  K <- length(x$class)
  for (k in 1:K){
    plot(x$trees[[k]],main=paste0("Tree for class = ",x$class[k]))
  }
}
#' @export
predict.PrInDTMulev <- function(object,classname,newdata,...){
K <- nlevels(newdata[,classname])
ctpreds <- matrix(0,nrow=dim(newdata)[1],ncol=K)
  for (k in 1:K)  {
    y <- t(array(unlist(stats::predict(object$trees[[k]],type="prob",newdata=newdata)),dim=c(2,dim(newdata)[1])))
    if (table(newdata[,classname])[k] < dim(newdata)[1]/2){
      ctpreds[,k] <- y[,2]
   } else {
      ctpreds[,k] <- y[,1]
   }
  }
  predclassno <- rep(0,dim(newdata)[1])
  for (j in 1:dim(newdata)[1]){
    predclassno[j] <- which.max(ctpreds[j,1:K])
  }
  predclass <- names(table(newdata[,classname]))[predclassno]
  result <-  as.factor(predclass)
  result
}

#' Interdependent estimation for classification
#'
#' @description The function SimCPrInDT applies interdependent estimation (endogenous case) for finding an optimal model for relationships between the two-class factor variables
#' specified as column indices of 'data' in the vector 'inddep' and all other factor and numerical variables in the data frame 'data' by means of
#' 'N' repetitions of random subsampling with percentages 'percl' for the large classes and 'percs' for the small classes. One percentage of observations 
#' for each dependent variable has to be specified for the larger and the smaller class. For example, for three dependent variables, 'percl' consists of 
#' three percentages specified in the order in which the dependent variables appear in 'inddep'.\cr
#' The dependent variables have to be specified as dummies, i.e. as 'property absent' (value 0) or 'property present' (value 1).\cr
#' The optimization citerion is the balanced accuracy on the full sample. \cr
#' In an additional step, the mean balanced accuracy over all class variables is optimized (joint optimization).\cr
#' The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'. \cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' @usage SimCPrInDT(data,ctestv=NA,inddep,percl,percs,N=99,M,psize,conf.level=0.95,
#'                                   minsplit=NA,minbucket=NA)
#'
#' @param data Input data frame with class factor variables and the influential variables,\cr
#'    which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param inddep indices of dependent variables
#' @param percl list of undersampling percentages of larger class (numerical, > 0 and <= 1): one per dependent class variable in the same order as in 'inddep'
#' @param percs list of undersampling percentage of smaller class (numerical, > 0 and <= 1); one per dependent class variable in the same order as in 'inddep'
#' @param N no. of repetitions of subsampling of observations (integer > 0); default = 99
#' @param M no. of repetitions of subsampling of predictors
#' @param psize no. of predictors in the subsamples of the predictors
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return
#' \describe{
#'   \item{models1}{Best trees at stage 1} 
#'   \item{models2}{Best trees at stage 2} 
#'   \item{models3}{Best trees from mean maximization} 
#'   \item{classnames}{names of classification variables}
#'   \item{baAll}{balanced accuracies of best trees at both stages} 
#' }
#'
#' @details
#' See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.2, for further information.
#'
#' Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name)} where 'name' is the output data 
#' frame of the function.\cr
#
#' @export SimCPrInDT
#'
#' @examples
#' data <- PrInDT::data_land # load data
#' dataclean <- data[,c(1:7,23:24,11:13,22,8:10)]  # only relevant features
#' inddep <- c(14:16) # dependent variables
#' dataland <- na.omit(dataclean)
#' ctestv <- NA
#' perc <- c(0.45,0.05,0.25)   # percentages of observations of larger class, 
#' # 1 per dependent class variable
#' perc2 <- c(0.75,0.95,0.75)  # percentages of observations of smaller class, 
#' # 1 per dependent class variable
#' outland <- SimCPrInDT(dataland,percl=perc,percs=perc2,N=9,inddep=inddep,M=2,psize=10)
#' outland
#' plot(outland)
#'
SimCPrInDT <- function(data,ctestv=NA,inddep,percl,percs,N=99,M,psize,conf.level=0.95,minsplit=NA,minbucket=NA){
## input check
if ( typeof(data) != "list" || !(typeof(ctestv) %in% c("logical", "character")) 
    || !all(0 < percl & percl <= 1) || !all(0 < percs & percs <= 1) 
    || !(typeof(N) %in% c("integer","double")) || !(typeof(M) %in% c("integer","double")) 
    || !(typeof(inddep) %in% c("integer","double")) 
    || !(typeof(psize) %in% c("integer","double")) || !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) 
    || !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
#
## SimulPrInDT: classification
#
message ("1st stage: full sample")
depnames <- colnames(data)[inddep]
anz <- length(inddep)
D <- dim(data)[2]
models <- list()
modmax <- list()
equal <- matrix(TRUE,ncol=anz,nrow=dim(data)[1])
ctpreds <- matrix(0,ncol=anz,nrow=dim(data)[1])
ba <- c(1:anz)*0
for (i in 1:anz){
  outland <- PrInDTAll(data,depnames[i],ctestv=ctestv,conf.level=conf.level)
  ctpredsVL <- predict(outland$treeAll,newdata=data)
  equal[,i] <- (as.integer(data[,9])-1) == ctpredsVL
  ctpreds[,i] <- ctpredsVL
  ba[i] <- outland$baAll
  models[[i]] <- outland$treeAll
}
# Joint assessment: 01, Hamming
#  acc01 <- rowSums(equal)
#  ba[(anz+2)] <- sum(acc01) / (3*dim(data)[1])
#  acc01[acc01 < anz] <- 0
#  acc01[acc01 == anz] <- 1
#  ba[(anz+1)] <- sum(acc01)
#  ba[(anz+1)] <- ba[(anz+1)] / dim(data)[1]
##
#cat("Balanced accuracies for interdependent Models on full sample","\n")
max01 <- ba
#print(max01)
#print(mean(ba[1:anz]))
modmax <- models
#plot(unlist(modmax[[1]]))
#plot(unlist(modmax[[2]]))
#plot(unlist(modmax[[3]]))
baP <- c(ba[1:anz],mean(ba[1:anz]))
modelsP <- modmax
###############
##### 2nd stage: individual 
message("\n","2nd stage: individual optimization")
set.seed(7654321)
dataMend <- data[,-inddep]  ## all exogenous variables used
for (i in psize){
message("Number of predictors: ",i)
for (n in 1:M) {
  message("repetition ",n)
##  datas <- dataland[sample(rownames(dataland))[1:ssize],]  ## alternative: varying number of observations
  datas <- cbind(dataMend[,sample(colnames(dataMend))[1:psize]],data[,inddep])
  equal <- matrix(TRUE,ncol=anz,nrow=dim(data)[1])
  ba <- c(1:anz)*0
  for (j in 1:anz){
    perc <- percl[j]
    perc2 <- percs[j]
#    message(depnames[j])
#  outland <- PrInDT(dataland[,indall],"French",ctestv=NA,N,perc,perc2,conf.level=cf)
    outland <- PrInDT(datas,depnames[j],ctestv=ctestv,N=N,percl=perc,percs=perc2,conf.level=conf.level,seedl=FALSE,minsplit=minsplit,minbucket=minbucket)
#  print(outland)
#  plot(outland$tree1st)
    class.pred <- predict(outland$tree1st,newdata=data)
    class.pred <- relevel(class.pred,ref=levels(data[,inddep[j]])[1])
    conti <- table(class.pred,data[,inddep[j]])
    ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
    ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
    ba[j] <- (ba1 + ba2)/2
    models[[j]] <- outland$tree1st
    if (ba[j] > max01[j]) { 
      max01[j] <- ba[j]
#      print(max01)
      modmax[[j]] <- models[[j]]
    }
  }
##
# Joint assessment: 01
#  acc01 <- rowSums(equal)
#  ba[(anz+2)] <- sum(acc01) / (3*dim(data)[1])
#  acc01[acc01 < 3] <- 0
#  acc01[acc01 == 3] <- 1
#  ba[(anz+1)] <- sum(acc01)
#  ba[(anz+1)] <- ba[(anz+1)] / dim(data)[1]
}
}
#print(max01[1:anz])
#print(sum(max01[1:anz]) / 3)
#plot(unlist(modmax[[1]]))
#plot(unlist(modmax[[2]]))
#plot(unlist(modmax[[3]]))
baI <- c(max01[1:anz],mean(max01[1:anz]))
modelsI <- modmax
##########################
###########  mean maximization
message("\n","3rd stage: joint optimization")
set.seed(7654321)
dataMend <- data[,-inddep]  ## all exogenous variables used
bam <- 0
for (i in psize){
message("Number of predictors: ",i)
for (n in 1:M) {
  message("repetition ",n)
##  datas <- dataland[sample(rownames(dataland))[1:ssize],]  ## alternative: varying number of observations
  datas <- cbind(dataMend[,sample(colnames(dataMend))[1:psize]],data[,inddep])
  equal <- matrix(TRUE,ncol=anz,nrow=dim(data)[1])
  ba <- c(1:anz)*0
  for (j in 1:anz){
    perc <- percl[j]
    perc2 <- percs[j]
#    message(depnames[j])
#  outland <- PrInDT(dataland[,indall],"French",ctestv=NA,N,perc,perc2,conf.level=cf)
    outland <- PrInDT(datas,depnames[j],ctestv=ctestv,N=N,percl=perc,percs=perc2,conf.level=conf.level,seedl=FALSE,minsplit=minsplit,minbucket=minbucket)
#  print(outland)
#  plot(outland$tree1st)
    class.pred <- predict(outland$tree1st,newdata=data)
    class.pred <- relevel(class.pred,ref=levels(data[,inddep[j]])[1])
    conti <- table(class.pred,data[,inddep[j]])
    ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
    ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
    ba[j] <- (ba1 + ba2)/2
    models[[j]] <- outland$tree1st
  }
  if (mean(ba) > bam) { 
    bam <- mean(ba)
    max01[1:anz] <- ba[1:anz]
#    print(c(max01[1:anz],bam))
    for (j in 1:anz){
      modmax[[j]] <- models[[j]]
    }
  }
}
}
baJ <- c(max01[1:anz],bam)
#print(baJ)
baAll <- rbind(baP,baI,baJ)
result <- list(models1=modelsP,models2=modelsI,models3=modmax,classnames=depnames,baAll=baAll) 
class(result) <- "SimCPrInDT"
result
}
#' @export
print.SimCPrInDT <- function(x,...){
## output
cat("Accuracies of best models","\n")
colnames(x$baAll) <- c(x$classnames,"mean")
rownames(x$baAll) <- c("full sample","indiv. opt.","joint opt.")
print(x$baAll)
cat("\n\n")
for (i in 1:length(x$classnames)){
  cat("Best model 2nd stage: ",x$classnames[i], "\n")
  print(x$models2[[i]])
  cat("\n")
}
}
#' @export
plot.SimCPrInDT <- function(x,...){
for (i in 1:length(x$classnames)){
  plot(x$models2[[i]],main=x$classnames[i])
}
}

#' The basic undersampling loop for classification
#'
#' @description The function PrInDT uses ctrees (conditional inference trees from the package "party") for optimal modeling of
#' the relationship between the two-class factor variable 'classname' and all other factor and numerical variables
#' in the data frame 'datain' by means of 'N' repetitions of undersampling. The optimization citerion is the balanced accuracy 
#' on the validation sample 'valdat' (default = full input sample 'datain'). The trees generated from undersampling can be restricted by not accepting trees 
#' including split results specified in the character strings of the vector 'ctestv'.\cr
#' The undersampling percentages are 'percl' for the larger class and 'percs' for the smaller class (default = 1).\cr
#' A probability threshold 'thres' for the prediction of the smaller class may be specified (default = 0.5).\cr
#' Undersampling may be stratified in two ways by the option 'stratvers'.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr\cr
#' In the case of repeated measurements ('indrep=1'), the values of the substructure variable have to be given in 'repvar'. 
#' Only one value of 'classname' is allowed for each value of 'repvar'.
#' If for a value of 'repvar' the percentage 'thr' of the observed occurence of a value of 'classname' is not reached by the number of predictions of the value of 'classname', a misclassification is detected.
#'
#' @usage PrInDT(datain,classname,ctestv=NA,N,percl,percs=1,conf.level=0.95,thres=0.5,
#'          stratvers=0,strat=NA,seedl=TRUE,minsplit=NA,minbucket=NA,repvar=NA,indrep=0,
#'          valdat=datain,thr=0.5)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
#'     character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
#'     For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
#'     Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
#'     A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
#'     If no restrictions exist, the default = NA is used.
#' @param N Number (> 2) of repetitions (integer)
#' @param percl Undersampling percentage of larger class (numerical, > 0 and <= 1);\cr
#'     if percs = default, default of percl = percentage of large class resulting in the same number of observations as in the small class
#' @param percs Undersampling percentage of smaller class (numerical, > 0 and <= 1);\cr
#'     default = 1
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param thres Probability threshold for prediction of smaller class (numerical, >= 0 and < 1); default = 0.5
#' @param stratvers Version of stratification;\cr
#'     = 0: none (default),\cr
#'     = 1: stratification according to the percentages of the values of the factor variable 'strat',\cr
#'     > 1: stratification with minimum number "stratvers" of observations per value of "strat"
#' @param strat Name of one (!) stratification variable for undersampling (character);\cr
#'     default = NA (no stratification)
#' @param seedl Should the seed for random numbers be set (TRUE / FALSE)?\cr
#'     default = TRUE
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#' @param repvar Values of variable defining the substructure in the case of repeated measurements, length = dim(valdat)[1] necessary; default = NA
#' @param indrep Indicator of repeated measurements ('indrep=1'); default = 0
#' @param valdat Validation data; default = datain
#' @param thr threshold for element classification: minimum percentage of correct class entries; default = 0.5
#'
#' @return
#' \describe{
#' \item{tree1st}{best tree on validation sample}
#' \item{tree2nd}{2nd-best tree on validation sample}
#' \item{tree3rd}{3rd-best tree on validation sample}
#' \item{treet1st}{best tree on test sample}
#' \item{treet2nd}{2nd-best tree on test sample}
#' \item{treet3rd}{3rd-best tree on test sample}
#' \item{ba1st}{accuracies: largeClass, smallClass, balanced of 'tree1st', both for validation and test sample}
#' \item{ba2nd}{accuracies: largeClass, smallClass, balanced of 'tree2nd', both for validation and test sample}
#' \item{ba3rd}{accuracies: largeClass, smallClass, balanced of 'tree3rd', both for validation and test sample}
#' \item{baen}{accuracies: largeClass, smallClass, balanced of ensemble of all interpretable, 3 best acceptable, and all acceptable trees on validation sample}
#' \item{bafull}{vector of balanced accuracies of all trees from undersampling}
#' \item{batest}{vector of test accuracies of all trees from undersampling}
#' \item{treeAll}{tree based on all observations}
#' \item{baAll}{balanced accuracy of 'treeAll'}
#' \item{interpAll}{criterion of interpretability of 'treeall' (TRUE / FALSE)}
#' \item{confAll}{confusion matrix of 'treeAll'}
#' \item{acc1AE}{Accuracy of full sample tree on Elements of large class} 
#' \item{acc2AE}{Accuracy of full sample tree on Elements of small class }  
#' \item{bamaxAE}{balanced accuracy of full sample tree on Elements} 
#' \item{namA1}{Names of misclassified Elements by full sample tree of large class}
#' \item{namA2}{Names of misclassified Elements by full sample tree of small class}
#' \item{acc1E}{Accuracy of best tree on Elements of large class} 
#' \item{acc2E}{Accuracy of best tree on Elements of small class }  
#' \item{bamaxE}{balanced accuracy of best tree on Elements} 
#' \item{nam1}{Names of misclassified Elements by best tree of large class}
#' \item{nam2}{Names of misclassified Elements by best tree of small class}
#' \item{lablarge}{Label of large class} 
#' \item{labsmall}{Label of small class} 
#' \item{valdat}{Validation set after edit}
#' \item{thr}{Threshold for repeated measurements}
#' }
#'
#' @details
#' For the optimzation of the trees, we employ a method we call Sumping (Subsampling umbrella of 
#' model parameters), a variant of Bumping (Bootstrap umbrella of model parameters) (Tibshirani 
#' & Knight, 1999) which uses subsampling instead of bootstrapping. The aim of the 
#' optimization is to identify conditional inference trees with maximum predictive power
#' on the full sample under interpretability restrictions.
#'
#' \strong{References} \cr
#' -- Tibshirani, R., Knight, K. 1999. Model Search and Inference By Bootstrap "bumping".
#' Journal of Computational and Graphical Statistics, Vol. 8, No. 4 (Dec., 1999), pp. 671-686 \cr
#' -- Weihs, C., Buschfeld, S. 2021a. Combining Prediction and Interpretation in  Decision Trees (PrInDT) - 
#' a Linguistic Example. arXiv:2103.02336
#'
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
#' frame of the function.\cr
#' The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
#' \code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
#
#' @export PrInDT
#'
#' @examples
#' datastrat <- PrInDT::data_zero
#' data <- na.omit(datastrat) # cleaned full data: no NAs
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
#' N <- 41  # no. of repetitions
#' conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
#' percl <- 0.08  # undersampling percentage of the larger class
#' percs <- 0.95 # undersampling percentage of the smaller class
#' # calls of PrInDT
#' out <- PrInDT(data,"real",ctestv,N,percl,percs,conf.level) # unstratified
#' out # print best model and ensembles as well as all observations
#' plot(out)
#' out <- PrInDT(data,"real",ctestv,N,percl,percs,conf.level,stratvers=1,
#'               strat="SEX") # percentage stratification
#' out <- PrInDT(data,"real",ctestv,N,percl,percs,conf.level,stratvers=50,
#'               strat="SEX") # stratification with minimum no. of tokens
#' out <- PrInDT(data,"real",ctestv,N,percl,percs,conf.level,thres=0.4) # threshold = 0.4
#'
#' @importFrom stats relevel median predict
#' @importFrom party ctree ctree_control
#' @importFrom splitstackshape stratified
#' @importFrom graphics hist abline par
#'
PrInDT <- function(datain,classname,ctestv=NA,N,percl=NA,percs=1,conf.level=0.95,thres=0.5,stratvers=0,strat=NA,seedl=TRUE,
                               minsplit=NA,minbucket=NA,repvar=NA,indrep=0,valdat=datain,thr=0.5){
  ## input check
  if (typeof(datain) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) || N <= 0 ||
      !( (0 < percl & percl <= 1) | typeof(percl)=="logical") || !(0 < conf.level & conf.level <= 1) || !(0 <= thres & thres < 1) ||
      !(0 <= stratvers) || !(typeof(strat) %in% c("logical", "character")) || typeof(seedl) != "logical" || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double")) || !(typeof(repvar) %in% c("logical","integer","character")) || !(typeof(indrep) %in% c("integer","double")) ||
      typeof(valdat) != "list" || typeof(thr) != "double") {
    stop("irregular input")
  }
  if (!(all(names(datain) %in% names(valdat)))){
    stop("validation data variables unequal input data variables")
  }
  if (indrep == 1 & all(is.na(repvar)) == TRUE){
    message("\n","Note: repeated measurement analysis disabled: no repeated measurement variable specified","\n")
    indrep <- 0
  }   
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
  if (N <= 2){
      stop("Number of repetitions too low (must be > 2)")
    }
  if (seedl){
    set.seed(7654321)  # set seed of random numbers
  }
  ###
  #print(.Random.seed[1:5])  # possible check of random seed
  data <- datain
  repv <- repvar
  names(data)[names(data)==classname] <- "class"
  n_class1 <- table(data$class)[1] # no. of elements of larger class 1
  n_class2 <- table(data$class)[2] # no. of elements of smaller class 2
  if (min(n_class1,n_class2) == 0){
   stop("irregular input: only 1 class")
  }
  if (n_class1 < n_class2){
    # relevel of classes if smaller class first
    data$class <- stats::relevel(data$class, levels(data$class)[2]) # larger class now first
    n_class1 <- table(data$class)[1] # no. of elements of larger class 1
    n_class2 <- table(data$class)[2] # no. of elements of smaller class 2
  }
  if (is.na(percl) == TRUE){
    percl <- percs*n_class2/n_class1
  }
  n <- n_class1 + n_class2 # overall no. of observations in data
  ## reordering of classes: smaller class first in observations
#print(c(n_class1,n_class2))
 # if (n_class1 > n_class2){
   order_class <- order(as.numeric(data$class),decreasing=TRUE) # re-ordered line numbers
 # } else {
 #   order_class <- order(as.numeric(data$class))
 # }
  if (percl*n_class1 < 3 || percs*n_class2 < 3){
      stop("Number of observations in one of the class levels too low (< 3)")
  }
  data <- data[order_class,] # data now reordered: smaller class first
  ## validation set re-ordered
  if (identical(valdat, datain)){
    valdat <- data
    if (indrep == 1) {repv <- repv[order_class]}
  }
  names(valdat)[names(valdat)==classname] <- "class"
  valdat$class <- stats::relevel(valdat$class, levels(data$class)[1]) # larger class now first  ## NEWNEW
  if (!(identical(levels(data$class),levels(valdat$class)))){
    stop("levels of input class variable unequal levels of validation class variable")
  }
  n_class1v <- table(valdat$class)[1] # no. of elements of larger class 1
  n_class2v <- table(valdat$class)[2] # no. of elements of smaller class 2
  if (min(n_class1v,n_class2v) == 0){
   stop("irregular input: only 1 class in valdation set")
  }
  nv <- n_class1v + n_class2v
  if (indrep == 1 & length(repv) != dim(valdat)[1]){
    stop("irregular input: no. of observations different in 'repvar' and validation set")
  }
######
  ## model with all original observations
  ######
  ct <- party::ctree(class ~ ., data = data, control = party::ctree_control(mincriterion=conf.level,minsplit=minsplit,minbucket=minbucket))
  if (thres != 0.5){
    predsprob <- stats::predict(ct,type="prob",newdata=valdat)
    ctpreds <- as.factor(sapply( 1:dim(valdat)[1],
                                 function(x)
                                   ifelse(predsprob[[x]][2] > thres,
                                          levels(valdat$class)[2], levels(valdat$class)[1] )))
  } else {
    ctpreds <- stats::predict(ct,newdata=valdat) # predictions for all validation observations
  }
  if (dim(table(ctpreds))[1] > 1){
    confAll <- table(ctpreds, valdat$class)  # confusion matrix calculation
  # balanced accuracy
    crit2 <- (confAll[1,1] / n_class1v + confAll[2,2] / n_class2v)/2
  } else {
    confAll <- matrix(c(table(valdat$class),0,0),nrow=2,byrow=TRUE)
    colnames(confAll) <- names(table(valdat$class))
    rownames(confAll) <- colnames(confAll)
    crit2 <- 0.5
  }
  crit1All <- "FALSE"
  if (all(is.na(ctestv)) == FALSE) {
    crit1All <- FindSubstr(ct,ctestv) # call of FindSubstr for overall tree
  }
  treeAll <- ct
  baAll <- crit2
acc1AE <- 0
acc2AE <- 0
bamaxAE <- 0
namA1 <-  list()
namA2 <- list()
if (indrep == 1){
  pred <- predict(ct,newdata=valdat)  ## NEWNEW
#print(c(length(repv),length(pred)))
  ch <- table(repv,pred)
  conti <- cbind(table(repv,valdat$class),ch)  ## NEWNEW
#print(conti)
  no1 <- sum(conti[,1] > 0)
  no2 <- sum(conti[,2] > 0)
  if ((no1 + no2) > dim(conti)[1]){
#    cat("Note: repeated measurement variable has more than one class per value")
    acc2AE <- -1
    acc1AE <- -1
  } else {
    for (i in 1:dim(conti)[1]){
      if (conti[i,2] > 0 & conti[i,4] <= (conti[i,2]*thr)){ 
        namA2 <- c(namA2,rownames(ch)[i])
      }
      if (conti[i,1] > 0 & conti[i,3] < (conti[i,1]*thr)){ 
        namA1 <- c(namA1,rownames(ch)[i])
      }
    }
    acc2AE <- 1 - length(namA2) / no2
    acc1AE <- 1 - length(namA1) / no1  
    bamaxAE <- (acc1AE + acc2AE)/2
    if (length(namA1) == 0) {namA1 <- "-"}
    if (length(namA2) == 0) {namA2 <- "-"} 
  }
}
################################################
  ## resampling loop
  ######
  ## initialization
  chs <- matrix(0,nrow=N,ncol=8) # matrix of characteristics
  s_ind <- matrix(0,nrow=N,ncol=n) # matrix for resampled indices
  preds <- matrix(0,nrow=N,ncol=n) # matrix for predictions
  preds_ordered <- preds   # matrix for ordered predictions
  preds_orderedv <- matrix(0,nrow=N,ncol=nv)  # matrix for validation predictions
  crit1 <- rep("FALSE",N) # vector for crit1 (interpretability)
  crit2 <- rep(0,N)      # vector for crit2 (predictive power on validation sample)
  crit3 <- rep(0,N)      # vector for crit3 (predictive power on test sample)
  trees <- ct              # initialization of vector of trees (to be concatenated)
  if (is.na(strat) == FALSE){
    df1 <- data.frame( cbind( 1:n_class1,data[data$class == levels(data$class)[1],names(data)==strat] ) )
    colnames(df1) <- c("x","strat")
    df2 <- data.frame( cbind( 1:n_class2,data[data$class == levels(data$class)[2],names(data)==strat] ) )
    colnames(df2) <- c("x","strat")
    stratv <- as.factor(as.numeric(data[,strat]))
    nmin <- min(sum(df1[,2] == levels(stratv)[1]) + sum(df2[,2] == levels(stratv)[1]),
                sum(df1[,2] == levels(stratv)[2]) + sum(df2[,2] == levels(stratv)[2]))
    if (nmin < stratvers){
      stop("Minimum number of tokens of predictor value for stratification too big")
    }
  }
  ## column names of confusion matrix elements
  colnames(chs) <- c("train11","train21","train12","train22","test11","test21","test12","test22")
  if (is.na(strat) == FALSE & stratvers == 1){
#    message("\n")
    message("Percentage stratification")
    message("\n")
  }
  if (is.na(strat) == FALSE & stratvers > 1){
#    message("\n")
    message("Stratification with minimum number of tokens per predictor value = ",stratvers)
    message("\n")
  }
  ## start of resampling loop
  for (i in 1:N) {  # N repetitions
    # generation of sample indices for undersampling
    if (is.na(strat) == TRUE){
      sample_ind <- sample.int(n_class1,round(percl*n_class1)) + n_class2
      if (percs < 1){
        sample_ind <- c(sample.int(n_class2,round(percs*n_class2)),sample_ind)
      } else {
        sample_ind <- c(seq(1:n_class2),sample_ind)
      }
    }
    if (is.na(strat) == FALSE){
      if (stratvers == 1){
        sample_ind <- t(splitstackshape::stratified( df1,"strat",percl )[,1] + n_class2)
        if (percs < 1){
          suppressMessages(sample_ind <- c(t( splitstackshape::stratified( df2,"strat",percs )[,1] ), sample_ind))
        } else {
          suppressMessages(sample_ind <- c( seq(1:n_class2), sample_ind ))
        }
      } else {
        iter <- 0
        nmin <- 0
        while (nmin < stratvers & iter < 10){
          iter <- iter + 1
          sample_ind <- sample.int(n_class1,round(percl*n_class1)) + n_class2
          if (percs < 1){
            sample_ind <- c(sample.int(n_class2,round(percs*n_class2)),sample_ind)
          } else {
            sample_ind <- c(seq(1:n_class2),sample_ind)
          }
          nmin <- min(sum(data[sample_ind,strat] == levels(data[,strat])[1]),
                      sum(data[sample_ind,strat] == levels(data[,strat])[2]))
        }
        if (iter == 10){
          stop("Minimum number of tokens of predictor value specified for stratification not realizable")
        }
      }
    }
    ls <- length(sample_ind)
    s_ind[i,1:ls] <- sample_ind # chosen indices of larger class
    s_ind[i,(ls + 1):n] <- (seq(1:n))[-sample_ind]
    # generation of train and test samples
    data_train <- data[sample_ind,] # undersamples of both classes
    data_test <- data[-sample_ind,] # rest of sample
    ## modeling
    ct <- party::ctree(class ~ ., data = data_train,control = party::ctree_control(mincriterion=conf.level,minsplit=minsplit,minbucket=minbucket)) 
    trees <- c(trees,ct) # storing of new tree
    if (all(is.na(ctestv)) == FALSE) {
      crit1[i] <- FindSubstr(ct,ctestv) # deciding on interpretability
    }
    ## classification accuracy
    ## training data
    if (thres != 0.5){
      predsprob <- stats::predict(ct,type="prob")
      ctpreds <- as.factor(sapply( 1:dim(data_train)[1], function(x) ifelse( predsprob[[x]][2] > thres,
                                                                             levels(data$class)[2],levels(data$class)[1] )))
    } else {
      ctpreds <- stats::predict(ct)
    }
    if (dim(table(ctpreds))[1] > 1){
      conf_train <- table(ctpreds, data_train$class)
    }
    else {
      conf_train <- matrix(c(table(data_train$class),0,0),nrow=2,byrow=TRUE)
      colnames(conf_train) <- names(table(data$class))
      rownames(conf_train) <- colnames(conf_train)
    }
    preds[i,1:ls] <- ctpreds # training predictions
    chs[i,1:4] <- as.vector(conf_train) # confusion matrix from training
    ## test data 
    if (thres != 0.5){
      predsprob <- stats::predict(ct,newdata=data_test,type="prob")
      ctpreds <- factor(sapply( 1:dim(data_test)[1], function(x) ifelse( predsprob[[x]][2] > thres,
                                                                         levels(data$class)[2],levels(data$class)[1] )))
    } else {
      ctpreds <- stats::predict(ct,newdata=data_test)
    }
    if (dim(table(ctpreds))[1] > 1){
      conf_test <- table(ctpreds, data_test$class)
    }
    else {
      conf_test <- matrix(c(table(data_test$class),0,0),nrow=2,byrow=TRUE)
      colnames(conf_test) <- names(table(data$class))
      rownames(conf_test) <- colnames(conf_test)
    }
    preds[i,(ls+1):n] <- ctpreds # test predictions
    preds_ordered[i,] <- preds[i,order(s_ind[i,])] # correct order of predictions
    chs[i,5:8] <- as.vector(conf_test) # confusion matrix from testing
#   crit2[i] <- 0.5 * ( (chs[i,1] + chs[i,5]) / n_class1 + (chs[i,4] + chs[i,8]) / n_class2 ) # overall balanced accuracy
#if (i > 1 & crit2[i] > max(crit2[1:(i-1)])){
#print(c(i,crit2[i]))
#}
    if ( percs < 1 ){
      crit3[i] <- 0.5 * ( chs[i,5] / (chs[i,5] + chs[i,6]) + chs[i,8] / (chs[i,7] + chs[i,8]) ) # test balanced accuracy
    } else {
      crit3[i] <- chs[i,5] / (chs[i,5] + chs[i,6])  # test balanced accuracy
    }
    ###########################
    ## validation data 
    if (thres != 0.5){
      predsprob <- stats::predict(ct,newdata=valdat,type="prob")
      ctpreds <- factor(sapply( 1:dim(valdat)[1], function(x) ifelse( predsprob[[x]][2] > thres,
                                                                         levels(valdat$class)[2],levels(valdat$class)[1] )))
    } else {
      ctpreds <- stats::predict(ct,newdata=valdat)
    }
    if (dim(table(ctpreds))[1] > 1){
#print(length(ctpreds))
#print(str(valdat))
      conf_test <- table(ctpreds, valdat$class)
    }
    else {
      conf_test <- matrix(c(table(valdat$class),0,0),nrow=2,byrow=TRUE)
      colnames(conf_test) <- names(table(valdat$class))
      rownames(conf_test) <- colnames(conf_test)
    }
    ba1 <- conf_test[1,1] / (conf_test[1,1] + conf_test[2,1])
    ba2 <- conf_test[2,2] / (conf_test[1,2] + conf_test[2,2])
    crit2[i] <- (ba1 + ba2)/2
    preds_orderedv[i,] <- ctpreds ## for validation set
#    preds[i,(ls+1):n] <- ctpreds # test predictions
#    preds_ordered[i,] <- preds[i,order(s_ind[i,])] # correct order of predictions
#    chs[i,5:8] <- as.vector(conf_test) # confusion matrix from testing
#    crit2[i] <- 0.5 * ( (chs[i,1] + chs[i,5]) / n_class1 + (chs[i,4] + chs[i,8]) / n_class2 ) # overall balanced accuracy
  }
  #### end of resampling loop
  ####
  ## analysis of resampled trees
  ####
  trees <- trees[-1]  # analyze without the tree from all data
  ## 3 best balanced accurary trees on validation sample
  crit2o <- sort(crit2,decreasing=TRUE,index.return=TRUE) # sorting the balanced accuracies
  three <- crit2o$ix[crit1[crit2o$ix] == "FALSE"][1:3]    # indices of the 3 best interpretable trees: overall
  crit3o <- sort(crit3,decreasing=TRUE,index.return=TRUE) # sorting the balanced accuracies
  threet <- crit3o$ix[crit1[crit3o$ix] == "FALSE"][1:3]    # indices of the 3 best intrepretable trees: test
  ba1st <- c(crit2[three[1]],crit3[threet[1]])
  ba2nd <- c(crit2o$x[crit1[crit2o$ix] == "FALSE"][2],crit3[threet[2]])
  ba3rd <- c(crit2o$x[crit1[crit2o$ix] == "FALSE"][3],crit3[threet[3]])
  ## accuracies on validation sample
  ## best tree
  classE <- as.integer(valdat$class)
  preds_orderedE <- preds_orderedv[three[1],]
  preds1 <- preds_orderedE[classE == 1]
  preds2 <- preds_orderedE[classE == 2]
  accE1 <- sum (preds1 == 1) / n_class1
  accE2 <- sum (preds2 == 2) / n_class2
  ## accuracies on test sample
  classE <- as.integer(data$class[s_ind[threet[1],(ls + 1):n]])
  preds_orderedE <- preds[threet[1],(ls+1):n]
  preds1 <- preds_orderedE[classE == 1]
  preds2 <- preds_orderedE[classE == 2]
  accT1 <- sum (preds1 == 1) / (chs[threet[1],5] + chs[threet[1],6])
  if ( percs < 1 ){
    accT2 <- sum (preds2 == 2) / (chs[threet[1],7] + chs[threet[1],8])
  } else {
    accT2 <- NA
  }
  ba1st <- c(accE1,accE2,ba1st[1],accT1,accT2,ba1st[2])
  names(ba1st) <- c(names(ba1st)[1:2],"balanced",paste0("test:",names(ba1st)[1]),paste0("test:",names(ba1st)[2]),"balanced")
  ## 2nd best tree
  classE <- as.integer(valdat$class)
  preds_orderedE <- preds_orderedv[three[2],]
  preds1 <- preds_orderedE[classE == 1]
  preds2 <- preds_orderedE[classE == 2]
  accE1 <- sum (preds1 == 1) / n_class1
  accE2 <- sum (preds2 == 2) / n_class2
  ##
  classE <- as.integer(data$class[s_ind[threet[2],(ls + 1):n]])
  preds_orderedE <- preds[threet[2],(ls+1):n]
  preds1 <- preds_orderedE[classE == 1]
  preds2 <- preds_orderedE[classE == 2]
  accT1 <- sum (preds1 == 1) / (chs[threet[2],5] + chs[threet[2],6])
  if ( percs < 1 ){
    accT2 <- sum (preds2 == 2) / (chs[threet[2],7] + chs[threet[2],8])
  } else {
    accT2 <- NA
  }
  ba2nd <- c(accE1,accE2,ba2nd[1],accT1,accT2,ba2nd[2])
  names(ba2nd) <- names(ba1st)
  ## 3rd best tree
  classE <- as.integer(valdat$class)
  preds_orderedE <- preds_orderedv[three[3],]
  preds1 <- preds_orderedE[classE == 1]
  preds2 <- preds_orderedE[classE == 2]
  accE1 <- sum (preds1 == 1) / n_class1
  accE2 <- sum (preds2 == 2) / n_class2
  ##
  classE <- as.integer(data$class[s_ind[threet[3],(ls + 1):n]])
  preds_orderedE <- preds[threet[3],(ls+1):n]
  preds1 <- preds_orderedE[classE == 1]
  preds2 <- preds_orderedE[classE == 2]
  accT1 <- sum (preds1 == 1) / (chs[threet[3],5] + chs[threet[3],6])
  if ( percs < 1 ){
    accT2 <- sum (preds2 == 2) / (chs[threet[3],7] + chs[threet[3],8])
  } else {
    accT2 <- NA
  }
  ba3rd <- c(accE1,accE2,ba3rd[1],accT1,accT2,ba3rd[2])
  names(ba3rd) <- names(ba1st)
  ####
  ## ensembles: accuracies on validation and test sample
  ####
  baen <- matrix(0,nrow=3,ncol=5)
  ## all acceptable trees
  ## Mode function definition
  modus <- function(x){which.max(tabulate(x))} # definition of Mode function
  ##
  classE <- as.integer(valdat$class)
  no1 <- sum(classE == 1)
  no2 <- sum(classE == 2)
  ##
  treesE <- trees[crit1 == 'FALSE'] # trees fulfilling criterion 1
  preds_orderedEv <- preds_orderedv[crit1 == 'FALSE',]
  preds1 <- preds_orderedEv[,classE == 1]
  preds2 <- preds_orderedEv[,classE == 2]
  accE1 <- sum (apply(preds1,2,modus) == 1) / no1
  accE2 <- sum (apply(preds2,2,modus) == 2) / no2
  accE <- (accE1 + accE2) / 2
  accET <- mean(crit3[crit1 == 'FALSE'])
  baen[1,] <- c(length(treesE),accE1,accE2,accE,accET)  # save for print
  ## 3 best trees
 # classE <- as.integer(data$class)
  treesE <- trees[three]
  preds_orderedEv <- preds_orderedv[three,]
  preds1 <- preds_orderedEv[,classE == 1]
  preds2 <- preds_orderedEv[,classE == 2]
  accE1 <- sum (apply(preds1,2,modus) == 1) / no1
  accE2 <- sum (apply(preds2,2,modus) == 2) / no2
  accE <- (accE1 + accE2) / 2
  accET <- mean(crit3[threet])
  baen[2,] <- c(length(treesE),accE1,accE2,accE,accET)  # save for print
  ## trees with overall accuracies better than median
  cthres <- stats::median(crit2) # trees with overall balanced accuracy greater or equal median
  cthrest <- stats::median(crit3) # trees with test balanced accuracy greater of equal median
 # classE <- as.integer(data$class)
  treesE <- trees[crit1 == 'FALSE' & crit2 >= cthres]
  treesET <- trees[crit1 == 'FALSE' & crit3 >= cthrest]
  ##treesEU <- unique(treesE) # (restiction to) unique trees
  ##length(treesEU) # no. of unique trees
  preds_orderedEv <- preds_orderedv[crit1 == 'FALSE' & crit2 >= cthres,]
  preds1 <- preds_orderedEv[,classE == 1]
  preds2 <- preds_orderedEv[,classE == 2]
  accE1 <- sum (apply(preds1,2,modus) == 1) / no1
  accE2 <- sum (apply(preds2,2,modus) == 2) / no2
  accE <- (accE1 + accE2) / 2
  accET <- mean(crit3[crit1 == 'FALSE' & crit3 >= cthrest])
  baen[3,] <- c(length(treesE),accE1,accE2,accE,accET)  # save for print
lablarge <- names(table(data$class))[1] # class 1 = large
labsmall <- names(table(data$class))[2] # class 2 = small 
  ####
acc1E <- 0
acc2E <- 0
bamaxE <- 0
nam1 <-  list()
nam2 <- list()
###
if (indrep == 1){ 
  pred <- predict(trees[[three[1]]],newdata=valdat)  ## NEWNEW
#print(table(pred))
  ch <- table(repv,pred)
#print(dim(ch))
  conti <- cbind(table(repv,valdat$class),ch)  ## NEWNEW
#print(conti)
  no1 <- sum(conti[,1] > 0)
  no2 <- sum(conti[,2] > 0)
  if ((no1 + no2) > dim(conti)[1]){
#    cat("Note: repeated measurement variable has more than one class per value")
    acc2E <- -1
    acc1E <- -1 
  } else {
    for (i in 1:dim(conti)[1]){
      if (conti[i,2] > 0 & conti[i,4] <= (conti[i,2]*thr)){ 
        nam2 <- c(nam2,rownames(ch)[i])
      }
      if (conti[i,1] > 0 & conti[i,3] < (conti[i,1]*thr)){ 
        nam1 <- c(nam1,rownames(ch)[i])
      }
    }
    acc2E <- 1 - length(nam2) / no2
    acc1E <- 1 - length(nam1) / no1  
    bamaxE <- (acc1E + acc2E)/2
    if (length(nam1) == 0) {nam1 <- "-"}
    if (length(nam2) == 0) {nam2 <- "-"}
  }
}
  result <- list(tree1st = trees[[three[1]]], treet1st = trees[[threet[1]]], ba1st = ba1st, tree2nd = trees[[three[2]]], treet2nd = trees[[threet[2]]],
                 ba2nd = ba2nd, tree3rd = trees[[three[3]]], treet3rd = trees[[threet[3]]], ba3rd = ba3rd, baen = baen, bafull= crit2, batest=crit3, 
                 treeAll=treeAll, baAll=baAll, interpAll=crit1All, confAll = confAll,lablarge=lablarge,labsmall=labsmall,
                 acc1AE=acc1AE,acc2AE=acc2AE,bamaxAE=bamaxAE,namA1=namA1,namA2=namA2,
                 acc1E=acc1E,acc2E=acc2E,bamaxE=bamaxE,nam1=nam1,nam2=nam2,indrep=indrep,valdat=valdat,thr=thr)
  class(result) <- c("PrInDT", "PrInDTAll")
#  gc(full=TRUE)
  result
}
#' @export
print.PrInDTAll <- function(x, ...){
  # output for model on all observations
  cat("******************************","\n")
  cat("Tree built on all observations","\n")
  cat("******************************","\n")
  print(x$treeAll) # print of tree
  cat("\n")
  cat("****************","\n")
  cat("Confusion matrix","\n")
  cat("****************","\n")
  print(x$confAll)
  cat("***********************************************","\n")
  cat("Criteria: interpretable    balanced accuracy","\n")
  cat("***********************************************","\n")
  cat("             ",unname(!as.logical(x$interpAll)),"              ",unname(x$baAll),"\n")
if (x$indrep == 1){
  cat("\n","Repeated measurements: Threshold = ",x$thr)
  if (x$acc1AE == -1){
    cat("\n","no analysis: repeated measurement variable has more than one class per value","\n")
  } else {  
    accs <- c(round(x$acc1AE,digits=6),round(x$acc2AE,digits=6),round(x$bamaxAE,digits=6))
    names(accs) <- c(colnames(x$confAll)[1:2],"balanced")
    cat("\n","Full sample accuracies of best tree (Elements)","\n")
    print(accs,row.names=FALSE)
 # cat("\n","Full sample accuracies of best tree (Elements)","\n"," ",colnames(x$confAll)[1]," \t",colnames(x$confAll)[2],"\t","balanced","\n")
 # cat(" ",c(round(x$acc1AE,digits=6),"\t",round(x$acc2AE,digits=6),"\t",round(x$bamaxAE,digits=6)),"\n\n")
    cat("\n","Wrongly predicted Elements:","\n")
    cat(x$lablarge,"\n")
    cat(unlist(x$namA1),"\n")
    cat(x$labsmall,"\n")
    cat(unlist(x$namA2),"\n")
  }
}
  invisible(x)
}
#' @export
print.PrInDT <- function(x, ...){
  NextMethod()
  cat("\n")
  cat("********************************************","\n")
  cat("Best acceptable tree from undersampling according to full sample accuracies","\n")
  cat("********************************************","\n")
  ## best tree: accuracies on full sample
  print(x$tree1st) # structure of tree with best overall accuracy
  accs <- c(round(x$ba1st[1],digits=6),round(x$ba1st[2],digits=6),round(x$ba1st[3],digits=6))
  names(accs) <- c(names(x$ba1st)[1:2],"balanced")
  cat("\n","Validation sample accuracies","\n")
  print(accs,row.names=FALSE)
#  cat("Validation sample accuracies: ",names(x$ba1st)[1],"\t","  ",names(x$ba1st)[2],"\t","  balanced","\n")
#  cat("                              ",unname(x$ba1st)[1],"\t","  ",unname(x$ba1st)[2],"\t","  ",unname(x$ba1st)[3],"\n") 
  cat("\n")
if (x$indrep == 1){
  cat("\n","Repeated measurements: Threshold = ",x$thr)
  if (x$acc1AE == -1){
    cat("\n","no analysis: repeated measurement variable has more than one class per value","\n")
  } else {  
    accs <- c(round(x$acc1E,digits=6),round(x$acc2E,digits=6),round(x$bamaxE,digits=6))
    names(accs) <- c(colnames(x$confAll)[1:2],"balanced")
    cat("\n","Input sample accuracies of best tree (Elements)","\n")
    print(accs,row.names=FALSE)
    cat("\n","Wrongly predicted Elements:","\n")
    cat(x$lablarge,"\n")
    cat(unlist(x$nam1),"\n")
    cat(x$labsmall,"\n")
    cat(unlist(x$nam2),"\n")
  }
}
  cat("\n","*************************************","\n")
  cat("Best acceptable tree from undersampling according to test accuracies","\n")
  cat("*************************************","\n")
  print(x$treet1st) # structure of tree with best test accuracy
  accs <- c(round(x$ba1st[4],digits=6),round(x$ba1st[5],digits=6),round(x$ba1st[6],digits=6))
  names(accs) <- c(names(x$ba1st)[4:5],"balanced")
  cat("\n","Test sample accuracies","\n")
  print(accs,row.names=FALSE)
#  cat("Test sample accuracies: ",names(x$ba1st)[4],"  ",names(x$ba1st)[5],"   balanced","\n")
#  cat("                        ",unname(x$ba1st)[4],"      ",unname(x$ba1st)[5],"         ",unname(x$ba1st)[6],"\n") # balanced accuracy of best tree printed
  cat("\n")
  cat("\n")
  cat("************************************************************************************************","\n")
  cat("Ensemble of all acceptable trees","\n")
  accs <- c(x$baen[1,1:5])
  names(accs) <- c("no. of trees",names(x$ba1st)[1:2],"balanced","  mean test balanced")
  print(accs,row.names=FALSE)
  #,names(x$ba1st)[1],"  ",names(x$ba1st)[2],"    balanced (mean test balanced)","\n"
  #cat("     ",unname(x$baen)[1,1],"                              ",unname(x$baen)[1,2],unname(x$baen)[1,3],unname(x$baen)[1,4],"   (",unname(x$baen)[1,5],")\n") # print no. of trees and accuracies
  cat("\n")
  cat("************************************************************************************************","\n")
  cat("Ensemble of 3 best acceptable trees","\n")
  accs <- c(x$baen[2,1:5])
  names(accs) <- c("no. of trees",names(x$ba1st)[1:2],"balanced","  mean test balanced")
  print(accs,row.names=FALSE)
 # cat("************************************************************************************************","\n")
 # cat("Ensemble of acceptable 3 best trees on full or test sample","\n"," no. of trees; full sample accuracies:",names(x$ba1st)[1],"  ",names(x$ba1st)[2],"   balanced (mean test balanced)","\n")
 # cat("******************************************************************************************","\n")
 # cat("     ",unname(x$baen)[2,1],"                               ",unname(x$baen)[2,2],unname(x$baen)[2,3],unname(x$baen)[2,4],"  (",unname(x$baen)[2,5],")\n") # print no. of trees and accuracies
  cat("\n")
  cat("************************************************************************************************","\n")
  cat("Ensemble of acceptable trees with full or test sample accuracies >= median ","\n")
  accs <- c(x$baen[3,1:5])
  names(accs) <- c("no. of trees",names(x$ba1st)[1:2],"balanced","  mean test balanced")
  print(accs,row.names=FALSE)
#  cat("*******************************************************************************************************","\n")
#  cat("Ensemble of acceptable trees with full or test sample accuracies >= median","\n"," no. of trees: full (test); full sample accuracies:",names(x$ba1st)[1],"  ",names(x$ba1st)[2],"   balanced (mean test #balanced)","\n")
#  cat("*******************************************************************************************************","\n")
#  cat("     ",unname(x$baen)[3,1],"                                          ",unname(x$baen)[3,2],unname(x$baen)[3,3],unname(x$baen)[3,4],"  (",unname(x$baen)[3,5],")\n") # print no. of trees and accuracies
  invisible(x)
}
#' @export
plot.PrInDTAll <- function(x, ...){
  plot(x$treeAll,main="Tree based on all observations")
  invisible(x)
}
#' @export
plot.PrInDT <- function(x, ...){
#  NextMethod()
  ## plot histograms of balanced accuracies
#  par(mfrow=c(1,2))
  hist(x$bafull,main=" ",xlab="Full sample balanced accuracies",cex.axis=1.5,cex.lab=1.5)
  abline(v = stats::median(x$bafull),col='red',lwd = 3)
  hist(x$batest,main=" ",xlab="Test sample balanced accuracies",cex.axis=1.5,cex.lab=1.5)
  abline(v = stats::median(x$batest),col='red',lwd = 3)
## Interpretable tree with highest full sample accuracy
  plot(x$tree1st,main="Interpretable tree from undersampling with highest full sample accuracy")
## Interpretable tree with highest test sample accuracy
  plot(x$treet1st,main="Interpretable tree from undersampling with highest test sample accuracy") 
  invisible(x)
}
#' @export
predict.PrInDT <- function(object,...){
  predict(object$tree1st,...)
}
#' @export
predict.PrInDTAll <- function(object,...){
  predict(object$treeAll,...)
}

#' @noRd
#'
PrInDTstruc1 <- function(data,class,ctestv=NA,name,check,labs,Ni,N=NA,Eit,Pit,undersamp=TRUE,co,crit="ba",ktest=0,stest=integer(length=0),p1,p2,minsplit=20,minbucket=7,repvar,valdat,indrep=0,thr=0.5){
lMit <- length(Eit)
lDit <- length(Pit)
if (Ni > 1){
  N <- Ni
}
if (is.na(N) == TRUE){
  N <- 99
}
lablarge <- names(table(data$class))[1] # class 1 = large
labsmall <- names(table(data$class))[2] # class 2 = small 
nam1max <- list()
nam2max <- list()
bamax <- 0
tamax <- 0
D <- dim(data)[2] - 1
kumin <-  length(stest)
summin <- length(stest)
mindlong <- list()
interp <- 0
# print(c(Eit,N,Ni))
for (M in Eit){
  message("Number of substructure elements per category: ",M)
  set.seed(7654321)
  for (i in 1:N){
#          ind1 <- sample(unique(name[check %in% labs[1,]]),prob=p1)[1:M] 
#          ind2 <- sample(unique(name[check %in% labs[2,]]),prob=p2)[1:M]
    ind1 <- sample(names(p1),prob=p1)[1:M] 
    ind2 <- sample(names(p2),prob=p2)[1:M]
    gTrain <- data[(name %in% ind1 | name %in% ind2),]
    gTest <- data[!(name %in% ind1 | name %in% ind2),]
    gname <- name[(name %in% ind1 | name %in% ind2)]
    gnameT <- name[!(name %in% ind1 | name %in% ind2)]
    gcheck <- check[(name %in% ind1 | name %in% ind2)]
    if (undersamp == TRUE & crit != "tal"){  ## for "tal" no effect of undersamp! NEWNEW
      gTrainr <- gTrain[gTrain$class==lablarge,]
      gnamer <- gname[gTrain$class==lablarge]
      gcheckr <- gcheck[gTrain$class==lablarge]
      gTrainz <- gTrain[gTrain$class==labsmall,]
      gnamez <- gname[gTrain$class==labsmall]
      gcheckz <- gcheck[gTrain$class==labsmall]
      samp <- sample(1:nrow(gTrainr))
      zTrain <- gTrainr[samp,]
      zname <- gnamer[samp]
      zcheck <- gcheckr[samp]
      gTest <- data[!(name %in% ind1 | name %in% ind2),]
      zTTrain <- rbind(zTrain[1:dim(gTrainz)[1],],gTrainz) # check if dim(gTrainz)[1] > 0 !!!!
      zTname <- c(as.character(zname[1:dim(gTrainz)[1]]),as.character(gnamez))
      nt <- nrow(zTTrain)
      zTcheck <- c(as.character(zcheck[1:dim(gTrainz)[1]]),as.character(gcheckz))
      gTest <- rbind(gTest,zTrain[(dim(gTrainz)[1]+1):(dim(gTrainr)[1]),])
    } else {
        zTTrain <- gTrain
        nt <- nrow(zTTrain)
        zTname <- gname
        zTcheck <- gcheck
    }
    if (Ni == 0){  
      datat <- zTTrain
      outAll <- PrInDTAll(datat,class,ctestv=ctestv,conf.level=co,minsplit=minsplit,minbucket=minbucket)
      if (outAll$interpAll == FALSE){
        interp <- interp + 1
      }
      indmax <- colnames(datat)
      if (crit == "ba"){
        class.pred <- predict(outAll$treeAll,newdata=valdat)
        class.pred <- relevel(class.pred,ref=levels(valdat$class)[1])
      # sum(class.pred == data$class) / length(class.pred)
        nam1 <- list()
        nam2 <- list()
        if (indrep == 2){
        #  acc1E <- 0
        #  acc2E <- 0
        #  pred <- predict(out$modbest,newdata=data)
          ch <- table(repvar,class.pred)
        # class <- datain[,names(datain)==classname]
          conti <- cbind(table(repvar,valdat$class),ch)
          no1 <- sum(conti[,1] > 0)
          no2 <- sum(conti[,2] > 0)
          for (i in 1:dim(conti)[1]){
            if (conti[i,2] > 0 & conti[i,4] <= (conti[i,2]*thr)){ 
              nam2 <- c(nam2,rownames(ch)[i])
            }
            if (conti[i,1] > 0 & conti[i,3] < (conti[i,1]*thr)){ 
              nam1 <- c(nam1,rownames(ch)[i])
            }
          }
          ba2 <- 1- length(nam2) / no2
          ba1 <- 1 - length(nam1) / no1  
          ba <- (ba1 + ba2)/2
          if (length(nam1) == 0) {nam1 <- "-"}
          if (length(nam2) == 0) {nam2 <- "-"}
        } else {
          conti <- table(class.pred,valdat$class)
          if (dim(conti)[1] == 1){                                                ## NEWNEW
            conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
          }
          ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
          ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
          ba <- (ba1 + ba2)/2
        }
        if (ba > bamax & outAll$interpAll == FALSE){
          bamax <- ba
          acc1 <- ba1
          acc2 <- ba2
          ntmax <- nt
          dmax <- D
          indmax <- colnames(datat)[1:D]
          ind1max <- ind1
          ind2max <- ind2
          nam1max <- nam1
          nam2max <- nam2
          outmax <- outAll
          gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
          colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
          gmaxTest <- gTest
        }
      }
      if (crit == "bat"){
        class.pred <- predict(outAll$treeAll,newdata=gTest)
        class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
      # sum(class.pred == gTest$class) / length(class.pred)
        conti <- table(class.pred,gTest$class)
        if (dim(conti)[1] == 1){                                                ## NEWNEW
          conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
        }
        ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
        ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
        ba <- (ba1 + ba2)/2
        if (ba > bamax & outAll$interpAll == FALSE){
          bamax <- ba
          acc1 <- ba1
          acc2 <- ba2
          ntmax <- nt 
          dmax <- D
          indmax <- colnames(datat)[1:D]
          ind1max <- ind1
          ind2max <- ind2
          outmax <- outAll
          gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
          colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
          gmaxTest <- gTest
        }
      }
      if (crit == "ta"){
        class.pred <- predict(outAll$treeAll,newdata=gTest)
        class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
        conti <- table(class.pred,gTest$class)
        if (dim(conti)[1] == 1){                                                ## NEWNEW
          conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
        }
        ta1 <- conti[1,1] / (conti[1,1] + conti[2,1])
        ta2 <- conti[2,2] / (conti[1,2] + conti[2,2])
        ta <- sum(class.pred == gTest$class) / length(class.pred)
        bat <- (ta1 + ta2)/2
        if (ta > tamax & outAll$interpAll == FALSE){
          bamax <- bat
          tamax <- ta
          acc1 <- ta1
          acc2 <- ta2
          ntmax <- nt 
          dmax <- D
          indmax <- colnames(datat)[1:D]
          ind1max <- ind1
          ind2max <- ind2
          outmax <- outAll
          gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
          colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
          gmaxTest <- gTest
        }
      }
      if (crit == "tal"){
        sTest <- stest[!(name %in% ind1 | name %in% ind2)]  ### NEWNEW
        splitTest <- table(gnameT)  
        for (i in 2:length(splitTest)){
          splitTest[i] <- splitTest[i] + splitTest[i-1]  #### splits according to gTest
        }
        class.pred <- predict(outAll$treeAll,newdata=gTest)
        class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
        conti <- table(class.pred,gTest$class)
        if (dim(conti)[1] == 1){                                                ## NEWNEW
          conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
        }
        ta1 <- conti[1,1] / (conti[1,1] + conti[2,1])
        ta2 <- conti[2,2] / (conti[1,2] + conti[2,2])
        bat <- (ta1 + ta2)/2
        ta <- sum(class.pred == gTest$class) / length(class.pred)
        predind <- class.pred == gTest$class
        indfal <- which(predind == FALSE)
        indfal <- indfal[!(indfal %in% splitTest)]
        indsplit <- split(indfal, cumsum(c(1, diff(indfal) != 1)))
        j <- 0
        indlong <- list()
        indstart <- vector()
        indlen <- vector()
        sumlen <- 0
        for (l in 1:length(indsplit)){
          if (length(unlist(indsplit[as.character(l)])) > ktest){ 
            j <- j + 1
            indlong[j] <- indsplit[as.character(l)]
            indstart[j] <- unlist(indlong[j])[1]
            indlen[j] <- length(unlist(indlong[j]))
            sumlen <- sumlen + indlen[j]
          }
        }
        kstart <- length(indstart)
        kustart <- length(unique(sTest[indstart]))
        if ((kustart < kumin & outAll$interpAll == FALSE) |
            (kustart == kumin & ta > tamax & outAll$interpAll == FALSE) |
            (kustart == kumin & ta == tamax & sumlen < summin & outAll$interpAll == FALSE) ){
          kumin <- kustart
          mindstart <- indstart  ## NEWNEW
          mindlong <- indlong
          elems <- unique(sTest[mindstart])
          summin <- sumlen
#print(c(i,kumin))
          tamax <- ta
          bamax <- bat
          acc1 <- ta1
          acc2 <- ta2
          ntmax <- nt 
          dmax <- D
          indmax <- colnames(datat)[1:D]
          ind1max <- ind1
          ind2max <- ind2
          outmax <- outAll
          gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
          colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
          gmaxTest <- gTest
        } 
      } 
    } else {
      for (d in Pit){
        for (n in 1:Ni){
#           datat <- zTTrain[,-1]
          datat <- zTTrain
          datat$class <- NULL
          ind <- sample(1:D)[1:d]
          datat <- datat[,ind]
#            datat <- cbind(datat,zTTrain[,1])
          datat <- cbind(datat,zTTrain$class)
          colnames(datat)[d+1] <- "class"
#print(colnames(datat))
          outAll <- PrInDTAll(datat,class,ctestv=ctestv,conf.level=co,minsplit=minsplit,minbucket=minbucket)
          if (outAll$interpAll == FALSE){
            interp <- interp + 1
          }
          if (crit == "ba"){
            class.pred <- predict(outAll$treeAll,newdata=valdat)
            class.pred <- relevel(class.pred,ref=levels(valdat$class)[1])
      # sum(class.pred == data$class) / length(class.pred)
            nam1 <- list()
            nam2 <- list()
            if (indrep == 2){
            #  acc1E <- 0
            #  acc2E <- 0
            #  pred <- predict(out$modbest,newdata=data)
              ch <- table(repvar,class.pred)
            # class <- datain[,names(datain)==classname]
              conti <- cbind(table(repvar,valdat$class),ch)
              no1 <- sum(conti[,1] > 0)
              no2 <- sum(conti[,2] > 0)
              for (i in 1:dim(conti)[1]){
                if (conti[i,2] > 0 & conti[i,4] <= (conti[i,2]*thr)){ 
                  nam2 <- c(nam2,rownames(ch)[i])
                }
                if (conti[i,1] > 0 & conti[i,3] < (conti[i,1]*thr)){ 
                  nam1 <- c(nam1,rownames(ch)[i])
                }
              }
              ba2 <- 1 - length(nam2) / no2
              ba1 <- 1 - length(nam1) / no1  
              ba <- (ba1 + ba2)/2
              if (length(nam1) == 0) {nam1 <- "-"}
              if (length(nam2) == 0) {nam2 <- "-"}
            } else {
              conti <- table(class.pred,valdat$class)
              if (dim(conti)[1] == 1){                                                ## NEWNEW
                conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
              }
              ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
              ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
              ba <- (ba1 + ba2)/2
            }
            if (ba > bamax & outAll$interpAll == FALSE){
              bamax <- ba
#print(c(i,d,bamax))
              acc1 <- ba1
              acc2 <- ba2
              ntmax <- nt
              dmax <- d
              indmax <- colnames(datat)[1:d]
              ind1max <- ind1
              ind2max <- ind2
              nam1max <- nam1
              nam2max <- nam2
              outmax <- outAll
              gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
              colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
              gmaxTest <- gTest
            }
          }
          if (crit == "bat"){
            class.pred <- predict(outAll$treeAll,newdata=gTest)
            class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
      # sum(class.pred == gTest$class) / length(class.pred)
            conti <- table(class.pred,gTest$class)
            if (dim(conti)[1] == 1){                                                ## NEWNEW
              conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
            }
            ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
            ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
            ba <- (ba1 + ba2)/2
            if (ba > bamax & outAll$interpAll == FALSE){
              bamax <- ba
              acc1 <- ba1
              acc2 <- ba2
              ntmax <- nt 
              dmax <- d
              indmax <- colnames(datat)[1:d]
              ind1max <- ind1
              ind2max <- ind2
              outmax <- outAll
              gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
              colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
              gmaxTest <- gTest
            }
          }
          if (crit == "ta"){
            class.pred <- predict(outAll$treeAll,newdata=gTest)
            class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
            conti <- table(class.pred,gTest$class)
            if (dim(conti)[1] == 1){                                                ## NEWNEW
              conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
            }
            ta1 <- conti[1,1] / (conti[1,1] + conti[2,1])
            ta2 <- conti[2,2] / (conti[1,2] + conti[2,2])
            ta <- sum(class.pred == gTest$class) / length(class.pred)
            bat <- (ta1 + ta2)/2
            if (ta > tamax & outAll$interpAll == FALSE){
              bamax <- bat
              tamax <- ta
              acc1 <- ta1
              acc2 <- ta2
              ntmax <- nt 
              dmax <- d
              indmax <- colnames(datat)[1:d]
              ind1max <- ind1
              ind2max <- ind2
              outmax <- outAll
              gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
              colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
              gmaxTest <- gTest
            }
          }
          if (crit == "tal"){
            sTest <- stest[!(name %in% ind1 | name %in% ind2)]  ### NEWNEW
            splitTest <- table(gnameT)  
            for (i in 2:length(splitTest)){
              splitTest[i] <- splitTest[i] + splitTest[i-1]  #### splits according to gTest
            }
            class.pred <- predict(outAll$treeAll,newdata=gTest)
            class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
            conti <- table(class.pred,gTest$class)
            if (dim(conti)[1] == 1){                                                ## NEWNEW
              conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
            }
            ta1 <- conti[1,1] / (conti[1,1] + conti[2,1])
            ta2 <- conti[2,2] / (conti[1,2] + conti[2,2])
            bat <- (ta1 + ta2)/2
            ta <- sum(class.pred == gTest$class) / length(class.pred)
            predind <- class.pred == gTest$class
            indfal <- which(predind == FALSE)
            indfal <- indfal[!(indfal %in% splitTest)]
            indsplit <- split(indfal, cumsum(c(1, diff(indfal) != 1)))
            j <- 0
            indlong <- list()
            indstart <- vector()
            indlen <- vector()
            sumlen <- 0
            for (l in 1:length(indsplit)){
              if (length(unlist(indsplit[as.character(l)])) > ktest){ 
                j <- j + 1
                indlong[j] <- indsplit[as.character(l)]
                indstart[j] <- unlist(indlong[j])[1]
                indlen[j] <- length(unlist(indlong[j]))
                sumlen <- sumlen + indlen[j]
              }
            }
            kstart <- length(indstart)
            kustart <- length(unique(sTest[indstart]))
            if ((kustart < kumin & outAll$interpAll == FALSE) |
              (kustart == kumin & ta > tamax & outAll$interpAll == FALSE) |
              (kustart == kumin & ta == tamax & sumlen < summin & outAll$interpAll == FALSE) ){
              kumin <- kustart
              mindstart <- indstart  ## NEWNEW
              mindlong <- indlong
#print(c(kumin,mindstart))
#print(unique(sTest[mindstart]))
              elems <- unique(sTest[mindstart])
              summin <- sumlen
              tamax <- ta
              bamax <- bat
              acc1 <- ta1
              acc2 <- ta2
              ntmax <- nt 
              dmax <- d
              indmax <- colnames(datat)[1:d]
              ind1max <- ind1
              ind2max <- ind2
              outmax <- outAll
              gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
              colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
              gmaxTest <- gTest
            }
          }
        }
      }
    } 
  }  
  if (crit == "ba"){
    message("current optimal balanced accuracy: ",round(bamax,digits=4),"\n")
    tamax <- 0
    kumin <- 0
    elems <- 0
  }
  if (crit == "bat"){
    message("current optimal balanced test accuracy: ",round(bamax,digits=4),"\n")
    tamax <- 0
    kumin <- 0
    elems <- 0
  }
  if (crit == "ta"){
    message("current optimal test accuracy: ",round(tamax,digits=4),"\n")
    elems <- 0
    kumin <- 0
  }
  if (crit == "tal"){
    message("current optimal number of long misclassified parts: ",kumin)
    message("current optimal test accuracy: ",round(tamax,digits=4),"\n")
  }
}
if (Ni > 0){
  interp <- c(interp,(lMit*N*Ni*length(Pit)))
} else {
  interp <- c(interp,(lMit*N))
}
##
  result <- list(interp=interp, dmax=dmax, ntmax = ntmax, acc1 = acc1, acc2 = acc2, nam1=nam1max, nam2=nam2max,
     bamax=bamax,tamax=tamax,kumin=kumin,elems=elems,mindlong=mindlong,
     ind1max = ind1max, ind2max = ind2max, indmax=indmax, modbest = outmax$treeAll, gmaxTrain = 
     gmaxTrain, gmaxTest = gmaxTest)
}

#' @noRd
#'
PrInDTstruc2 <- function(data,class,ctestv=NA,name,check,labs,N,Ni=NA,Pit,Eit,undersamp=TRUE,co,crit="ba",ktest=0,stest=integer(length=0),p1,p2,minsplit=20,minbucket=7,repvar,valdat,indrep=0,thr=0.5){
lMit <- length(Eit)
lDit <- length(Pit)
if (N > 0){
  Ni <- N
}
if (is.na(Ni) == TRUE){
  Ni <- 99
}
lablarge <- names(table(data$class))[1]  # class 1 = large
labsmall <- names(table(data$class))[2]  # class 2 = small
nam1max <- list()
nam2max <- list()
bamax <- 0
tamax <- 0
D <- dim(data)[2] - 1
kumin <- length(stest)
summin <- length(stest)
mindlong <- list()
interp <- 0
############
for (d in Pit){
  set.seed(7654321)  ## for single equations
message("Number of predictors: ",d)
  for (n in 1:Ni){
#  set.seed(321*n + n) # 17:4:0.99: 0.7097482, 17:4:0.95: 0.709784  # changed for what reason?
# set.seed(54321*n + n) # not used anymore
# print(n)
    datat <- data
    datat$class <- NULL
    ind <- sample(1:D)[1:d]
    datat <- datat[,ind]
    datat <- cbind(datat,data$class)
    colnames(datat)[d+1] <- "class"
#print(n)
    if (N > 0){
      for (M in Eit){
#      print(M)
#      set.seed(7654321)
        for (i in 1:N){
#          ind1 <- sample(unique(name[check %in% labs[1,]]),prob=p1)[1:M] 
#          ind2 <- sample(unique(name[check %in% labs[2,]]),prob=p2)[1:M]
          ind1 <- sample(names(p1),prob=p1)[1:M] 
          ind2 <- sample(names(p2),prob=p2)[1:M]
          gTrain <- datat[(name %in% ind1 | name %in% ind2),]
          gTest <- datat[!(name %in% ind1 | name %in% ind2),]
          sTest <- stest[!(name %in% ind1 | name %in% ind2)]
          gname <- name[(name %in% ind1 | name %in% ind2)]
          gnameT <- name[!(name %in% ind1 | name %in% ind2)]
          gcheck <- check[(name %in% ind1 | name %in% ind2)]
          if (undersamp == TRUE & crit != "tal"){
            gTrainr <- gTrain[gTrain$class==lablarge,]
            gnamer <- gname[gTrain$class==lablarge]
            gcheckr <- gcheck[gTrain$class==lablarge]
            gTrainz <- gTrain[gTrain$class==labsmall,]
            gnamez <- gname[gTrain$class==labsmall]
            gcheckz <- gcheck[gTrain$class==labsmall]
            samp <- sample(1:nrow(gTrainr))
            zTrain <- gTrainr[samp,]
            zname <- gnamer[samp]
            zcheck <- gcheckr[samp]
            gTest <- datat[!(name %in% ind1 | name %in% ind2),]
            zTTrain <- rbind(zTrain[1:dim(gTrainz)[1],],gTrainz) # check if dim(gTrainz)[1] > 0 !!!!
            zTname <- c(as.character(zname[1:dim(gTrainz)[1]]),as.character(gnamez))
            nt <- nrow(zTTrain)
            zTcheck <- c(as.character(zcheck[1:dim(gTrainz)[1]]),as.character(gcheckz))
            gTest <- rbind(gTest,zTrain[(dim(gTrainz)[1]+1):(dim(gTrainr)[1]),])
          } else {
            zTTrain <- gTrain
            nt <- nrow(zTTrain)
            zTname <- gname
            zTcheck <- gcheck
          }
          outAll <- PrInDTAll(zTTrain,class,ctestv=ctestv,conf.level=co,minsplit=minsplit,minbucket=minbucket)
#print(outAll$treeAll)
          if (outAll$interpAll == FALSE){
            interp <- interp + 1
          } 
          if (crit == "ba"){
            class.pred <- predict(outAll$treeAll,newdata=valdat)
            class.pred <- relevel(class.pred,ref=levels(valdat$class)[1])
          # sum(class.pred == data$class) / length(class.pred)
            nam1 <- list()
            nam2 <- list()
            if (indrep == 2){
        #   acc1E <- 0
        #  acc2E <- 0
        #  pred <- predict(out$modbest,newdata=data)
              ch <- table(repvar,class.pred)
        # class <- datain[,names(datain)==classname]
              conti <- cbind(table(repvar,valdat$class),ch)
              no1 <- sum(conti[,1] > 0)
              no2 <- sum(conti[,2] > 0)
              for (i in 1:dim(conti)[1]){
                if (conti[i,2] > 0 & conti[i,4] <= (conti[i,2]*thr)){ 
                  nam2 <- c(nam2,rownames(ch)[i])
                }
                if (conti[i,1] > 0 & conti[i,3] < (conti[i,1]*thr)){ 
                  nam1 <- c(nam1,rownames(ch)[i])
                }
              }
              ba2 <- 1- length(nam2) / no2
              ba1 <- 1 - length(nam1) / no1  
              ba <- (ba1 + ba2)/2
              if (length(nam1) == 0) {nam1 <- "-"}
              if (length(nam2) == 0) {nam2 <- "-"}
            } else {
              conti <- table(class.pred,valdat$class)
              if (dim(conti)[1] == 1){                                                ## NEWNEW
                conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
              }
              ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
              ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
              ba <- (ba1 + ba2)/2
            }
            if (ba > bamax & outAll$interpAll == FALSE){
              bamax <- ba
              acc1 <- ba1
              acc2 <- ba2
              ntmax <- nt
              dmax <- D
              indmax <- colnames(zTTrain)[1:d]
              ind1max <- ind1
              ind2max <- ind2
              nam1max <- nam1
              nam2max <- nam2
              outmax <- outAll
              gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
              colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
              gmaxTest <- gTest
            }
          }
          if (crit == "bat"){
            class.pred <- predict(outAll$treeAll,newdata=gTest)
            class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
            # sum(class.pred == gTest$class) / length(class.pred)
            conti <- table(class.pred,gTest$class)
            if (dim(conti)[1] == 1){                                                ## NEWNEW
              conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
            }
            ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
            ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
            ba <- (ba1 + ba2)/2
            if (ba > bamax & outAll$interpAll == FALSE){
              bamax <- ba
              acc1 <- ba1
              acc2 <- ba2
              ntmax <- nt 
              dmax <- d
              indmax <- colnames(zTTrain)[1:d]
#print(indmax)
              ind1max <- ind1
              ind2max <- ind2
              outmax <- outAll
              gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
              colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
              gmaxTest <- gTest
            }
          }
          if (crit == "ta"){
            class.pred <- predict(outAll$treeAll,newdata=gTest)
            class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
            conti <- table(class.pred,gTest$class)
            if (dim(conti)[1] == 1){                                                ## NEWNEW
              conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
            }
            ta1 <- conti[1,1] / (conti[1,1] + conti[2,1])
            ta2 <- conti[2,2] / (conti[1,2] + conti[2,2])
            ta <- sum(class.pred == gTest$class) / length(class.pred)
            bat <- (ta1 + ta2)/2
            if (ta > tamax & outAll$interpAll == FALSE){
              bamax <- bat
              tamax <- ta
              acc1 <- ta1
              acc2 <- ta2
              ntmax <- nt 
              dmax <- d
              indmax <- colnames(zTTrain)[1:d]
              ind1max <- ind1
              ind2max <- ind2
              outmax <- outAll
              gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
              colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
              gmaxTest <- gTest
            }
          }
          if (crit == "tal"){
            sTest <- stest[!(name %in% ind1 | name %in% ind2)]  ### NEWNEW
            splitTest <- table(gnameT)  
            for (i in 2:length(splitTest)){
              splitTest[i] <- splitTest[i] + splitTest[i-1]  #### splits according to gTest
            }
            class.pred <- predict(outAll$treeAll,newdata=gTest)
            class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
            conti <- table(class.pred,gTest$class)
            if (dim(conti)[1] == 1){                                                ## NEWNEW
              conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
            }
            ta1 <- conti[1,1] / (conti[1,1] + conti[2,1])
            ta2 <- conti[2,2] / (conti[1,2] + conti[2,2])
            bat <- (ta1 + ta2)/2
            ta <- sum(class.pred == gTest$class) / length(class.pred)
            predind <- class.pred == gTest$class
            indfal <- which(predind == FALSE)
            indfal <- indfal[!(indfal %in% splitTest)]
            indsplit <- split(indfal, cumsum(c(1, diff(indfal) != 1)))
            j <- 0
            indlong <- list()
            indstart <- vector()
            indlen <- vector()
            sumlen <- 0
            for (l in 1:length(indsplit)){
              if (length(unlist(indsplit[as.character(l)])) > ktest){ 
                j <- j + 1
                indlong[j] <- indsplit[as.character(l)]
                indstart[j] <- unlist(indlong[j])[1]
                indlen[j] <- length(unlist(indlong[j]))
                sumlen <- sumlen + indlen[j]
              }
            }
            kstart <- length(indstart)
            labku <- unique(sTest[indstart])
            kustart <- length(labku)
            if ((kustart < kumin & outAll$interpAll == FALSE) |
                (kustart == kumin & ta > tamax & outAll$interpAll == FALSE) |
                (kustart == kumin & ta == tamax & sumlen < summin & outAll$interpAll == FALSE) ){
              kumin <- kustart
              labkumin <- labku
              summin <- sumlen
              mindstart <- indstart  ## NEWNEW
              mindlong <- indlong
#print(kumin)
#print(labkumin)
#print(ta)
              tamax <- ta
              bamax <- bat
              acc1 <- ta1
              acc2 <- ta2
              ntmax <- nt 
              dmax <- d
              indmax <- colnames(zTTrain)[1:d]
              ind1max <- ind1
              ind2max <- ind2
              outmax <- outAll
              gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
              colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
              gmaxTest <- gTest
            }
          }
        }
      } 
    } else {  ## N = 0
#print(n)
      gTrain <- datat
      zTTrain <- gTrain
      nt <- nrow(zTTrain)
      outAll <- PrInDTAll(zTTrain,class,ctestv=ctestv,conf.level=co,minsplit=minsplit,minbucket=minbucket)
      if (outAll$interpAll == FALSE){
        interp <- interp + 1
      }
#plot(outAll$treeAll)
      class.pred <- predict(outAll$treeAll,newdata=data)
#print(class.pred)
      class.pred <- relevel(class.pred,ref=levels(data$class)[1])
      conti <- table(class.pred,data$class)
      if (dim(conti)[1] == 1){                                                ## NEWNEW
        conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
      }
#### setting standards
      sTest <- stest
      gTest <- data            
      zTname <- name
      zTcheck <- check
      if (crit == "ba"){
        class.pred <- predict(outAll$treeAll,newdata=valdat)
        class.pred <- relevel(class.pred,ref=levels(valdat$class)[1])
      # sum(class.pred == data$class) / length(class.pred)
        nam1 <- list()
        nam2 <- list()
        if (indrep == 2){
        #  acc1E <- 0
        #  acc2E <- 0
        #  pred <- predict(out$modbest,newdata=data)
          ch <- table(repvar,class.pred)
        # class <- datain[,names(datain)==classname]
          conti <- cbind(table(repvar,valdat$class),ch)
          no1 <- sum(conti[,1] > 0)
          no2 <- sum(conti[,2] > 0)
          for (i in 1:dim(conti)[1]){
            if (conti[i,2] > 0 & conti[i,4] <= (conti[i,2]*thr)){ 
              nam2 <- c(nam2,rownames(ch)[i])
            }
            if (conti[i,1] > 0 & conti[i,3] < (conti[i,1]*thr)){ 
              nam1 <- c(nam1,rownames(ch)[i])
            }
          }
          ba2 <- 1- length(nam2) / no2
          ba1 <- 1 - length(nam1) / no1  
          ba <- (ba1 + ba2)/2
          if (length(nam1) == 0) {nam1 <- "-"}
          if (length(nam2) == 0) {nam2 <- "-"}
        } else {
          conti <- table(class.pred,valdat$class)
          if (dim(conti)[1] == 1){                                                ## NEWNEW
            conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
          }
          ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
          ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
          ba <- (ba1 + ba2)/2
        }
        if (ba > bamax & outAll$interpAll == FALSE){
          bamax <- ba
          acc1 <- ba1
          acc2 <- ba2
          ntmax <- nt
          dmax <- D
          indmax <- colnames(zTTrain)[1:D]
          ind1max <- ind1
          ind2max <- ind2
          nam1max <- nam1
          nam2max <- nam2
          outmax <- outAll
          gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
          colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
          gmaxTest <- gTest
        }
      }
      if (crit == "bat"){
        class.pred <- predict(outAll$treeAll,newdata=gTest)
        class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
        # sum(class.pred == gTest$class) / length(class.pred)
        conti <- table(class.pred,gTest$class)
        if (dim(conti)[1] == 1){                                                ## NEWNEW
          conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
        }
        ba1 <- conti[1,1] / (conti[1,1] + conti[2,1])
        ba2 <- conti[2,2] / (conti[1,2] + conti[2,2])
        ba <- (ba1 + ba2)/2
        if (ba > bamax & outAll$interpAll == FALSE){
          bamax <- ba
          acc1 <- ba1
          acc2 <- ba2
          ntmax <- nt 
          dmax <- d
          indmax <- colnames(zTTrain)[1:d]
#print(indmax)
          ind1max <- indmax
          ind2max <- indmax
          outmax <- outAll
          gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
          colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
          gmaxTest <- gTest
        }
      }
      if (crit == "ta"){
        class.pred <- predict(outAll$treeAll,newdata=gTest)
        class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
        conti <- table(class.pred,gTest$class)
        if (dim(conti)[1] == 1){                                                ## NEWNEW
          conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
        }
        ta1 <- conti[1,1] / (conti[1,1] + conti[2,1])
        ta2 <- conti[2,2] / (conti[1,2] + conti[2,2])
        ta <- sum(class.pred == gTest$class) / length(class.pred)
        bat <- (ta1 + ta2)/2
        if (ta > tamax & outAll$interpAll == FALSE){
          bamax <- bat
          tamax <- ta
          acc1 <- ta1
          acc2 <- ta2
          ntmax <- nt 
          dmax <- d
          indmax <- colnames(zTTrain)[1:d]
          ind1max <- indmax
          ind2max <- indmax
          outmax <- outAll
          gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
          colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
          gmaxTest <- gTest
        }
      }
      if (crit == "tal"){
        splitTest <- table(sTest) ## no element selection! 
        for (i in 2:length(splitTest)){
          splitTest[i] <- splitTest[i] + splitTest[i-1]  #### splits according to sTest
        }
        class.pred <- predict(outAll$treeAll,newdata=gTest)
        class.pred <- relevel(class.pred,ref=levels(gTest$class)[1])
        conti <- table(class.pred,gTest$class)
#print(conti)
        if (dim(conti)[1] == 1){                                                ## NEWNEW
          conti <- matrix(c(conti,0,0),nrow=2,byrow=TRUE)
        }
        ta1 <- conti[1,1] / (conti[1,1] + conti[2,1])
        ta2 <- conti[2,2] / (conti[1,2] + conti[2,2])
        bat <- (ta1 + ta2)/2
        if (bat == 0.5){
          kustart <- kumin + 1  # whole small class is wrongly predicted
        } else {
          ta <- sum(class.pred == gTest$class) / length(class.pred)
          predind <- class.pred == gTest$class
          indfal <- which(predind == FALSE)
          indfal <- indfal[!(indfal %in% splitTest)]
          indsplit <- split(indfal,cumsum(c(1, diff(indfal) != 1))) 
#print(length(indsplit))
          j <- 0
          indlong <- list()
          indstart <- vector()
          indlen <- vector()
          sumlen <- 0
          for (l in 1:length(indsplit)){
            if (length(unlist(indsplit[as.character(l)])) > ktest){ 
              j <- j + 1
              indlong[j] <- indsplit[as.character(l)]
              indstart[j] <- unlist(indlong[j])[1]
              indlen[j] <- length(unlist(indlong[j]))
              sumlen <- sumlen + indlen[j]
            }
          }
          kstart <- length(indstart)
          labku <- unique(sTest[indstart])
          kustart <- length(labku)
          if ((kustart < kumin & outAll$interpAll == FALSE) |
              (kustart == kumin & ta > tamax & outAll$interpAll == FALSE) |
              (kustart == kumin & ta == tamax & sumlen < summin & outAll$interpAll == FALSE) ){
            kumin <- kustart
            summin <- sumlen
            labkumin <- labku
            mindstart <- indstart  ## NEWNEW
            mindlong <- indlong
#print(kumin)
#print(labkumin)
#print(ta)
            tamax <- ta
            bamax <- bat
            acc1 <- ta1
            acc2 <- ta2
            ntmax <- nt 
            dmax <- d
            indmax <- colnames(zTTrain)[1:d]
            ind1max <- indmax
            ind2max <- indmax
            outmax <- outAll
            gmaxTrain <- cbind(zTTrain,as.vector(zTname),as.vector(zTcheck))
            colnames(gmaxTrain) <- c(colnames(zTTrain),"SubStruc","check")
            gmaxTest <- data
          }
        }
      }
    } 
  }
  if (crit == "ba"){
    message("current optimal balanced accuracy: ",round(bamax,digits=4),"\n")
    tamax <- 0
    kumin <- 0
    elems <- 0
  }
  if (crit == "bat"){
    message("current optimal balanced test accuracy: ",round(bamax,digits=4),"\n")
    tamax <- 0
    kumin <- 0
    elems <- 0
  }
  if (crit == "ta"){
    message("current optimal test accuracy: ",round(tamax,digits=4),"\n")
    elems <- 0
    kumin <- 0
  }
  if (crit == "tal"){
    message("current optimal number of long misclassified parts: ",kumin)
    elems <- labkumin
    message("current optimal test accuracy: ",round(tamax,digits=4),"\n")
  }
#if (crit == "ba"){message("current optimal balanced accuracy: ",round(bamax,digits=4))}
#if (crit == "bat"){message("current optimal balanced test accuracy: ",round(bamax,digits=4))}
#if (crit == "ta"){message("current optimal test accuracy: ",round(tamax,digits=4))} # {print(tamax)}
#if (crit == "tal"){message("current optimal number of wrongly predicted test squencies longer than ",ktest,": ",round(kumin))} # {print(kumin)}
}
if (N > 0){
  interp <- c(interp,(lMit*N*Ni*lDit))
} else {
  interp <- c(interp,(Ni*lDit))
}
result <- list(interp=interp,dmax=dmax,ntmax=ntmax,acc1=acc1,acc2=acc2,nam1=nam1max,nam2=nam2max,
     bamax=bamax,tamax=tamax,kumin=kumin,elems=elems,mindlong=mindlong,
     ind1max=ind1max,ind2max=ind2max,indmax=indmax,modbest = outmax$treeAll,gmaxTrain=gmaxTrain,gmaxTest=gmaxTest)
}

#' Posterior analysis of conditional inference trees: distribution of a specified variable in the terminal nodes.
#'
#' @description The conditional inference tree 'ct' is analyzed according to the distribution of a variable 'var' in its terminal nodes.\cr
#' In the case of a discrete variable 'var', the appearance of the different levels is considered for each terminal node. \cr
#' In the case of a continuous variable 'var', means and standard deviations of 'var' or the target variable are considered for each terminal node.\cr
#' In particular, this function can be used for the posterior analysis of a tree regarding the distribution of a variable not present in the tree.
#' 
#' @usage PostPrInDT(datain, ct, target, var, vardata, vt)
#'
#' @param datain input data frame with the observatios of all variables used in the model 
#' @param ct conditional inference tree to be analyzed
#' @param target name of target variable of 'ct' (character)
#' @param var name of variable of interest (character)
#' @param vardata observations of 'var'
#' @param vt type of variables: 'dd' for discrete target (classification) and discrete variable 'var', 'dc' for discrete target (classification) and continuous 'var',
#' 'cd' for continuous target (regression) and discrete 'var', and 'cc' for continuous target (regression) and continuous 'var'.
#'
#' @return None: Relevant output is produced by the function.
#'
#' @export PostPrInDT 
#'
#' @examples
#' data <- PrInDT::data_zero
#' data <- na.omit(data)
#' outAll <- PrInDTAll(data,"real") 
#' PostPrInDT(data,outAll$treeAll,"real","ETH",data$ETH,vt="dd")
#' PostPrInDT(data,outAll$treeAll,"real","AGE",data$AGE,vt="dc")
#' datareg <- PrInDT::data_vowel
#' outregAll <- PrInDTregAll(datareg,"target") 
#' PostPrInDT(datareg,outregAll$treeAll,"target","Nickname",datareg$Nickname,vt="cd")
#' PostPrInDT(datareg,outregAll$treeAll,"target","AGE",datareg$AGE,vt="cc")
#'
#' @import party
#'
PostPrInDT <- function(datain,ct,target,var,vardata,vt){
  ## input check
  if (typeof(datain) != "list" || typeof(ct) != "S4" || typeof(var) != "character" || typeof(target) != "character" ||
      !(typeof(vardata) %in% c("integer","double")) || typeof(vt) != "character"){
    stop("irregular input")
  }
####
  data <- datain
  colnames(data)[colnames(data)==target] <- "target"
  if (vt == "dd" | vt == "cd"){
    check <- cbind( as.character(predict(ct,newdata=data)) , where(ct,newdata = data), as.character(vardata))
    tvar <- table(check[,2],check[,3])
    cat("\n","Occurence of levels of variable *** ",var," *** in terminal nodes of the following tree")
    print(ct)
    cat("\n","Nodes in rows, levels in columns")
    print(tvar)
  }
  if (vt == "dc" | vt == "cc"){
    cat("\n","Means and standard deviations of values of variable *** ",var," *** in terminal nodes of the following tree","\n")
    print(ct)
    tt <- table(where(ct,newdata=data))
    means <- rep( 0, dim(tt)[1] )
    stdevs <- means
    for (i in 1:dim(tt)[1]) {
        mm <- mean( vardata[ where(ct,newdata=data) == names(tt)[i] ])
        if (is.na(mm) == FALSE) {means[i] <- mm}
        st <- sd( vardata[ where(ct,newdata=data) == names(tt)[i] ] )
        if (is.na(st) == FALSE) {stdevs[i] <- st}
    }
    tvar <- rbind(means,stdevs)
    rownames(tvar) <- c("means","stdevs")
    colnames(tvar) <- names(tt)
    cat("\n","Means and standard deviations in nodes","\n")
    print(tvar)
  }
  if (vt == "cd" |vt == "cc"){
    cat("\n","Means and standard deviations of values of target variable *** ",target," *** in terminal nodes of the following tree","\n")
    print(ct)
    tt <- table(where(ct,newdata=data))
    means <- rep( 0, dim(tt)[1] )
    stdevs <- means
    for (i in 1:dim(tt)[1]) {
        mm <- mean( data$target[ where(ct,newdata=data) == names(tt)[i] ])
        if (is.na(mm) == FALSE) {means[i] <- mm}
        st <- sd( data$target[ where(ct,newdata=data) == names(tt)[i] ] )
        if (is.na(st) == FALSE) {stdevs[i] <- st}
    }
    tvar <- rbind(means,stdevs)
    rownames(tvar) <- c("means","stdevs")
    colnames(tvar) <- names(tt)
    cat("\n","Means and standard deviations in nodes","\n")
    print(tvar)
  }
}

#' Repeated \code{\link{PrInDT}} for specified percentage combinations
#'
#' @description
#' By the function \code{\link{RePrInDT}}, the function \code{\link{PrInDT}} is called repeatedly according to all combinations of the percentages specified in the vectors 'plarge' and 'psmall'.\cr
#' The relationship between the two-class factor variable 'classname' and all other factor and numerical variables
#' in the data frame 'datain' is optimally modeled by means of 'N' repetitions of undersampling.\cr 
#' The optimization citerion is the balanced accuracy on the validation sample 'valdat' (default = full input sample 'datain').\cr
#' The trees generated from undersampling can be restricted by rejecting 
#' unacceptable trees which include split results specified in the character strings of the vector 'ctestv'.\cr
#' The probability threshold 'thres' for the prediction of the smaller class may be specified (default = 0.5).\cr
#' Undersampling may be stratified in two ways by the feature 'strat'.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' \strong{Reference}\cr Weihs, C., Buschfeld, S. 2021c. Repeated undersampling in PrInDT (RePrInDT): Variation in undersampling and prediction, 
#' and ranking of predictors in ensembles. arXiv:2108.05129
#'
#' @usage RePrInDT(datain, classname, ctestv=NA, N, plarge, psmall, conf.level=0.95,
#'        thres=0.5, stratvers=0, strat=NA, seedl=TRUE,minsplit=NA,minbucket=NA,
#'        valdat=datain)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param N Number of repetitions (integer > 0)
#' @param plarge Vector of undersampling percentages of larger class (numerical, > 0 and <= 1)
#' @param psmall Vector of undersampling percentages of smaller class (numerical, > 0 and <= 1)
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param thres Probability threshold for prediction of smaller class (numerical, >= 0 and < 1); default = 0.5
#' @param stratvers Version of stratification;\cr
#'     = 0: none (default),\cr
#'     = 1: stratification according to the percentages of the values of the factor variable 'strat',\cr
#'     > 1: stratification with minimum number 'stratvers' of observations per value of 'strat'
#' @param strat Name of one (!) stratification variable for undersampling (character);\cr
#'     default = NA (no stratification)
#' @param seedl Should the seed for random numbers be set (TRUE / FALSE)?\cr
#'     default = TRUE
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#' @param valdat Validation data; default = datain
#'
#' @return
#' \describe{
#' \item{treesb}{best trees for the different percentage combinations; refer to an individual tree as \code{treesb[[k]]}, k = 1, ..., length(plarge)*length(psmall)}
#' \item{acc1st}{accuracies of best trees on full sample}
#' \item{acc3en}{accuracies of ensemble of 3 best trees on full sample}
#' \item{simp_m}{mean of permutation losses for the predictors}
#' }
#'
#' @details
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
#' frame of the function.\cr
#' The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
#' \code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
#'
#' @export RePrInDT
#' @importFrom graphics barplot

#' @examples
#' datastrat <- PrInDT::data_zero
#' data <- na.omit(datastrat) # cleaned full data: no NAs
#' # interpretation restrictions (split exclusions)
#' ctestv <- rbind('ETH == {C2a, C1a}', 'MLU == {1, 3}')
#' N <- 51  # no. of repetitions
#' conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
#' psmall <- c(0.95,1)     # percentages of the small class
#' plarge <- c(0.09,0.1)  # percentages of the large class
#' outRe <- RePrInDT(data,"real",ctestv,N,plarge,psmall,conf.level) 
#' outRe
#' plot(outRe)
#'
RePrInDT <- function(datain,classname,ctestv=NA,N,plarge,psmall,conf.level=0.95,thres=0.5,stratvers=0,strat=NA,seedl=TRUE,minsplit=NA,minbucket=NA,valdat=datain){
  ## input check
  if (typeof(datain) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) || N <= 0 ||
      !(all(0 < plarge & plarge <= 1)) || !(all(0 < psmall & psmall <= 1)) ||
      !(0 < conf.level & conf.level <= 1) | !(0 <= thres & thres < 1) ||
      !(0 <= stratvers) || !(typeof(strat) %in% c("logical", "character")) || typeof(seedl) != "logical" || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double")) || typeof(valdat) != "list" ) {
    stop("irregular input")
  }
  if (!(all(names(datain) %in% names(valdat)))){
    stop("validation data variables unequal input data variables")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
  if (seedl == TRUE){
    set.seed(7654321)  # set seed of random numbers
  }
  ###
  data <- datain
  names(data)[names(data)==classname] <- "class"
  names(valdat)[names(valdat)==classname] <- "class"
  if (!(identical(levels(data$class),levels(valdat$class)))){
    stop("levels of input class variable unequal levels of validation class variable")
  }
  ## initializations
  norep <- length(psmall) * length(plarge)
  acc1st <- array(0,dim=c(length(plarge),length(psmall),8))
  acc3en <- array(0,dim=c(length(plarge),length(psmall),6)) # !!!
  simp <- array(0,dim=c(length(plarge),length(psmall),(dim(data)[2])) )
  ## permuting columns
  data_per <- valdat
  for (j in 2:(dim(valdat)[2])) {
    data_per[,j] <- sample(valdat[,j])
  }
  k <- 0
  K <- length(plarge) * length(psmall)
  for (i in 1:length(plarge)) {
    for (j in 1:length(psmall)) {
      k <- k + 1
#      message("\n")
      message("\n","sampling percentage for larger class: ",plarge[i])
      message("sampling percentage for smaller class:",psmall[j])
      ## call of PrInDT
      out <- PrInDT(data,classname,ctestv,N,plarge[i],psmall[j],conf.level,thres,stratvers,strat,seedl=TRUE,minsplit=minsplit,minbucket=minbucket,valdat=valdat)
      if (any(is.na(out)) == TRUE){ # !!!
        return(NA)
      }
      acc1st[i,j,] <- c(plarge[i],psmall[j],out$ba1st)
      acc3en[i,j,] <- c(plarge[i],psmall[j],out$baen[2,2:5])
      if (k == 1){
        treesb <- out$tree1st
        n_class1 <- table(out$valdat$class)[1] # no. of elements of larger class 1
        n_class2 <- table(out$valdat$class)[2] # no. of elements of smaller class 2
      } else {
        treesb <- c(treesb,out$tree1st)
      }
      
#      if (out$ba1st[3] > 0){
        for (l in 2:dim(out$valdat)[2]){
          data_imp <- out$valdat
          data_imp[,l] <- data_per[,l]
          ctpreds_imp <- predict(out$tree1st,newdata=data_imp)
          conf_imp <- table(ctpreds_imp, data_imp$class)
          simp[i,j,l] <- simp[i,j,l] + out$ba1st[3] - (conf_imp[1,1] / n_class1 + conf_imp[2,2] / n_class2)/2
          ctpreds_imp <- predict(out$tree2nd,newdata=data_imp)
          conf_imp <- table(ctpreds_imp, data_imp$class)
          simp[i,j,l] <- simp[i,j,l] + out$ba2nd[3] - (conf_imp[1,1] / n_class1 + conf_imp[2,2] / n_class2)/2
          ctpreds_imp <- predict(out$tree3rd,newdata=data_imp)
          conf_imp <- table(ctpreds_imp, data_imp$class)
          simp[i,j,l] <- simp[i,j,l] + out$ba3rd[3] - (conf_imp[1,1] / n_class1 + conf_imp[2,2] / n_class2)/2
        }
#      }
#      if (k < K){
#        remove(out,data_imp,ctpreds_imp,conf_imp)
#        gc(full=TRUE,reset=TRUE)
#        gc(full=TRUE,reset=TRUE)
#      }
    }
  }
 
## preparation of print
  dimnames(acc1st) <- list(1:length(plarge),1:length(psmall),c("plarge","psmall",paste0("validation ",levels(out$valdat$class)[1]),
                                                               paste0("validation ",levels(out$valdat$class)[2]),"validation balanced",
                                                               paste0("test ",levels(out$valdat$class)[1]),paste0("test ",levels(out$valdat$class)[2]),"test balanced"))
  dimnames(acc3en) <- list(1:length(plarge),1:length(psmall),c("plarge","psmall",paste0("validation ",levels(out$valdat$class)[1]),
                                                               paste0("validation ",levels(out$valdat$class)[2]),"validation balanced","mean test balanced")) # !!!
  simp_m <- sapply(2:dim(data)[2],function(x) mean(simp[,,x]))
  names(simp_m) <- colnames(out$valdat)[-(colnames(out$valdat) == "class")]
###
  result <- list(treesb=treesb,acc1st=acc1st,acc3en=acc3en,simp_m=simp_m)
  class(result) <- "RePrInDT"
  result
}
#' @export
print.RePrInDT <- function(x, ...){
  ####
  ## print accuracies of best trees
  ####
  cat("\n","Accuracies of best trees for different combinations of resampling percentages: on full and test samples","\n")
  apply(x$acc1st,1,print)
  ## print overall best trees
  cat("\n\n","Best trees of percentage combinations","\n")
  maxba <- as.vector(t(x$acc1st[,,5] == max(x$acc1st[,,5])))
  mm <- which(maxba)
  k <- 0 
  for (i in 1:dim(x$acc1st)[1]) {
    for (j in 1:dim(x$acc1st)[2]) {
      k <- k + 1
      if (k %in% mm) {
        cat("\n","tree for  ( plarge  psmall  bal.acc. ) = (",x$acc1st[i,j,c(1,2,5)],")")
        print(x$treesb[[k]])
      }
    }
  }
  ## print accuracies of ensembles of 3 best trees
  cat("\n","Accuracies of ensembles of 3 best trees for different combinations of resampling percentages","\n")
  apply(x$acc3en,1,print)
  ####
  ## ranking of predictors
  ####
  cat("\n\n","Mean of permutation losses for the predictors","\n")
  print(x$simp_m)
}
#' @export
plot.RePrInDT <- function(x, ...){
  ####
  ## plot best trees
  ####
  maxba <- as.vector(t(x$acc1st[,,5] == max(x$acc1st[,,5])))
  mm <- which(maxba)
  lmm <- length(mm)
  k <- 0 
  for (i in 1:dim(x$acc1st)[1]) {
    for (j in 1:dim(x$acc1st)[2]) {
      k <- k + 1
      if (k %in% mm) {
        x$acc1st[i,j,5] <- round(x$acc1st[i,j,5],4)
        v <- toString(x$acc1st[i,j,c(1,2,5)])
        plot(x$treesb[[k]],main=paste0("One of the ",lmm," best trees: ( plarge,  psmall,  bal.acc. ) = (",v,")"))
      }
    }
  }
  ####
  ## ranking of predictors
  ####
  cat("\n")
  barplot(sort(x$simp_m[x$simp_m>0] / max(x$simp_m)),horiz=TRUE,main="Normed means of permutation losses")
}

#' Regression tree resampling by the PrInDT method
#'
#' @description The function \code{\link{PrInDTreg}} realizes regression tree optimization to identify the best interpretable tree; interpretability is checked (see 'ctestv').\cr
#' The relationship between the target variable 'regname' and all other factor and numerical variables
#' in the data frame 'datain' is optimally modeled by means of 'N' repetitions of subsampling.\cr 
#' The optimization criterion is the R2 of the model on the validation sample 'valdat'.\cr
#' Default for the validation sample is the full sample 'datain'.\cr
#' Multiple subsampling percentages of observations and predictors can be specified (in 'pobs' and 'ppre', correspondingly).\cr
#' The trees generated from subsampling can be restricted by
#' rejecting unacceptable trees which include split results specified in the character strings of the vector 'ctestv'.\cr
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#' Besides the maximal R2, the minimal MAE (Mean Absolute Error) is reported.
#'
#' @usage PrInDTreg(datain,regname,ctestv=NA,N,pobs=c(0.9,0.7),ppre=c(0.9,0.7),
#'                conf.level=0.95,seedl=TRUE,minsplit=NA,minbucket=NA,valdat=datain)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param regname name of regressand variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param N Number of repetitions (integer > 0)
#' @param pobs Vector of resampling percentages of observations (numerical, > 0 and <= 1)
#' @param ppre Vector of resampling percentages of predictor variables (numerical, > 0 and <= 1)
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
#'     default = 0.95
#' @param seedl Should the seed for random numbers be set (TRUE / FALSE)?\cr
#'     default = TRUE
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#' @param valdat Validation data; default = datain
#'
#' @return 
#' \describe{
#' \item{meanint}{Mean number of interpretable trees over the combinations of individual percentages in 'pobs' and 'ppre'}
#' \item{ctmax}{best resampled regression tree on the validation data set}
#' \item{percmax}{Best model achieved for \%observations}
#' \item{perfeamax}{Best model achieved for \%predictors}
#' \item{maxR2}{maximum R2 on the validation data set for resampled regression trees (for 'ctmax')} 
#' \item{minMAE}{minimum MAE (Mean Absolute Error) on the validation data set for resampled regression trees (for 'ctmax')} 
#' \item{interpmax}{interpretability of best tree 'ctmax'}
#' \item{ctmax2}{second best resampled regression tree on the full data set}
#' \item{percmax2}{second best model achieved for \%observations}
#' \item{perfeamax2}{second best model achieved for \%features}
#' \item{max2R2}{second best R2 on the validation data set for resampled regression trees (for 'ctmax2')}
#' \item{min2MAE}{second best MAE on the validation data set for resampled regression trees (for 'ctmax2')}
#' \item{interp2max}{interpretability of second-best tree 'ctmax2'}
#' }
#'
#' @details
#' For the optimzation of the trees, we employ a method we call Sumping (Subsampling umbrella of 
#' model parameters), a variant of Bumping (Bootstrap umbrella of model parameters) (Tibshirani 
#' & Knight, 1999). We use subsampling instead of bootstrapping. The aim of the 
#' optimization is to identify conditional inference trees with maximum predictive power
#' on the full sample under interpretability restrictions. 
#'
#' \strong{Reference}\cr Tibshirani, R., Knight, K. 1999. Model Search and Inference By Bootstrap "bumping".
#'            Journal of Computational and Graphical Statistics, Vol. 8, No. 4 (Dec., 1999), pp. 671-686 
#'  
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
#' frame of the function.\cr
#' The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
#' \code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
#'
#' @export PrInDTreg
#'
#' @examples
#' data <- PrInDT::data_vowel
#' data <- na.omit(data)
#' ctestv <- 'vowel_maximum_pitch <= 320'
#' N <- 30 # no. of repetitions
#' pobs <- c(0.70,0.60) # percentages of observations
#' ppre <- c(0.90,0.70) # percentages of predictors
#' outreg <- PrInDTreg(data,"target",ctestv,N,pobs,ppre)
#' outreg
#' plot(outreg)
#'
#' @importFrom party ctree ctree_control
#' @importFrom stats formula
#'
PrInDTreg <- function(datain,regname,ctestv=NA,N,pobs=c(0.9,0.7),ppre=c(0.9,0.7),
    conf.level=0.95,seedl=TRUE,minsplit=NA,minbucket=NA,valdat=datain){
  ## input check
  if (typeof(datain) != "list" || typeof(regname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) || N <= 0 ||
      !all(0 < pobs & pobs <= 1) || !all(0 < ppre & ppre <= 1) 
      || !(0 < conf.level & conf.level <= 1) || typeof(seedl) != "logical"
      || !(typeof(minsplit) %in% c("logical","double")) || !(typeof(minbucket) %in% c("logical", "double")) || typeof(valdat) != "list" ){
    stop("irregular input")
  }
  if (!(all(names(datain) %in% names(valdat)))){
    stop("validation data variables unequal input data variables")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
####
  data <- datain
  maxR2 <- 0
  max2R2 <- 0
#  R2mean <- matrix(0,nrow = length(pobs), ncol = length(ppre))
  interpmax <- FALSE
  interp2max <- FALSE
  meanint <- 0
  if (seedl == TRUE){
    set.seed(1234567)  # set seed of random numbers
  }
#  set.seed(1234567)
  if (any(pobs*dim(data)[1] < 10)){
      stop("For one of the percentages: Number of tokens too small (< 10)")
  }
  if (any(ppre*dim(data)[2] < 1)){
      stop("For one of the percentages: Number of predictors too small (< 1)")
  }
  message("combinations of percentages of","\n","observations & predictors")
  for (i in 1:length(pobs)) {
    perc <- pobs[i]
    for (j in 1:length(ppre)) {
      perfea <- ppre[j]
      message("    ",perc,"         ",perfea)
  #    R2 <- 0
      crit1 <- rep(FALSE,N) # vector for crit1 (interpretability)
      for (k in 1:N) {
        samplerows <- sample(1:nrow(data), perc*dim(data)[1],replace=FALSE)
        datasub <- data[samplerows,]
#        datatest <- data[-samplerows,]
        datasub2 <- datasub[,sample(1:ncol(data), perfea*dim(data)[2],replace=FALSE)]
        datasub2[,regname] <- datasub[,regname]
        ct <- party::ctree(formula(paste(regname, "~ .")), data = datasub2,control = 
                 party::ctree_control(mincriterion=conf.level,minsplit=minsplit,minbucket=minbucket))
        if (all(is.na(ctestv)) == FALSE) {
          crit1[k] <- FindSubstr(ct,ctestv) # deciding on interpretability
        }
        if (crit1[k] == FALSE){
          meanint <- meanint + 1
        } 
        ctpreds <- predict(ct,newdata=valdat)
        R2 <- 1 - sum((valdat[,regname] - ctpreds)^2) / sum((valdat[,regname]-mean(data[,regname]))^2)
        MAE <- sum(abs(valdat[,regname] - ctpreds))/ dim(valdat)[1]
        if (maxR2 == 0 & crit1[k] == FALSE) {
          maxR2 <- R2
          minMAE <- MAE
          ctmax <- ct
          ctmax2 <- ct
          max2R2 <- maxR2
          min2MAE <- minMAE
          percmax <- perc
          percmax2 <- percmax
          perfeamax <- perfea
          perfeamax2 <- perfeamax
          interpmax <- crit1[k]
          interp2max <- interpmax
        }
        if (R2 > maxR2 & crit1[k] == FALSE) {
#print(maxR2)
          max2R2 <- maxR2
          maxR2 <- R2
          min2MAE <- minMAE
          minMAE <- MAE
          ctmax2 <- ctmax
          ctmax <- ct
          percmax2 <- percmax
          perfeamax2 <- perfeamax
          percmax <- perc
          perfeamax <- perfea
          interp2max <- interpmax
          interpmax <- crit1[k]
        }
#        ctpreds <- predict(ct,newdata=datatest)
#        R2 <- 1 - sum((datatest[,regname] - ctpreds)^2) / sum((datatest[,regname]-mean(datatest[,regname]))^2)
#        R2mean[i,j] <- R2mean[i,j] + R2
      }
#      R2mean[i,j] <- R2mean[i,j] / N
    }
  }
  meanint <- meanint / (length(pobs)*length(ppre)) # mean number of interpretable trees over the combinations
  # if (crit == "MAE"){
  #  maxR2 <- -maxR2 # MAE had to be minimized
  #  max2R2 <- -max2R2
  # }
#  colnames(R2mean) <- as.character(ppre)
#  rownames(R2mean) <- as.character(pobs)
#  result <- list(meanint = meanint, R2mean = R2mean, ctmax = ctmax, percmax = percmax, perfeamax = perfeamax, maxR2 = maxR2, 
#    interpmax = interpmax, ctmax2 = ctmax2, percmax2 = percmax2, perfeamax2 = perfeamax2, max2R2 = max2R2, interp2max = interp2max)
  result <- list(meanint = meanint, ctmax = ctmax, percmax = percmax, perfeamax = perfeamax, maxR2 = maxR2, minMAE = minMAE,
    interpmax = interpmax, ctmax2 = ctmax2, percmax2 = percmax2, perfeamax2 = perfeamax2, max2R2 = max2R2, min2MAE = min2MAE,
    interp2max = interp2max)
  class(result) <- "PrInDTreg"
  result
}
#' @export
print.PrInDTreg <- function(x,...){
  cat("\n","Mean number of interpretable trees over the combinations: ","\n")
  print(x$meanint)
#  cat("\n","Mean R2 on test sets over the combinations: ","\n")
#  print(x$R2mean)
#
  cat("\n\n","****Best interpretable regression tree****","\n")
  cat("Characteristics","\n")
  cat("  %obs","   %pred","   : ","R2 on validation sample","\n") #      interpretable","\n")
  cat("  ",x$percmax,"   ",x$perfeamax,"    :    ",x$maxR2,"\n") #"    ",!as.logical(x$interpmax),"\n")
  cat("      ","        ","   : ","MAE on validation sample","\n") #      interpretable","\n")
  cat("      ","        ","   :    ",x$minMAE,"\n") #"    ",!as.logical(x$interpmax),"\n")
  print(x$ctmax)
#
  cat("\n\n","****Second best interpretable regression tree****","\n")
  cat("Characteristics","\n")
  cat("  %obs","   %pred","   : ","R2 on validation sample","\n") #      interpretable","\n")
  cat("  ",x$percmax2,"   ",x$perfeamax2,"    :    ",x$max2R2,"\n") #"    ",!as.logical(x$interp2max),"\n")
  cat("      ","        ","   : ","MAE on validation sample","\n") #      interpretable","\n")
  cat("      ","        ","   :    ",x$min2MAE,"\n") #"    ",!as.logical(x$interp2max),"\n")
  print(x$ctmax2)
}
#' @export
plot.PrInDTreg <- function(x,...){
  plot(x$ctmax,main="Best interpretable regression tree")
  plot(x$ctmax2,main="Second best interpretable regression tree")
}
#' @export
predict.PrInDTreg <- function(object,...){
  predict(object$ctmax,...)
}

#' Conditional inference tree (ctree) for multiple classes on all observations
#'
#' @description ctree for more than 2 classes on all observations. Interpretability is checked (see 'ctestv').
#' The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
#'
#' @usage PrInDTMulevAll(datain,classname,ctestv=NA,conf.level=0.95,minsplit=NA,minbucket=NA)
#'
#' @param datain Input data frame with class factor variable 'classname' and the\cr
#'    influential variables, which need to be factors or numericals (transform logicals and character variables to factors) 
#' @param classname Name of class variable (character)
#' @param ctestv Vector of character strings of forbidden split results;\cr
#'     (see function \code{\link{PrInDT}} for details.)\cr
#'     If no restrictions exist, the default = NA is used.
#' @param conf.level (1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1)\cr
#'     (default = 0.95)
#' @param minsplit Minimum number of elements in a node to be splitted;\cr
#'     default = 20
#' @param minbucket Minimum number of elements in a node;\cr
#'     default = 7
#'
#' @return 
#' \describe{
#' \item{treeall}{ctree based on all observations}
#' \item{baAll}{balanced accuracy of 'treeall'}
#' \item{interpAll}{criterion of interpretability of 'treeall' (TRUE / FALSE)}
#' \item{confAll}{confusion matrix of 'treeall'}
#' }
#'
#' @details
#' Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
#' frame of the function.
#'
#' @export PrInDTMulevAll
#'
#' @examples
#' datastrat <- PrInDT::data_zero
#' data <- na.omit(datastrat)
#' ctestv <- rbind('ETH == {C2a,C1a}', 'MLU == {1, 3}')
#' data$rel[data$ETH %in% c("C1a","C1b","C1c") & data$real == "zero"] <- "zero1"
#' data$rel[data$ETH %in% c("C2a","C2b","C2c") & data$real == "zero"] <- "zero2"
#' data$rel[data$real == "realized"] <- "real"
#' data$rel <- as.factor(data$rel) # rel is new class variable
#' data$real <- NULL # remove old class variable
#' conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
#' outAll <- PrInDTMulevAll(data,"rel",ctestv,conf.level) 
#' outAll # print model based on all observations
#' plot(outAll)
#'
#' @importFrom stats relevel predict
#' @importFrom party ctree ctree_control
#'
PrInDTMulevAll <- function(datain,classname,ctestv=NA,conf.level=0.95,minsplit=NA,minbucket=NA){
  ## input check
  if (typeof(datain) != "list" || typeof(classname) != "character" || !(typeof(ctestv) %in% c("logical", "character")) || 
      !(0 < conf.level & conf.level <= 1) || !(typeof(minsplit) %in% c("logical","double")) || 
      !(typeof(minbucket) %in% c("logical", "double"))  ) {
    stop("irregular input")
  }
  if ((is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minsplit <- 20
    minbucket <- 7
  }
  if (!(is.na(minsplit) == TRUE) & (is.na(minbucket) == TRUE)){
    minbucket <- minsplit / 3
  }
  if ((is.na(minsplit) == TRUE) & !(is.na(minbucket) == TRUE)){
    minsplit <- minbucket * 3
  }
  data <- datain
  names(data)[names(data)==classname] <- "class"
  K <- length(table(data$class))
  ######
  ## model with all observations
  ######
  ct <- party::ctree(class ~ ., data = data, control = party::ctree_control(mincriterion=conf.level,minsplit=minsplit,minbucket=minbucket))
  crit1 <- FALSE
  if (is.na(ctestv[1]) == FALSE) {
    crit1 <- FindSubstr(ct,ctestv) # call of the above function for overall tree
  }
  ctpreds <- stats::predict(ct) # predictive fit
  conf <- table(ctpreds,data$class)  # confusion matrix
  ## balanced accuracy
  acc <- rep(0,K)
  for (j in 1:K){
        acc[j] <- conf[j,j] / colSums(conf)[j]
  }
  crit2 <- sum(acc) / K
  result <- list(treeAll = ct, baAll=crit2, confAll = conf, interpAll=crit1)
  class(result) <- "PrInDTMulevAll"
  result
}
#' @export
print.PrInDTMulevAll <- function(x, ...){
  # output for model on all observations
  cat("******************************","\n")
  cat("Tree built on all observations","\n")
  cat("******************************","\n")
  print(x$treeAll) # print of tree
  cat("\n")
  cat("****************","\n")
  cat("Confusion matrix","\n")
  cat("****************")
  print(x$confAll)
  cat("***********************************************","\n")
  cat("Criteria: balanced accuracy","\n")
  cat("***********************************************","\n")
  cat("             ",unname(x$baAll),"\n")
  cat("***********************************************","\n")
  cat("Criteria: interpretable tree","\n")
  cat("***********************************************","\n")
  cat("             ",unname(!(x$interpAll)),"\n")
}
#' @export
plot.PrInDTMulevAll <- function(x, ...){
  plot(x$treeAll,main="Multi-class problem: Tree on full sample") # plot of tree
}
#' @export
predict.PrInDTMulevAll <- function(object,...){
  predict(object$ct,...)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTreg.R
\name{PrInDTreg}
\alias{PrInDTreg}
\title{Regression tree resampling by the PrInDT method}
\usage{
PrInDTreg(datain,regname,ctestv=NA,N,pobs=c(0.9,0.7),ppre=c(0.9,0.7),
               conf.level=0.95,seedl=TRUE,minsplit=NA,minbucket=NA,valdat=datain)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{regname}{name of regressand variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{N}{Number of repetitions (integer > 0)}

\item{pobs}{Vector of resampling percentages of observations (numerical, > 0 and <= 1)}

\item{ppre}{Vector of resampling percentages of predictor variables (numerical, > 0 and <= 1)}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{seedl}{Should the seed for random numbers be set (TRUE / FALSE)?\cr
default = TRUE}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}

\item{valdat}{Validation data; default = datain}
}
\value{
\describe{
\item{meanint}{Mean number of interpretable trees over the combinations of individual percentages in 'pobs' and 'ppre'}
\item{ctmax}{best resampled regression tree on the validation data set}
\item{percmax}{Best model achieved for \%observations}
\item{perfeamax}{Best model achieved for \%predictors}
\item{maxR2}{maximum R2 on the validation data set for resampled regression trees (for 'ctmax')} 
\item{minMAE}{minimum MAE (Mean Absolute Error) on the validation data set for resampled regression trees (for 'ctmax')} 
\item{interpmax}{interpretability of best tree 'ctmax'}
\item{ctmax2}{second best resampled regression tree on the full data set}
\item{percmax2}{second best model achieved for \%observations}
\item{perfeamax2}{second best model achieved for \%features}
\item{max2R2}{second best R2 on the validation data set for resampled regression trees (for 'ctmax2')}
\item{min2MAE}{second best MAE on the validation data set for resampled regression trees (for 'ctmax2')}
\item{interp2max}{interpretability of second-best tree 'ctmax2'}
}
}
\description{
The function \code{\link{PrInDTreg}} realizes regression tree optimization to identify the best interpretable tree; interpretability is checked (see 'ctestv').\cr
The relationship between the target variable 'regname' and all other factor and numerical variables
in the data frame 'datain' is optimally modeled by means of 'N' repetitions of subsampling.\cr 
The optimization criterion is the R2 of the model on the validation sample 'valdat'.\cr
Default for the validation sample is the full sample 'datain'.\cr
Multiple subsampling percentages of observations and predictors can be specified (in 'pobs' and 'ppre', correspondingly).\cr
The trees generated from subsampling can be restricted by
rejecting unacceptable trees which include split results specified in the character strings of the vector 'ctestv'.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
Besides the maximal R2, the minimal MAE (Mean Absolute Error) is reported.
}
\details{
For the optimzation of the trees, we employ a method we call Sumping (Subsampling umbrella of 
model parameters), a variant of Bumping (Bootstrap umbrella of model parameters) (Tibshirani 
& Knight, 1999). We use subsampling instead of bootstrapping. The aim of the 
optimization is to identify conditional inference trees with maximum predictive power
on the full sample under interpretability restrictions. 

\strong{Reference}\cr Tibshirani, R., Knight, K. 1999. Model Search and Inference By Bootstrap "bumping".
           Journal of Computational and Graphical Statistics, Vol. 8, No. 4 (Dec., 1999), pp. 671-686 
 
Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
frame of the function.\cr
The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
\code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
}
\examples{
data <- PrInDT::data_vowel
data <- na.omit(data)
ctestv <- 'vowel_maximum_pitch <= 320'
N <- 30 # no. of repetitions
pobs <- c(0.70,0.60) # percentages of observations
ppre <- c(0.90,0.70) # percentages of predictors
outreg <- PrInDTreg(data,"target",ctestv,N,pobs,ppre)
outreg
plot(outreg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RePrInDT.R
\name{RePrInDT}
\alias{RePrInDT}
\title{Repeated \code{\link{PrInDT}} for specified percentage combinations}
\usage{
RePrInDT(datain, classname, ctestv=NA, N, plarge, psmall, conf.level=0.95,
       thres=0.5, stratvers=0, strat=NA, seedl=TRUE,minsplit=NA,minbucket=NA,
       valdat=datain)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{N}{Number of repetitions (integer > 0)}

\item{plarge}{Vector of undersampling percentages of larger class (numerical, > 0 and <= 1)}

\item{psmall}{Vector of undersampling percentages of smaller class (numerical, > 0 and <= 1)}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{thres}{Probability threshold for prediction of smaller class (numerical, >= 0 and < 1); default = 0.5}

\item{stratvers}{Version of stratification;\cr
= 0: none (default),\cr
= 1: stratification according to the percentages of the values of the factor variable 'strat',\cr
> 1: stratification with minimum number 'stratvers' of observations per value of 'strat'}

\item{strat}{Name of one (!) stratification variable for undersampling (character);\cr
default = NA (no stratification)}

\item{seedl}{Should the seed for random numbers be set (TRUE / FALSE)?\cr
default = TRUE}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}

\item{valdat}{Validation data; default = datain}
}
\value{
\describe{
\item{treesb}{best trees for the different percentage combinations; refer to an individual tree as \code{treesb[[k]]}, k = 1, ..., length(plarge)*length(psmall)}
\item{acc1st}{accuracies of best trees on full sample}
\item{acc3en}{accuracies of ensemble of 3 best trees on full sample}
\item{simp_m}{mean of permutation losses for the predictors}
}
}
\description{
By the function \code{\link{RePrInDT}}, the function \code{\link{PrInDT}} is called repeatedly according to all combinations of the percentages specified in the vectors 'plarge' and 'psmall'.\cr
The relationship between the two-class factor variable 'classname' and all other factor and numerical variables
in the data frame 'datain' is optimally modeled by means of 'N' repetitions of undersampling.\cr 
The optimization citerion is the balanced accuracy on the validation sample 'valdat' (default = full input sample 'datain').\cr
The trees generated from undersampling can be restricted by rejecting 
unacceptable trees which include split results specified in the character strings of the vector 'ctestv'.\cr
The probability threshold 'thres' for the prediction of the smaller class may be specified (default = 0.5).\cr
Undersampling may be stratified in two ways by the feature 'strat'.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr

\strong{Reference}\cr Weihs, C., Buschfeld, S. 2021c. Repeated undersampling in PrInDT (RePrInDT): Variation in undersampling and prediction, 
and ranking of predictors in ensembles. arXiv:2108.05129
}
\details{
Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
frame of the function.\cr
The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
\code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
}
\examples{
datastrat <- PrInDT::data_zero
data <- na.omit(datastrat) # cleaned full data: no NAs
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}', 'MLU == {1, 3}')
N <- 51  # no. of repetitions
conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
psmall <- c(0.95,1)     # percentages of the small class
plarge <- c(0.09,0.1)  # percentages of the large class
outRe <- RePrInDT(data,"real",ctestv,N,plarge,psmall,conf.level) 
outRe
plot(outRe)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OptPrInDT.R
\name{OptPrInDT}
\alias{OptPrInDT}
\title{Optimisation of undersampling percentages for classification}
\usage{
OptPrInDT(data,classname,ctestv=NA,N=99,plmax=0.09,psmax=0.9,
               distl=0.01,dists=0.1,conf.level=0.95,minsplit=NA,minbucket=NA,
               valdat=data,steps=3)
}
\arguments{
\item{data}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{N}{Number (> 7) of repetitions (integer)}

\item{plmax}{Maximal undersampling percentage of larger class (numerical, > 0 and <= 1);\cr
default = 0.09}

\item{psmax}{Maximal undersampling percentage of smaller class (numerical, > 0 and <= 1);\cr
default = 0.9}

\item{distl}{Distance to the next lower undersampling percentage of larger class (numerical, > 0 and < 1);\cr
default = 0.01}

\item{dists}{Distance to the next lower undersampling percentage of smaller class (numerical, > 0 and < 1);\cr
default = 0.1}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}

\item{valdat}{validation data; default = data}

\item{steps}{number of optimization steps = 1, 2, 3; default = 3}
}
\value{
\describe{
\item{besttree}{best tree on full sample}
\item{bestba}{balanced accuracy of best tree on full sample}
\item{percl}{undersampling percentage of large class of best tree on full sample}
\item{percs}{undersampling percentage of small class of best tree on full sample}
}
}
\description{
The function \code{\link{OptPrInDT}} applies an iterative technique for finding optimal undersampling percentages 
 'percl' for the larger class and 'percs' for the smaller class by a nested grid search for the use of the function \code{\link{PrInDT}} for
the relationship between the two-class factor variable 'classname' and all other factor and numerical variables
in the data frame 'data' by means of 'N' repetitions of undersampling. The optimization citerion is the balanced accuracy 
on the validation sample 'valdat' (default = full sample 'data'). The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'.\cr
The inputs plmax and psmax determine the maximal values of the percentages and the inputs distl and dists the
the distances to the next smaller percentage to be tried.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
The parameter 'steps' controls, how many of the 3 possible optimization steps should be carried out; default=3.\cr
}
\details{
See help("RePrInDT") and help("PrInDT") for further information.

Standard output can be produced by means of \code{print(name$besttree)} or just \code{name$besttree} as well as \code{plot(name$besttree)} where 'name' is the output data 
frame of the function.\cr
}
\examples{
datastrat <- PrInDT::data_zero
data <- na.omit(datastrat) # cleaned full data: no NAs
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
# call of OptPrInDT
out <- OptPrInDT(data,"real",ctestv,N=24,conf.level=0.99,steps=1) # unstratified
out # print best model and ensembles as well as all observations
plot(out)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTregAll.R
\name{PrInDTregAll}
\alias{PrInDTregAll}
\title{Regression tree based on all observations}
\usage{
PrInDTregAll(datain,regname,ctestv=NA,conf.level=0.95,minsplit=NA,minbucket=NA)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{regname}{name of regressand variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}
}
\value{
\describe{
\item{treeall}{tree based on all observations}
\item{R2All}{goodness of fit of 'treeall' based on all observations}
\item{MAEAll}{MAE of 'treeall' based on all observations}
\item{interpAll}{criterion of interpretability of 'treeall' (TRUE / FALSE)}
}
}
\description{
Regression tree based on the full sample; interpretability is checked (see 'ctestv').\cr
The relationship between the target variable 'regname' and all other factor and numerical variables
in the data frame 'datain' is modeled based on all observations. \cr 
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
Besides the maximal R2, the minimal MAE (Mean Absolute Error) is reported.
}
\details{
Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
frame of the function.
}
\examples{
data <- PrInDT::data_vowel
data <- na.omit(data)
ctestv <- 'vowel_maximum_pitch <= 320'
outreg <- PrInDTregAll(data,"target",ctestv)
outreg
plot(outreg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PostPrInDT.R
\name{PostPrInDT}
\alias{PostPrInDT}
\title{Posterior analysis of conditional inference trees: distribution of a specified variable in the terminal nodes.}
\usage{
PostPrInDT(datain, ct, target, var, vardata, vt)
}
\arguments{
\item{datain}{input data frame with the observatios of all variables used in the model}

\item{ct}{conditional inference tree to be analyzed}

\item{target}{name of target variable of 'ct' (character)}

\item{var}{name of variable of interest (character)}

\item{vardata}{observations of 'var'}

\item{vt}{type of variables: 'dd' for discrete target (classification) and discrete variable 'var', 'dc' for discrete target (classification) and continuous 'var',
'cd' for continuous target (regression) and discrete 'var', and 'cc' for continuous target (regression) and continuous 'var'.}
}
\value{
None: Relevant output is produced by the function.
}
\description{
The conditional inference tree 'ct' is analyzed according to the distribution of a variable 'var' in its terminal nodes.\cr
In the case of a discrete variable 'var', the appearance of the different levels is considered for each terminal node. \cr
In the case of a continuous variable 'var', means and standard deviations of 'var' or the target variable are considered for each terminal node.\cr
In particular, this function can be used for the posterior analysis of a tree regarding the distribution of a variable not present in the tree.
}
\examples{
data <- PrInDT::data_zero
data <- na.omit(data)
outAll <- PrInDTAll(data,"real") 
PostPrInDT(data,outAll$treeAll,"real","ETH",data$ETH,vt="dd")
PostPrInDT(data,outAll$treeAll,"real","AGE",data$AGE,vt="dc")
datareg <- PrInDT::data_vowel
outregAll <- PrInDTregAll(datareg,"target") 
PostPrInDT(datareg,outregAll$treeAll,"target","Nickname",datareg$Nickname,vt="cd")
PostPrInDT(datareg,outregAll$treeAll,"target","AGE",datareg$AGE,vt="cc")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDT.R
\name{PrInDT}
\alias{PrInDT}
\title{The basic undersampling loop for classification}
\usage{
PrInDT(datain,classname,ctestv=NA,N,percl,percs=1,conf.level=0.95,thres=0.5,
         stratvers=0,strat=NA,seedl=TRUE,minsplit=NA,minbucket=NA,repvar=NA,indrep=0,
         valdat=datain,thr=0.5)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{N}{Number (> 2) of repetitions (integer)}

\item{percl}{Undersampling percentage of larger class (numerical, > 0 and <= 1);\cr
if percs = default, default of percl = percentage of large class resulting in the same number of observations as in the small class}

\item{percs}{Undersampling percentage of smaller class (numerical, > 0 and <= 1);\cr
default = 1}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{thres}{Probability threshold for prediction of smaller class (numerical, >= 0 and < 1); default = 0.5}

\item{stratvers}{Version of stratification;\cr
= 0: none (default),\cr
= 1: stratification according to the percentages of the values of the factor variable 'strat',\cr
> 1: stratification with minimum number "stratvers" of observations per value of "strat"}

\item{strat}{Name of one (!) stratification variable for undersampling (character);\cr
default = NA (no stratification)}

\item{seedl}{Should the seed for random numbers be set (TRUE / FALSE)?\cr
default = TRUE}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}

\item{repvar}{Values of variable defining the substructure in the case of repeated measurements, length = dim(valdat)[1] necessary; default = NA}

\item{indrep}{Indicator of repeated measurements ('indrep=1'); default = 0}

\item{valdat}{Validation data; default = datain}

\item{thr}{threshold for element classification: minimum percentage of correct class entries; default = 0.5}
}
\value{
\describe{
\item{tree1st}{best tree on validation sample}
\item{tree2nd}{2nd-best tree on validation sample}
\item{tree3rd}{3rd-best tree on validation sample}
\item{treet1st}{best tree on test sample}
\item{treet2nd}{2nd-best tree on test sample}
\item{treet3rd}{3rd-best tree on test sample}
\item{ba1st}{accuracies: largeClass, smallClass, balanced of 'tree1st', both for validation and test sample}
\item{ba2nd}{accuracies: largeClass, smallClass, balanced of 'tree2nd', both for validation and test sample}
\item{ba3rd}{accuracies: largeClass, smallClass, balanced of 'tree3rd', both for validation and test sample}
\item{baen}{accuracies: largeClass, smallClass, balanced of ensemble of all interpretable, 3 best acceptable, and all acceptable trees on validation sample}
\item{bafull}{vector of balanced accuracies of all trees from undersampling}
\item{batest}{vector of test accuracies of all trees from undersampling}
\item{treeAll}{tree based on all observations}
\item{baAll}{balanced accuracy of 'treeAll'}
\item{interpAll}{criterion of interpretability of 'treeall' (TRUE / FALSE)}
\item{confAll}{confusion matrix of 'treeAll'}
\item{acc1AE}{Accuracy of full sample tree on Elements of large class} 
\item{acc2AE}{Accuracy of full sample tree on Elements of small class }  
\item{bamaxAE}{balanced accuracy of full sample tree on Elements} 
\item{namA1}{Names of misclassified Elements by full sample tree of large class}
\item{namA2}{Names of misclassified Elements by full sample tree of small class}
\item{acc1E}{Accuracy of best tree on Elements of large class} 
\item{acc2E}{Accuracy of best tree on Elements of small class }  
\item{bamaxE}{balanced accuracy of best tree on Elements} 
\item{nam1}{Names of misclassified Elements by best tree of large class}
\item{nam2}{Names of misclassified Elements by best tree of small class}
\item{lablarge}{Label of large class} 
\item{labsmall}{Label of small class} 
\item{valdat}{Validation set after edit}
\item{thr}{Threshold for repeated measurements}
}
}
\description{
The function PrInDT uses ctrees (conditional inference trees from the package "party") for optimal modeling of
the relationship between the two-class factor variable 'classname' and all other factor and numerical variables
in the data frame 'datain' by means of 'N' repetitions of undersampling. The optimization citerion is the balanced accuracy 
on the validation sample 'valdat' (default = full input sample 'datain'). The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'.\cr
The undersampling percentages are 'percl' for the larger class and 'percs' for the smaller class (default = 1).\cr
A probability threshold 'thres' for the prediction of the smaller class may be specified (default = 0.5).\cr
Undersampling may be stratified in two ways by the option 'stratvers'.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr\cr
In the case of repeated measurements ('indrep=1'), the values of the substructure variable have to be given in 'repvar'. 
Only one value of 'classname' is allowed for each value of 'repvar'.
If for a value of 'repvar' the percentage 'thr' of the observed occurence of a value of 'classname' is not reached by the number of predictions of the value of 'classname', a misclassification is detected.
}
\details{
For the optimzation of the trees, we employ a method we call Sumping (Subsampling umbrella of 
model parameters), a variant of Bumping (Bootstrap umbrella of model parameters) (Tibshirani 
& Knight, 1999) which uses subsampling instead of bootstrapping. The aim of the 
optimization is to identify conditional inference trees with maximum predictive power
on the full sample under interpretability restrictions.

\strong{References} \cr
-- Tibshirani, R., Knight, K. 1999. Model Search and Inference By Bootstrap "bumping".
Journal of Computational and Graphical Statistics, Vol. 8, No. 4 (Dec., 1999), pp. 671-686 \cr
-- Weihs, C., Buschfeld, S. 2021a. Combining Prediction and Interpretation in  Decision Trees (PrInDT) - 
a Linguistic Example. arXiv:2103.02336

Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
frame of the function.\cr
The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
\code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
}
\examples{
datastrat <- PrInDT::data_zero
data <- na.omit(datastrat) # cleaned full data: no NAs
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
N <- 41  # no. of repetitions
conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
percl <- 0.08  # undersampling percentage of the larger class
percs <- 0.95 # undersampling percentage of the smaller class
# calls of PrInDT
out <- PrInDT(data,"real",ctestv,N,percl,percs,conf.level) # unstratified
out # print best model and ensembles as well as all observations
plot(out)
out <- PrInDT(data,"real",ctestv,N,percl,percs,conf.level,stratvers=1,
              strat="SEX") # percentage stratification
out <- PrInDT(data,"real",ctestv,N,percl,percs,conf.level,stratvers=50,
              strat="SEX") # stratification with minimum no. of tokens
out <- PrInDT(data,"real",ctestv,N,percl,percs,conf.level,thres=0.4) # threshold = 0.4

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTAllparts.R
\name{PrInDTAllparts}
\alias{PrInDTAllparts}
\title{Conditional inference trees (ctrees) based on consecutive parts of the full sample}
\usage{
PrInDTAllparts(datain, classname, ctestv=NA, conf.level=0.95, thres=0.5,
       nesvar, divt,minsplit=NA,minbucket=NA)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1); default = 0.95}

\item{thres}{Probability threshold for prediction of smaller class (numerical, >= 0 and < 1); default = 0.5}

\item{nesvar}{Name of nesting variable (character)}

\item{divt}{Number of parts of nesting variable nesvar for which models should be determined individually}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}
}
\value{
\describe{
\item{baAll}{balanced accuracy of tree on full sample}
\item{nesvar}{name of nesting variable}
\item{divt}{number of consecutive parts of the sample}
\item{badiv}{balanced accuracy of trees on 'divt' consecutive parts of the sample}
}
}
\description{
ctrees based on the full sample of the smaller class and consecutive parts of the larger class of the nesting variable 'nesvar'. 
The variable 'nesvar' has to be part of the data frame 'datain'.\cr   
Interpretability is checked (see 'ctestv'); probability threshold can be specified.
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr

\strong{Reference}\cr Weihs, C., Buschfeld, S. 2021b. NesPrInDT: Nested undersampling in PrInDT. 
arXiv:2103.14931
}
\details{
Standard output can be produced by means of \code{print(name)} or just \code{ name } where 'name' is the output data 
frame of the function.
}
\examples{
data <- PrInDT::data_speaker
data <- na.omit(data)
nesvar <- "SPEAKER"
outNesAll <- PrInDTAllparts(data,"class",ctestv=NA,conf.level=0.95,thres=0.5,nesvar,divt=8)
outNesAll

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTCstruc.R
\name{PrInDTCstruc}
\alias{PrInDTCstruc}
\title{Structured subsampling for classification}
\usage{
PrInDTCstruc(datain,classname,ctestv=NA,Struc=NA,vers="d",weight=FALSE,
                Eit=NA,Pit=NA,N=99,Ni=99,undersamp=TRUE,crit="ba",ktest=0,
                stest=integer(length=0),conf.level=0.95,indrep=0,
                minsplit=NA,minbucket=NA,repvar=NA,valdat=datain,thr=0.5)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{Struc}{= list(name,check,labs), cf. description for explanations; Struc not needed for vers="b"; for indrep=1, Struc = list(name);}

\item{vers}{Version of structured subsampling: "a", "b", "c", "d", cf. description; \cr
default = "d"}

\item{weight}{Weights to be used for subsampling of elements of substructure (logical, TRUE or FALSE);
default = FALSE}

\item{Eit}{List of number of elements of substructure (integers); \cr
default = c((Cl-4):Cl), Cl = maximum number of elements in both categories}

\item{Pit}{List of number of predictors (integers)\cr
default = c(max(1,(D-3)):D), D = maximum number of predictors}

\item{N}{Number of repetitions of subsampling from substructure (integer)\cr
default = 0 for vers="b", = 99 otherwise; if vers="b", any input is overwritten by the default}

\item{Ni}{Number of repetitions of subsampling from predictors\cr
default = 0 for vers="a", = 99 for vers="b", = N otherwise; if vers="a", any input is overwritten by the default}

\item{undersamp}{Undersampling of categories of 'classname' to be used ((logical, TRUE or FALSE)\cr
default = TRUE; if indrep=1 or vers="b", undersamp = FALSE}

\item{crit}{Optimisation criterion: "ba" for balanced accuracy, "bat" for balanced accuracy on test sets, "ta" for test accuracy, 
"tal" for test accuracy of continuing parts of length 'ktest' in substructure elements 'stest';\cr
default = "ba"}

\item{ktest}{Length of continuing parts to be tested (for crit="tal");\cr
default = 0}

\item{stest}{Part of substructure to be tested (for crit="tal")(integer vector);\cr
default = integer vector of length 0}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{indrep}{Indicator for repeated measurements, i.e. more than one observation with the same class for each element;\cr 
for indrep=1, Struc=list(name) only; default = 0}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}

\item{repvar}{Values of variable defining the substructure corresponding to valdat in the case of repeated measurements, length = dim(valdat)[1] necessary; default = NA}

\item{valdat}{validation data; default = datain}

\item{thr}{threshold for element classfication: minimum percentage of correct class entries; default = 0.5}
}
\value{
\describe{
  \item{modbest}{Best tree} 
  \item{interp}{Number of interpretable trees, overall number of trees} 
  \item{dmax}{Number of predictors in training set for best tree} 
  \item{ntmax}{Size of training set for best tree} 
  \item{acc1}{Accuracy of best tree on large class} 
  \item{acc2}{Accuracy of best tree on small class }  
  \item{bamax}{Balanced accuracy of best tree} 
  \item{tamax}{Test accuracy of best tree}
  \item{kumin}{Number of elements with misclassified parts longer than 'ktest' for best tree}
  \item{elems}{Elements with long misclassified parts for best tree}
  \item{mindlong}{Indices of long misclassified parts for best tree}
  \item{ind1max}{Elements of 1st category of substructure used by best tree} 
  \item{ind2max}{Elements of 2nd category of substructure used by best tree} 
  \item{indmax}{Predictors used by best tree} 
  \item{bestTrain}{Training set for best tree} 
  \item{bestTest}{Test set for best tree} 
  \item{labs}{labs from Struc} 
  \item{lablarge}{Label of large class} 
  \item{labsmall}{Label of small class} 
  \item{vers}{Version used for structured subsampling} 
  \item{acc1E}{Accuracy of large class on Elements of best tree} 
  \item{acc2E}{Accuracy of small class on Elements of best tree}  
  \item{bamaxE}{Balanced accuracy of best tree on Elements} 
  \item{nam1}{Names of misclassified Elements of large class}
  \item{nam2}{Names of misclassified Elements of small class}
  \item{thr}{Threshold for element classification}
}
}
\description{
The function PrInDTCstruc applies structured subsampling for finding an optimal subsample to model
the relationship between the two-class factor variable 'classname' and all other factor and numerical variables
in the data frame 'datain' by means of 'N' repetitions of subsampling from a substructure and 'Ni' repetitions of subsampling from the predictors.  
The optimization citerion is the balanced accuracy on the validation sample 'valdat' (default is the full input sample 'datain'). 
Other criteria are possible (cf. parameter description of 'crit'). The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'.\cr
The substructure of the observations used for subsampling in modelling is specified by the list 'Struc' which consists of the factor variable 'name' representing the substructure,
the name 'check' of the factor variable with the information about the categories of the substructure, and the matrix 'labs' which specifies the values of 'check'
corresponding to two categories in its rows, i.e. in 'labs[1,]' and 'labs[2,]'. The names of the categories have to be specified by \code{rownames(labs)}.\cr
See parameter description of 'Struc' for its specification for 'vers="b"' and 'indrep > 0'.\cr
The number of predictors 'Pit' to be included in the model and the number of elements of the substructure 'Eit' have to be specified (lists allowed), and 
undersampling of the categories of 'classname' can be controlled by 'undersamp=TRUE/FALSE'.\cr
In structured subsampling, 'N' repetitions of subsampling of the variable 'name' with 'Eit' different elements of each category in 'check' are realized. If 'Eit' is a list, each entry is employed individually. If 'Eit' is larger than the maximum available number of elements with a certain value of 'check', the maximum possible number of elements is used.\cr
Four different versions of structured subsampling exist: \cr
a) just of the elements in the substructure (possibly with additional undersampling) with parameters 'N' and 'Eit',\cr
b) just of the predictors with parameters 'Ni' and 'Pit', \cr
c) of the predictors and for each subset of predictors subsampling of the elements of the substructure (possibly with additional undersampling) 
with parameters 'N', 'Ni', 'Eit', and 'Pit', and\cr
d) of the elements of the substructure (possibly with additional undersampling) and for each of these subsets subsampling of the predictors 
with the same parameters as version c).\cr
Sampling of the elements of the substructure can be influenced by using weights of the elements ('weights=TRUE') according to the number of appearances of the smaller 
class of 'classnames'. This way, elements with more realisations in the smaller class are preferred in modelling.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr\cr
The parameter 'indrep' indicates repeated measurement situations ('indrep > 0', only implemented for 'crit="ba"'); default = 0.\cr
Repeated measurements are multiple measurements of the same variable taken on the same subjects (or objects) either under different conditions or over two or more time periods.\cr
The variable with the repeatedly observed subjects (or objects) has to be specified by 'name' in 'Struc'. Only one value of 'classname' is allowed for each value of 'Struc$name'.
In case of 'indrep > 0' it is automatically assumed that the same number of 
subjects (or objects) of the two classes under study have to be used for model building. Possible such numbers can be specified by 'Eit'.\cr
For indrep=1, models are optimized for individual observations and the classes of the substructure elements are subsequently determined via the parameter 'thr'.\cr
For indrep=2, models are optimized so that the criterion "ba" is maximal for the substructure elements taking 'thr' into account.\cr
For indrep > 0 and 'valdat' unequal 'datain', the variable 'repvar' defines the substructure for 'valdat', length = dim(valdat)[1] necessary; default = NA'
}
\details{
See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.1, for further information.

Standard output can be produced by means of \code{print(name$besttree)} or just \code{name$besttree} as well as \code{plot(name$besttree)} where 'name' is the output data 
frame of the function.\cr
}
\examples{
data <- PrInDT::data_zero
data <- na.omit(data) # cleaned full data: no NAs
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
# substructure
name <- PrInDT::participant_zero
check <- "data$ETH"
labs <- matrix(1:6,nrow=2,ncol=3)
labs[1,] <- c("C1a","C1b","C1c")
labs[2,] <- c("C2a","C2b","C2c")
rownames(labs) <- c("Children 1","Children 2")
Struc <- list(name=name,check=check,labs=labs)
out <- PrInDTCstruc(data,"real",ctestv,Struc,vers="c",weight=TRUE,N=5,Pit=5,conf.level=0.99)
out
plot(out)
# indrep = 1
Struc <- list(name=name)
out <- PrInDTCstruc(data,"real",Struc=Struc,vers="c",Pit=5,Eit=5,N=5,crit="ba",indrep=1,
           conf.level=0.99)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{data_land}
\alias{data_land}
\title{Landscape analysis}
\format{
A data frame with 149 observations and 28 columns
\describe{
  \item{coder}{who coded the sign: Factor (3 levels "C","E","S")  (anonymized)}
  \item{researcher}{who took a photograph of the sign: Factor (2 levels "L","S")  (anonymized)}
  \item{sign}{where the sign was found: Factor (11 levels "digi","door","graf",...)}
  \item{type.of.sign}{kind of sign: Factor (5 levels "com","commem","infra","reg","trans")}
  \item{permanent}{was the sign permanent? Factor (2 levels "no","yes")}
  \item{proper.noun}{kind of proper noun on the sign: Factor (9 levels "bn","bn+","cn",...)}
  \item{no.languages}{number of languages on sign: Factor (4 levels "1","2","3","4+")}
  \item{French}{French on sign? Factor (2 levels "0","1")}
  \item{Dutch}{Dutch on sign? Factor (2 levels "0","1")}
  \item{English}{English on sign? Factor (2 levels "0","1")}
  \item{Italian}{Italian on sign? Factor (2 levels "0","1")}
  \item{Spanish}{Spanish on sign? Factor (2 levels "0","1")}
  \item{German}{German on sign? Factor (2 levels "0","1")}
  \item{Indian,Mandarin,Greek,Indonesian,Portuguese,Libanese,Japanese,Russian,Danish,
            Norwegian,Hebrew,Catalan}{12 infrequent languages: Factor (2 levels "0","1")}
  \item{bn.unclear}{brand name unclear: Factor (2 levels "0","1")}
  \item{multilingual.type}{type of multilingualism on sign: Factor (6 levels "0","1","2","3","4","uc")}
  \item{location}{location of sign: Factor (2 levels "M","P")  (anonymized)}
}
}
\source{
Sarah Buschfeld, TU Dortmund
}
\usage{
data_land
}
\description{
The use of language(s) on public signs and categorization criteria.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{data_speaker}
\alias{data_speaker}
\title{Subject pronouns and a predictor with one very frequent level}
\format{
A data frame with 3370 observations and 6 columns
\describe{
  \item{class}{subject pronoun realized? Factor (2 levels "zero","realized")}
  \item{AGE}{age: Numerical (in months)}
  \item{ETHN_GROUP}{ethnic group: Factor (3 levels "C","I","n_a") (anonymized)}
  \item{MLU}{mean length of utterance: Factor (5 levels "1","2","3","adult","OL")}
  \item{PRN_TYPE}{pronoun type: Factor (5 levels "dem","it_con","it_ex","it_ref","refer")}
  \item{SPEAKER}{speaker: Factor (2 levels "adult","child")}
}
}
\source{
Sarah Buschfeld, TU Dortmund
}
\usage{
data_speaker
}
\description{
Usage of subject pronouns and its predictors; speaker level "adult" very frequent.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Mix2SPrInDT.R
\name{Mix2SPrInDT}
\alias{Mix2SPrInDT}
\title{Two-stage estimation for classification-regression mixtures}
\usage{
Mix2SPrInDT(datalist,ctestv=NA,N=99,percent=NA,conf.level=0.95,minsplit=NA,minbucket=NA)
}
\arguments{
\item{datalist}{list(datanames,targets,datastruc,summ) Input data: For specification see the above description}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{N}{Number of repetitions of subsampling for predictors (integer); default = 99}

\item{percent}{matrix of percent spefications: For specification see the above description; default: 'percent = NA' meaning default values for percentages.}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1); default = 0.95}

\item{minsplit}{Minimum number of elements in a node to be splitted; default = 20}

\item{minbucket}{Minimum number of elements in a node; default = 7}
}
\value{
\describe{
  \item{models1}{Best trees at stage 1} 
  \item{models2}{Best trees at stage 2} 
  \item{depnames}{names of dependent variables}
  \item{nmod}{number of models of tasks}
  \item{nlev}{levels of tasks}
  \item{accAll}{accuracies of best trees at both stages}  
}
}
\description{
The function Mix2SPrInDT applies 'N' repetitions of subsampling for finding an optimal subsample to model
the relationship between the dependent variables specified in the sublist 'targets' of the list 'datalist' and all other factor and numerical variables
in the corresponding data frame specified in the sublist 'datanames' of the list 'datalist' in the same order as 'targets'. \cr
The function is prepared to handle classification tasks with 2 or more classes and regression tasks. At first stage, the targets are estimated only on
the basis of the exogenous predictors. At second stage, summaries of the predictions of the endogenous features from the first stage models are added as new predictors.\cr
Subsampling of observations and predictors uses the percentages specified in the matrix 'percent', one row per estimation taks. For classification tasks, 
percentages 'percl' and 'percs' for the larger and the smaller class have to be specified. For regression tasks, 'pobs' and 'ppre' have to specified 
for observations and predictors, respectively.\cr
For generating summaries, the variables representing the substructure have to be specified in the sublist 'datastruc' of 'datalist'.\cr
The sublist 'summ' of 'datalist' includes for each discrete target the classes for which you want to calculate the summary percentages and for each 
continuous target just NA (in the same order as in the sublist 'targets'). For a discrete target in this list, you can provide a sublist of classes to be combined 
(for an example see the below example).\cr
The optimization citerion is the balanced accuracy for classification and R2 for regression, both evaluated on the full sample.  \cr
The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
}
\details{
See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.1, for further information.

Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name} where 'name' is the output data 
frame of the function.\cr
}
\examples{
# zero data
datazero <- PrInDT::data_zero
datazero <- na.omit(datazero) # cleaned full data: no NAs
names(datazero)[names(datazero)=="real"] <- "zero"
CHILDzero <- PrInDT::participant_zero
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
##
# multi-level data
datastrat <- PrInDT::data_zero
datamult <- na.omit(datastrat)
# ctestv <- NA
datamult$mult[datamult$ETH \%in\% c("C1a","C1b","C1c") & datamult$real == "zero"] <- "zero1"
datamult$mult[datamult$ETH \%in\% c("C2a","C2b","C2c") & datamult$real == "zero"] <- "zero2"
datamult$mult[datamult$real == "realized"] <- "real"
datamult$mult <- as.factor(datamult$mult) # mult is new class variable
datamult$real <- NULL # remove old class variable
CHILDmult <- CHILDzero
##
# vowel data
data <- PrInDT::data_vowel
data <- na.omit(data)
CHILDvowel <- data$Nickname
data$Nickname <- NULL
syllable <- 3 - data$syllables
data$syllabels <- NULL
data$syllables <- syllable
data$speed <- data$word_duration / data$syllables
names(data)[names(data) == "target"] <- "vowel"
datavowel <- data
##
# function preparation and call
datanames <- list("datazero","datamult","datavowel")
targets <- c("zero","mult","vowel")
datastruc <- list(CHILDzero,CHILDmult,CHILDvowel)
summult <- paste("zero1","zero2",sep=",")  
summ <- c("zero",summult,NA)
datalist <- list(datanames=datanames,targets=targets,datastruc=datastruc,summ=summ)
percent <- matrix(NA,nrow=3,ncol=2)
percent[1,] <- c("percl=0.075","percs=0.9") # percentages for datazero
# no percentages needed for datapast
percent[3,] <- c("pobs=0.9","ppre=c(0.9,0.8)") # percentages for datavowel
out2SMix <- Mix2SPrInDT(datalist,ctestv=ctestv,N=19,percent=percent,conf.level=0.99)
out2SMix
plot(out2SMix)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTMulev.R
\name{PrInDTMulev}
\alias{PrInDTMulev}
\title{PrInDT analysis for a classification problem with multiple classes.}
\usage{
PrInDTMulev(datain,classname,ctestv=NA,N,conf.level=0.95,seedl=FALSE,
                                    minsplit=NA,minbucket=NA)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{N}{Number of repetitions (integer > 0)}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1)\cr
(default = 0.95)}

\item{seedl}{Should the seed for random numbers be set (TRUE / FALSE)?\cr
default = FALSE}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}
}
\value{
\describe{
\item{class}{levels of class variable} 
\item{trees}{trees for the levels of the class variable; refer to an individual tree as \code{trees[[k]]}, k = 1, ..., no. of levels}
\item{ba}{balanced accuracy of combined predictions} 
\item{conf}{confusion matrix of combined predictions} 
\item{ninterp}{no. of non-interpretable trees} 
\item{acc}{balanced accuracies of best models for individual classes}
}
}
\description{
PrInDT analysis for a classification problem with more than 2 classes. For each combination of one class vs. 
the other classes a 2-class \code{\link{PrInDT}} analysis is carried out.\cr
The percentages for undersampling of the larger class ('percl' in \code{\link{PrInDT}}) are chosen so that the resulting sizes
are comparable with the size of the smaller classes for which all their observations are used in undersampling ('percs' = 1 in \code{\link{PrInDT}}).\cr
The class with the highest probability in the K (= number of classes) analyses is chosen for prediction.\cr
Interpretability is checked (see 'ctestv').
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
}
\details{
Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)}  where 'name' is the output data 
frame of the function.\cr
The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
\code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
}
\examples{
datastrat <- PrInDT::data_zero
data <- na.omit(datastrat)
ctestv <- NA
data$rel[data$ETH \%in\% c("C1a","C1b","C1c") & data$real == "zero"] <- "zero1"
data$rel[data$ETH \%in\% c("C2a","C2b","C2c") & data$real == "zero"] <- "zero2"
data$rel[data$real == "realized"] <- "real"
data$rel <- as.factor(data$rel) # rel is new class variable
data$real <- NULL # remove old class variable
N <- 51
conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
out <- PrInDTMulev(data,"rel",ctestv,N,conf.level) 
out # print best models based on subsamples
plot(out) # corresponding plots

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{participant_zero}
\alias{participant_zero}
\title{Participants of subject pronoun study}
\format{
A dataset with 1024 observations and 1 column
\describe{
  \item{participant}{participant of study (P1 - P51)}
}
}
\source{
Sarah Buschfeld, TU Dortmund
}
\usage{
participant_zero
}
\description{
Participants (random excerpt according to data_zero)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{data_vowel}
\alias{data_vowel}
\title{Vowel length}
\format{
A data frame with 82 observations and 22 columns
\describe{
  \item{Nickname}{nickname: Factor (43 levels "Nick1","Nick2",...,"Nick43") (anonymized)}
  \item{LiBa}{linguistic background: Factor (2 levels "mono","multi")}
  \item{MLU}{mean length of utterance: Factor (3 levels "1","2","3")}
  \item{phone_label}{phone label: Factor (2 levels "fleece","kit")}
  \item{lexeme}{lexeme: Factor (14 levels "bee","cheek","cheese","chicken", ...)}
  \item{phone_left_1_duration}{duration of phone to the left of the vowel: Numerical (in msec)}
  \item{phone_right_1_duration}{duration of phone to the right of the vowel: Numerical (in msec)}
  \item{word_duration}{duration of word: Numerical (in msec)}
  \item{vowel_minimum_pitch}{minimum pitch of vowel: Numerical (in Hertz)}
  \item{vowel_maximum_pitch}{maximum pitch of vowel: Numerical (in Hertz)}
  \item{vowel_intensity_mean}{mean intensity of vowel: Numerical (in decibel)}
  \item{f1_fifty}{first formant F1 at midpoint of vowel (50\%): Numerical (in Hertz)}
  \item{f2_fifty}{second formant F2 at midpoint of vowel (50\%): numercial (in Hertz)}
  \item{target}{vowel length: Numerical (in msec)}
  \item{cons_class_l}{class of consonant to the left of the vowel: Factor (6 levels "l","r","tsh",...)}
  \item{cons_class_r}{class of consonant to the right of the vowel: Factor (7 levels "?"(glottal stop),"empty","nas",...)}
  \item{ETH}{ethnic group: Factor (6 levels "C1a","C1b","C1c","C2a","C2b","C2c") (anonymized)}
  \item{SEX}{gender: Factor (2 levels "female","male)}
  \item{AGE}{age: Numerical (in months)}
  \item{syllables}{number of syllables in lexeme: integer (1,2)}
  \item{speed}{speed of speech: Numerical (word duration / syllables; in msec)}
  \item{country}{country: Factor (2 levels "E","S")  (anonymized)}
}
}
\source{
Sarah Buschfeld, TU Dortmund
}
\usage{
data_vowel
}
\description{
Vowel length and categorization criteria.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTAll.R
\name{PrInDTAll}
\alias{PrInDTAll}
\title{Conditional inference tree (ctree) based on all observations}
\usage{
PrInDTAll(datain, classname, ctestv=NA, conf.level=0.95, thres=0.5,
                 minsplit=NA,minbucket=NA,repvar=NA,indrep=0,thr=0.5)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1); default = 0.95}

\item{thres}{Probability threshold for prediction of smaller class (numerical, >= 0 and < 1); default = 0.5}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}

\item{repvar}{Values of variable defining the substructure in the case of repeated measurements, length = dim(datain)[1] necessary; default=NA}

\item{indrep}{Indicator of repeated measurements ('indrep=1'); default = 0}

\item{thr}{threshold for element classification: minimum percentage of correct class entries; default = 0.5}
}
\value{
\describe{
\item{treeall}{ctree based on all observations}
\item{baAll}{balanced accuracy of 'treeall'}
\item{interpAll}{criterion of interpretability of 'treeall' (TRUE / FALSE)}
\item{confAll}{confusion matrix of 'treeall'}
\item{acc1AE}{Accuracy of full sample tree on Elements of large class} 
\item{acc2AE}{Accuracy of full sample tree on Elements of small class}  
\item{bamaxAE}{balanced accuracy of full sample tree on Elements} 
\item{namA1}{Names of misclassified Elements by full sample tree of large class}
\item{namA2}{Names of misclassified Elements by full sample tree of small class}
\item{lablarge}{Label of large class} 
\item{labsmall}{Label of small class}
\item{thr}{Threshold for repeated measurements} 
}
}
\description{
ctree based on all observations in 'datain'. Interpretability is checked (see 'ctestv'); probability threshold can be specified.
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.

\strong{Reference} \cr
Weihs, C., Buschfeld, S. 2021a. Combining Prediction and Interpretation in  Decision Trees (PrInDT) - 
a Linguistic Example. arXiv:2103.02336

In the case of repeated measurements ('indrep=1'), the values of the substructure variable have to be given in 'repvar'.
Only one value of 'classname' is allowed for each value of 'repvar'.
If for a value of 'repvar' the percentage 'thr' of the observed occurence of a value of 'classname' is not reached by the number of predictions of the value of 'classname', a misclassification is detected.
}
\details{
Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
frame of the function.\cr
}
\examples{
datastrat <- PrInDT::data_zero
data <- na.omit(datastrat)
ctestv <- rbind('ETH == {C2a,C1a}','MLU == {1, 3}')
conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
outAll <- PrInDTAll(data,"real",ctestv,conf.level) 
print(outAll) # print model based on all observations
plot(outAll) # plot model based on all observations

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/R2SPrInDT.R
\name{R2SPrInDT}
\alias{R2SPrInDT}
\title{Two-stage estimation for regression}
\usage{
R2SPrInDT(data,ctestv=NA,inddep,N=99,pobs1=c(0.90,0.70),ppre1=c(0.90,0.70),
                pobs2=pobs1,ppre2=ppre1,conf.level=0.95,minsplit=NA,minbucket=NA)
}
\arguments{
\item{data}{Input data frame with continuous target variable 'regname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{inddep}{Column indices of target variables in datain}

\item{N}{Number of repetitions of subsampling from predictors (integer) in versions "b" and "c";\cr
default = 99}

\item{pobs1}{Percentage(s) of observations for  subsampling at stage 1;\cr
default=c(0.9,0.7)}

\item{ppre1}{Percentage(s) of predictors for subsampling at stage 1;\cr
default=c(0.9,0.7)}

\item{pobs2}{Percentage(s) of observations for  subsampling at stage 2";\cr
default=pobs1}

\item{ppre2}{Percentage(s) of predictors for subsampling at stage 2;\cr
default=ppre1}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}
}
\value{
\describe{
  \item{models1}{Best trees at stage 1} 
  \item{models2}{Best trees at stage 2} 
  \item{depnames}{names of dependent variables}
  \item{R2both}{R2s of best trees at both stages}  
}
}
\description{
The function R2SPrInDT applies 'N' repetitions of subsampling for finding an optimal subsample to model
the relationship between the continuous variables with indices 'inddep' and all other factor and numerical variables
in the data frame 'datain'. \cr
Subsampling of observations and predictors uses the percentages in 'pobs1' and 'ppre1', respectively, at stage 1, and the percentages 'pobs2' and 'ppre2' 
at stage 2, accordingly.
The optimization criterion is the goodness of fit R2 on the full sample.  \cr
The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
}
\details{
See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.1, for further information.

Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name} where 'name' is the output data 
frame of the function.\cr
}
\examples{
data <- PrInDT::data_vowel
data <- na.omit(data)
CHILDvowel <- data$Nickname
data$Nickname <- NULL
syllable <- 3 - data$syllables
data$syllabels <- NULL
data$syllables <- syllable
data$speed <- data$word_duration / data$syllables
names(data)[names(data) == "target"] <- "vowel_length"
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
inddep <- c(13,9) 
out2SR <- R2SPrInDT(data,ctestv=ctestv,inddep=inddep,N=9,conf.level=0.99)
out2SR
plot(out2SR)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTMulab.R
\name{PrInDTMulab}
\alias{PrInDTMulab}
\title{Multiple label classification based on resampling by \code{\link{PrInDT}}}
\usage{
PrInDTMulab(datain,classnames=NA,ctestv=NA,conf.level=0.95,percl,percs=1,N,
               indind=NA,indaddind=NA,inddep,minsplit=NA,minbucket=NA)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classnames}{names of class variables (character vector); default = NA: from old version: superfluous now}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{percl}{list of undersampling percentages of larger class (numerical, > 0 and <= 1): one per dependent class variable}

\item{percs}{list of undersampling percentage of smaller class (numerical, > 0 and <= 1); one per dependent class variable}

\item{N}{no. of repetitions (integer > 0)}

\item{indind}{indices of independent variables}

\item{indaddind}{indices of additional independent variables used in the case of dependent binary relevance modeling}

\item{inddep}{indices of dependent variables}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}
}
\value{
\describe{
\item{accbr}{model errors for Binary Relevance (single assessment) - only independent predictors are used for modeling one label at a time,
the other labels are not used as predictors. As the performance measure for the resulting classification rules,
the balanced accuracy of the best model from PrInDT is employed for each individual label.}
\item{errbin}{combined error for Binary Relevance (joint assessment) - the best prediction models for the different labels are combined  
to assess the combined prediction. The 01-accuracy counts a label combination as correct only if all labels are correctly predicted.
The hamming accuracy corresponds to the proportion of labels whose value is correctly predicted.}
\item{accdbr}{model errors for Dependent Binary Relevance (Extended Model) (single assessment) - each label is trained by means of an extended model which 
not only includes the independent predictors but also the other labels. For these labels, the truly observed values are used for 
estimation and prediction. In the extended model, other labels, which are not treated as dependent variables, can also be used as additional predictors.}
\item{errext}{combined errors for Dependent Binary Relevance (Extended Model) (joint assessment)}
\item{errtrue}{combined errors for Dependent Binary Relevance (True Prediction) - in the prediction phase, the values 
of all modeled labels are first predicted by the independent predictors only and then the predicted labels are used in the estimated  
extended model in a 2nd step to 
ultimately predict the labels.}
\item{coldata}{column names of input data}
\item{inddep}{indices of dependent variables (labels to be modeled)}
\item{treebr}{list of trees from Binary Relevance modeling, one tree for each label; refer to an individual tree as \code{treebr[[i]]}, 
                i = 1, ..., no. of labels}
\item{treedbr}{list of trees from Dependent Binary Relevance modeling, one for each label; refer to an individual tree as \code{treedbr[[i]]}, 
                i = 1, ..., no. of labels}
}
}
\description{
Multiple label classification based on resampling by \code{\link{PrInDT}}. We consider two ways of modeling (Binary relevance modeling, 
dependent binary modeling) and three ways of model evaluation: single assessment, joint assessment, and true prediction  
(see the Value section for more information).\cr
Variables should be arranged in 'datain' according to indices specified in 'indind', 'indaddind', and 'inddep'.\cr
Please note that the dependent variables have to be specified as dummies, i.e. as 'absent' (value 0) or 'present' (value 1).\cr
Undersampling is repeated 'N' times.\cr
Undersampling percentages 'percl' for the larger class and 'percs' for the smaller class can be 
specified, one each per dependent class variable.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr

\strong{References}\cr 
- Buschfeld, S., Weihs, C. and Ronan, P.. 2024. Modeling linguistic landscapes: A comparison of St Martin’s two capitals Philipsburg and Marigot Linguistic Landscape 10(3): 302–334. https://doi.org/10.1075/ll.23070.bus\cr
- Probst, P., Au, Q., Casalicchio, G., Stachl, C., and Bischl, B. 2017. Multilabel Classification with R Package mlr. arXiv:1703.08991v2
}
\details{
Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)}  where 'name' is the output data 
frame of the function.\cr
The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
\code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
}
\examples{
data <- PrInDT::data_land # load data
dataclean <- data[,c(1:7,23:24,11:13,22,8:10)]  # only relevant features
indind <- c(1:9) # original predictors
indaddind <- c(10:13) # additional predictors
inddep <- c(14:16) # dependent variables
dataclean <- na.omit(dataclean)
ctestv <- NA
N <- 21  # no. of repetitions
perc <- c(0.45,0.05,0.25)   # percentages of observations of larger class, 
#                             1 per dependent class variable
perc2 <- c(0.75,0.95,0.75)  # percentages of observations of smaller class, 
#                             1 per dependent class variable
##
# Call PrInDT: language by language
##
outmult <- PrInDTMulab(dataclean,ctestv=NA,conf.level=0.95,percl=perc,percs=perc2,N=N,
                  indind=indind,indaddind=indaddind,inddep=inddep)
print(outmult)
plot(outmult)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SimMixPrInDT.R
\name{SimMixPrInDT}
\alias{SimMixPrInDT}
\title{Interdependent estimation for classification-regression mixtures}
\usage{
SimMixPrInDT(datalist,ctestv=NA,Struc,M=12,N=99,nsub,percent=NA,conf.level=0.95,
                                      minsplit=NA,minbucket=NA)
}
\arguments{
\item{datalist}{list(datanames,targets,datastruc,summ) Input data: For specification see the above description}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{Struc}{list(name,check,labs) Paprametes for structured subsampling, as explained in the desciption above.}

\item{M}{Number of repetitions of subsampling of elements of substructure; default = 12}

\item{N}{Number of repetitions of subsampling for predictors (integer); default = 99}

\item{nsub}{(List of) numbers of different elements of substructure per subsample}

\item{percent}{matrix of percent spefications: For specification see the above description; default: 'percent = NA' meaning default values for percentages.}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1); default = 0.95}

\item{minsplit}{Minimum number of elements in a node to be splitted; default = 20}

\item{minbucket}{Minimum number of elements in a node; default = 7}
}
\value{
\describe{
  \item{modelsF}{Best trees at stage 1 (Full sample)} 
  \item{modelsI}{Best trees at stage 2 (Individual optimization)} 
  \item{modelsJ}{Best trees at stage 3 (Joint optimization)} 
  \item{depnames}{names of dependent variables}
  \item{nmod}{number of models of tasks}
  \item{nlev}{levels of tasks}
  \item{accAll}{accuracies of best trees at both stages}  
}
}
\description{
The function SimMixPrInDT applies structured subsampling for finding an optimal subsample to model
the relationship between the dependent variables specified in the sublist 'targets' of the list 'datalist' and all other factor and numerical variables
in the corresponding individual data frame specified in the sublist 'datanames' of the list 'datalist' in the same order as 'targets'. The data frames have to be different of each other! \cr
The function is prepared to handle classification tasks with 2 or more classes and regression tasks. At first stage, the targets are estimated based on
the full sample of all exogenous variables and on summaries of the observed endogenous variables.\cr
For generating summaries, the variables representing the substructure have to be specified in the sublist 'datastruc' of 'datalist'.\cr
The sublist 'summ' of 'datalist' includes for each discrete target the classes for which you want to calculate the summary percentages and for each 
continuous target just NA (in the same order as in the sublist 'targets'). For a discrete target in this list, you can provide a sublist of classes to be combined 
(for an example see the below example).\cr
At second stage, structured subsampling is used for improving the models from first stage.
The substructure of the observations used for structured subsampling is specified by the list 'Struc' which here only consists of 
the name 'check' of the variable with the information about the categories of the substructure (without specification of the dataset names already specified 
in 'datanames', see example below), and the matrix 'labs' which specifies the values of 'check' corresponding to two categories in its rows, i.e. in 'labs[1,]' and 'labs[2,]'. 
The names of the categories have to be specified by  \code{rownames(labs)}.\cr
In structured subsampling, first 'M' repetitions of subsampling of the variable 'name' with 'nsub' different elements of each category in 'check' are realized. If 'nsub' is a list, each entry is employed individually. If 'nsub' is larger than the maximum available number of elements with a certain value of 'check', the maximum possible number of elements is used. Then,
for each of the subsamples 'N' repetitions of subsampling in classification or regression with the specified percentages of classes, observations, and predictors are carried out.\cr 
These percentages are specified in the matrix 'percent', one row per estimation task. For binary classification tasks, 
percentages 'percl' and 'percs' for the larger and the smaller class have to be specified. For multilevel classification tasks, NA is specified (see the below example) 
since the percentages are generated automatically. For regression tasks, 'pobs' and 'ppre' have to be specified for observations and predictors, respectively.\cr
The optimization citerion is balanced accuracy for classification and goodness of fit R2 for regression on the full sample, respectively.  
At stage 2, the models are optimized individually. At stage 3, the models are optimized on the maximum of joint elements in their substructures.\cr
The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
}
\details{
See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.2, for further information.

Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name} where 'name' is the output data 
frame of the function.\cr
}
\examples{
# zero data
datazero <- PrInDT::data_zero
datazero <- na.omit(datazero) # cleaned full data: no NAs
names(datazero)[names(datazero)=="real"] <- "zero"
CHILDzero <- PrInDT::participant_zero
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
##
# multi-level data
datastrat <- PrInDT::data_zero
datamult <- na.omit(datastrat)
# ctestv <- NA
datamult$mult[datamult$ETH \%in\% c("C1a","C1b","C1c") & datamult$real == "zero"] <- "zero1"
datamult$mult[datamult$ETH \%in\% c("C2a","C2b","C2c") & datamult$real == "zero"] <- "zero2"
datamult$mult[datamult$real == "realized"] <- "real"
datamult$mult <- as.factor(datamult$mult) # mult is new class variable
datamult$real <- NULL # remove old class variable
CHILDmult <- CHILDzero
##
# vowel data
data <- PrInDT::data_vowel
data <- na.omit(data)
CHILDvowel <- data$Nickname
CHILDvowel <- as.factor(gsub("Nick","P",CHILDvowel))
data$Nickname <- NULL
syllable <- 3 - data$syllables
data$syllabels <- NULL
data$syllables <- syllable
data$speed <- data$word_duration / data$syllables
names(data)[names(data) == "target"] <- "vowel"
datavowel <- data
##
# function preparation and call
# datalist and percent
datanames <- list("datazero","datamult","datavowel")
targets <- c("zero","mult","vowel")
datastruc <- list(CHILDzero,CHILDmult,CHILDvowel)
summult <- paste("zero1","zero2",sep=",")  
summ <- c("zero",summult,NA)
datalist <- list(datanames=datanames,targets=targets,datastruc=datastruc,summ=summ)
percent <- matrix(NA,nrow=3,ncol=2)
percent[1,] <- c("percl=0.075","percs=0.9") # percentages for datazero
# no percentages needed for datapast
percent[3,] <- c("pobs=0.9","ppre=c(0.9,0.8)") # percentages for datavowel
# substructures
labs <- matrix(1:6,nrow=2,ncol=3)
labs[1,] <- c("C1a","C1b","C1c")
labs[2,] <- c("C2a","C2b","C2c")
rownames(labs) <- c("children 1","children 2")
Struc <- list(check="ETH",labs=labs)
outSimMix <- SimMixPrInDT(datalist,ctestv=ctestv,Struc=Struc,M=2,N=9,nsub=c(19,20),
                       percent=percent,conf.level=0.99)
outSimMix
plot(outSimMix)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/C2SPrInDT.R
\name{C2SPrInDT}
\alias{C2SPrInDT}
\title{Two-stage estimation for classification}
\usage{
C2SPrInDT(datain,ctestv=NA,conf.level=0.95,percl,percs,N=99,indind=NA,inddep,
                                 minsplit=NA,minbucket=NA)
}
\arguments{
\item{datain}{Input data frame with class factor variables and the influential variables,\cr
which need to be factors or numericals (transform logicals and character variables to factors)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{percl}{list of undersampling percentages of larger class (numerical, > 0 and <= 1): one per dependent class variable in the same order as in 'inddep'}

\item{percs}{list of undersampling percentage of smaller class (numerical, > 0 and <= 1); one per dependent class variable in the same order as in 'inddep'}

\item{N}{no. of repetitions (integer > 0); default = 99}

\item{indind}{indices of independent variables at stage 1; default = NA (means all independent variables used)}

\item{inddep}{indices of dependent variables}

\item{minsplit}{Minimum number of elements in a node to be splitted; default = 20}

\item{minbucket}{Minimum number of elements in a node; default = 7}
}
\value{
\describe{
  \item{models1}{Best trees at stage 1} 
  \item{models2}{Best trees at stage 2} 
  \item{classnames}{names of classification variables}
  \item{baAll}{balanced accuracies of best trees at both stages} 
}
}
\description{
The function C2SPrInDT applies two-stage estimation for finding an optimal model for relationships between the two-class factor variables
specified as column indices of 'datain' in the vector 'inddep' and all other factor and numerical variables in the data frame 'datain' by means of
'N' repetitions of random subsampling with percentages 'percl' for the large classes and 'percs' for the small classes. One percentage of observations 
for each dependent variable has to be specified for the larger and the smaller class. For example, for three dependent variables, 'percl' consists of 
three percentages specified in the order in which the dependent variables appear in 'inddep'.\cr
The dependent variables have to be specified as dummies, i.e. as 0 for 'property absent' or 1 for 'property present' for a certain property the dependent variable is representing.\cr
The indices of the predictors relevant at 1st stage modeling can be specified in the vector 'indind'. If 'indind' is not specied, then all variables in 'datain' 
which are not specified in 'inddep' are used as predictors at 1st stage. At 2nd stage, all such variables are used as predictors anyway.\cr
The optimization citerion is the balanced accuracy on the full sample. The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'. \cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
}
\details{
See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.1, for further information.

Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name)} where 'name' is the output data 
frame of the function.\cr
}
\examples{
data <- PrInDT::data_land # load data
dataclean <- data[,c(1:7,23:24,11:13,22,8:10)]  # only relevant features
indind <- c(1:9) # original predictors
inddep <- c(14:16) # dependent variables
dataland <- na.omit(dataclean)
ctestv <- NA
perc <- c(0.45,0.05,0.25)   # percentages of observations of larger class, 
# 1 per dependent class variable
perc2 <- c(0.75,0.95,0.75)  # percentages of observations of smaller class, 
# 1 per dependent class variable
outland <- C2SPrInDT(dataland,percl=perc,percs=perc2,N=19,indind=indind,inddep=inddep)
outland
plot(outland)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTRstruc.R
\name{PrInDTRstruc}
\alias{PrInDTRstruc}
\title{Structured subsampling for regression}
\usage{
PrInDTRstruc(datain,regname,ctestv=NA,Struc=NA,vers="d",M=NA,Mit=NA,N=99,Pit=NA,
               pobs=c(0.9,0.7),ppre=c(0.9,0.7),conf.level=0.95,minsplit=NA,minbucket=NA,
               valdat=datain,indrep=0)
}
\arguments{
\item{datain}{Input data frame with continuous target variable 'regname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{regname}{Name of target variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{Struc}{= list(name,check,labs), cf. description for explanations; Struc not needed for vers="b"}

\item{vers}{Version of structured subsampling: "a", "b", "c", "d", cf. description; \cr
default = "d"}

\item{M}{Number of repetitions of subsampling from substructure (integer) in versions "a" and "d";\cr
default = 99}

\item{Mit}{List of number of elements of substructure (integers); \cr
default = c((Cl-4):Cl), Cl = maximum number elements in both categories}

\item{N}{Number of repetitions of subsampling from predictors (integer) in versions "b" and "c";\cr
default = 99}

\item{Pit}{List of number of predictors (integers)\cr
default = c(max(1,(D-3)):D), D = maximum number of predictors}

\item{pobs}{Percentage(s) of observations for subsampling in versions "c" and "d";\cr
default=c(0.9,0.7)}

\item{ppre}{Percentage(s) of predictors for subsampling in versions "c" and "d";\cr
default=c(0.9,0.7)}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}

\item{valdat}{Validation data; default = datain}

\item{indrep}{Indicator for repeated measurements, i.e. more than one observation with the same class for each element;\cr 
indrep=1: Struc=list(name) only; default = 0}
}
\value{
\describe{
  \item{outmax}{Best tree} 
  \item{interp}{Number of interpretable trees, overall number of trees} 
  \item{ntmax}{Size of training set for best tree} 
  \item{R2max}{R squared of best tree}
  \item{R2sub}{Mean R squared of objects in substructure}
  \item{MAEmax}{MAE (Mean Absolute Error) of best tree}
  \item{MAEsub}{Mean MAE of objects in substructure}
  \item{ind1max}{Elements of 1st category of substructure used by best tree} 
  \item{ind2max}{Elements of 2nd category of substructure used by best tree} 
  \item{indmax}{Predictors used by best tree} 
  \item{gmaxTrain}{Training set for best tree} 
  \item{labs}{labs from Struc} 
  \item{vers}{Version used for structured subsampling} 
  \item{lMit}{Number of different numbers of substructure elements}
  \item{lPit}{Number of different numbers of predictors}
  \item{M}{Number of repetitions of selection of substructure elements} 
  \item{N}{Number of repetitions of selection of predictors}
  \item{indrep}{Indicator of repeated measurements: indrep=1} 
}
}
\description{
The function \code{\link{PrInDTRstruc}} applies structured subsampling for finding an optimal subsample to model
the relationship between the continuous variable 'regname' and all other factor and numerical variables
in the data frame 'datain' by means of 'M' repetitions of subsampling from a substructure and 'N' repetitions of subsampling from the predictors. 
The optimization citerion is the goodness of fit R2 on the validation sample 'valdat' (default = 'datain').  
The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'.\cr
The substructure of the observations used for subsampling is specified by the list 'Struc' which consists of the factor variable 'name' representing the substructure,
the name 'check' of the factor variable with the information about the categories of the substructure, and the matrix 'labs' which specifies the values of 'check'
corresponding to two categories in its rows, i.e. in 'labs[1,]' and 'labs[2,]'. The names of the categories have to be specified by \code{rownames(labs)}.\cr
In structured subsampling, 'M' repetitions of subsampling of the variable 'name' with 'Mit' different elements of each category in 'check' are realized. If 'Mit' is a list, each entry is employed individually. If 'Mit' is larger than the maximum available number of elements with a certain value of 'check', the maximum possible number of elements is used.\cr
The number of predictors 'Pit' to be included in the model and the number of elements of the substructure 'Mit' have to be specified (lists allowed).\cr
The percentages of involved observations and predictors can be controlled by the parameters 'pobs' and 'ppre', respectively.\cr
The parameter 'Struc' is needed for all versions of subsampling except "b". Four different versions of structured subsampling exist:\cr 
a) just of the elements in the substructure with parameters 'M' and 'Mit',\cr
b) just of the predictors with parameters 'N' and 'Pit',\cr
c) of the predictors and for each subset of predictors subsampling of the elements of the substructure with parameters 'M', 'N', 'Mit', 'Pit', 'pobs', and 'ppre', and\cr
d) of the elements of the substructure and for each of these subsets subsampling of the predictors with the same parameters as version c).\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
Besides the maximal R2, the minimal MAE (Mean Absolute Error) is reported. \cr\cr
Repeated measurements can also be handled by this function (indrep=1). They are multiple measurements of the same variable taken on the same subjects (or objects) either under different conditions 
or over two or more time periods.\cr
The variable with the repeatedly observed subjects (or objects) is assumed to be 'name' in 'Struc'.\cr
The measure MAE is split according to the values of 'name'.
}
\details{
See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.4, for further information.

Standard output can be produced by means of \code{print(name$besttree)} or just \code{name$besttree} as well as \code{plot(name$besttree)} where 'name' is the output data 
frame of the function.\cr
}
\examples{
data <- PrInDT::data_vowel
data <- na.omit(data)
CHILDvowel <- data$Nickname
data$Nickname <- NULL
data$syllables <- 3 - data$syllables
data$speed <- data$word_duration / data$syllables  ## NEW NEW
names(data)[names(data) == "target"] <- "vowel_length"
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
name <- CHILDvowel
check <- "data$ETH"
labs <- matrix(1:6,nrow=2,ncol=3)
labs[1,] <- c("C1a","C1b","C1c")
labs[2,] <- c("C2a","C2b","C2c")
rownames(labs) <- c("children 1","children 2")
Struc <- list(name=name,check=check,labs=labs)
outstruc <- PrInDTRstruc(data,"vowel_length",ctestv=ctestv,Struc=Struc,vers="d",
                  M=3,Mit=21,N=9,pobs=c(0.95,0.7),ppre=c(1,0.7),conf.level=0.99)
outstruc
plot(outstruc)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PrInDTMulevALL.R
\name{PrInDTMulevAll}
\alias{PrInDTMulevAll}
\title{Conditional inference tree (ctree) for multiple classes on all observations}
\usage{
PrInDTMulevAll(datain,classname,ctestv=NA,conf.level=0.95,minsplit=NA,minbucket=NA)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1)\cr
(default = 0.95)}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}
}
\value{
\describe{
\item{treeall}{ctree based on all observations}
\item{baAll}{balanced accuracy of 'treeall'}
\item{interpAll}{criterion of interpretability of 'treeall' (TRUE / FALSE)}
\item{confAll}{confusion matrix of 'treeall'}
}
}
\description{
ctree for more than 2 classes on all observations. Interpretability is checked (see 'ctestv').
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
}
\details{
Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)} where 'name' is the output data 
frame of the function.
}
\examples{
datastrat <- PrInDT::data_zero
data <- na.omit(datastrat)
ctestv <- rbind('ETH == {C2a,C1a}', 'MLU == {1, 3}')
data$rel[data$ETH \%in\% c("C1a","C1b","C1c") & data$real == "zero"] <- "zero1"
data$rel[data$ETH \%in\% c("C2a","C2b","C2c") & data$real == "zero"] <- "zero2"
data$rel[data$real == "realized"] <- "real"
data$rel <- as.factor(data$rel) # rel is new class variable
data$real <- NULL # remove old class variable
conf.level <- 0.99 # 1 - significance level (mincriterion) in ctree
outAll <- PrInDTMulevAll(data,"rel",ctestv,conf.level) 
outAll # print model based on all observations
plot(outAll)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{data_zero}
\alias{data_zero}
\title{Subject pronouns}
\format{
A data frame with 1024 observations and 7 columns
\describe{
  \item{real}{subject pronoun realized? Factor (2 levels "zero","realized")}
  \item{AGE}{age: Numerical (in months)}
  \item{LiBa}{linguistic background: Factor (3 levels "mono","multi", NA)}
  \item{ETH}{ethnic group: Factor (6 levels "C1a","C1b","C1c","C2a","C2b","C2c") (anonymized)}
  \item{SEX}{gender: Factor (2 levels "female","male")}
  \item{MLU}{mean length of utterance: Factor (4 levels "1","2","3","OL")}
  \item{PRN}{pronoun: Factor (7 levels "I","you_s","he","she","it","we","they")}
}
}
\source{
Sarah Buschfeld, TU Dortmund
}
\usage{
data_zero
}
\description{
Usage of subject pronouns and its predictors.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SimRPrInDT.R
\name{SimRPrInDT}
\alias{SimRPrInDT}
\title{Interdependent estimation for regression}
\usage{
SimRPrInDT(data,ctestv=NA,Struc=NA,inddep,N=99,pobs=0.9,ppre=c(0.9,0.7),
                M=1,nsub=1,conf.level=0.95,minsplit=NA,minbucket=NA)
}
\arguments{
\item{data}{Input data frame with continuous target variables with column indices 'inddep' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
Example: ctestv <- rbind('variable1 == \{value1, value2\}','variable2 <= value3'), where
character strings specified in 'value1', 'value2' are not allowed as results of a splitting operation in variable 1 in a tree.\cr
For restrictions of the type 'variable <= xxx', all split results in a tree are excluded with 'variable <= yyy' and yyy <= xxx.\cr
Trees with split results specified in 'ctestv' are not accepted during optimization.\cr
A concrete example is: 'ctestv <- rbind('ETH == \{C2a, C1a\}','AGE <= 20')' for variables 'ETH' and 'AGE' and values 'C2a','C1a', and '20';\cr
If no restrictions exist, the default = NA is used.}

\item{Struc}{= list(name,check,labs), cf. description for explanations}

\item{inddep}{Column indices of target variables in datain}

\item{N}{Number of repetitions of subsampling (integer) of predictors; default = 99}

\item{pobs}{Percentage(s) of observations for subsampling; default=0.9}

\item{ppre}{Percentage(s) of predictors for subsampling; default=c(0.9,0.7)}

\item{M}{Number of repetitions of subsampling of elements of substructure}

\item{nsub}{(List of) numbers of different elements of substructure per subsample}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}
}
\value{
\describe{
  \item{modelsF}{Best trees at stage 1} 
  \item{modelsI}{Best trees for the different values of 'nsub' at stage 2}
  \item{modelsJ}{Best trees for the different values of 'nsub' after mean optimization}
  \item{Struc}{Used structure}
  \item{msub}{Best numbers of elements in substructure: 2nd stage, 3rd stage}
  \item{depnames}{names of dependent variables}
  \item{R2All}{R2s of best trees at stages 1, 2, mean max.}  
}
}
\description{
The function \code{\link{SimRPrInDT}} applies structured subsampling for finding an optimal subsample to model
the relationship between the continuous variables with indices 'inddep' and all other factor and numerical variables
in the data frame 'datain'. \cr
The substructure of the observations used for subsampling is specified by the list 'Struc' which consists of the factor variable 'name' representing the substructure,
the name 'check' of the factor variable with the information about the categories of the substructure, and the matrix 'labs' which specifies the values of 'check'
corresponding to two categories in its rows, i.e. in 'labs[1,]' and 'labs[2,]'. The names of the categories have to be specified by \code{rownames(labs)}.\cr
In structured subsampling, first 'M' repetitions of subsampling of the variable 'name' with 'nsub' different elements of each category in 'check' are realized. If 'nsub' is a list, each entry is employed individually. If 'nsub' is larger than the maximum available number of elements with a certain value of 'check', the maximum possible number of elements is used. Then,
for each of the subsamples 'N' repetitions of subsampling of 'ppre' percentages of the predictors are carried out.\cr 
Subsampling of observations can additionally be restricted to 'pobs' percentages.\cr
The optimization citerion is the goodness of fit R2 on the full sample. At stage 2, the models are optimized individually. 
At stage 3, the mean of accuracies is optimized over all models.\cr
Struc=NA causes random subsampling of observations instead of structured subsampling.\cr
The trees generated from undersampling can be restricted by not accepting trees 
including split results specified in the character strings of the vector 'ctestv'.\cr
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr
}
\details{
See Buschfeld & Weihs (2025), Optimizing decision trees for the analysis of World Englishes and sociolinguistic data. Cambridge University Press, section 4.5.6.2, for further information.

Standard output can be produced by means of \code{print(name)} or just \code{name} as well as \code{plot(name} where 'name' is the output data 
frame of the function.\cr
}
\examples{
data <- PrInDT::data_vowel
data <- na.omit(data)
CHILDvowel <- data$Nickname
data$Nickname <- NULL
syllable <- 3 - data$syllables
data$syllabels <- NULL
data$syllables <- syllable
data$speed <- data$word_duration / data$syllables
names(data)[names(data) == "target"] <- "vowel_length"
# interpretation restrictions (split exclusions)
ctestv <- rbind('ETH == {C2a, C1a}','MLU == {1, 3}') # split exclusions
# structure definition
name <- CHILDvowel
check <- "data$ETH"
labs <- matrix(1:6,nrow=2,ncol=3)
labs[1,] <- c("C1a","C1b","C1c")
labs[2,] <- c("C2a","C2b","C2c")
rownames(labs) <- c("children 1","children 2")
Struc <- list(name=name,check=check,labs=labs)
# column indices of dependent variables
inddep <- c(13,9) 
outSimR <- SimRPrInDT(data,ctestv=ctestv,Struc=Struc,inddep=inddep,N=3,M=2,
                   nsub=c(19,20),conf.level=0.99)
outSimR
plot(outSimR)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NesPrInDT.R
\name{NesPrInDT}
\alias{NesPrInDT}
\title{Nested \code{\link{PrInDT}} with additional undersampling of a factor with two unbalanced levels}
\usage{
NesPrInDT(datain,classname,ctestv=NA,N,plarge,psmall=1.0,conf.level=0.95,
       thres=0.5,stratvers=0,strat=NA,seedl=TRUE,nesvar,nesunder,repin,
       minsplit=NA,minbucket=NA)
}
\arguments{
\item{datain}{Input data frame with class factor variable 'classname' and the\cr
influential variables, which need to be factors or numericals (transform logicals and character variables to factors)}

\item{classname}{Name of class variable (character)}

\item{ctestv}{Vector of character strings of forbidden split results;\cr
(see function \code{\link{PrInDT}} for details.)\cr
If no restrictions exist, the default = NA is used.}

\item{N}{Number of repetitions (integer > 0)}

\item{plarge}{Undersampling percentage of larger class (numerical, > 0 and <= 1)}

\item{psmall}{Undersampling percentage of smaller class (numerical, > 0 and <= 1);\cr
default = 1}

\item{conf.level}{(1 - significance level) in function \code{ctree} (numerical, > 0 and <= 1);\cr
default = 0.95}

\item{thres}{Probability threshold for prediction of smaller class; default = 0.5}

\item{stratvers}{Version of stratification;\cr
= 0: none (default),\cr
= 1: stratification according to the percentages of the values of the factor variable 'strat',\cr
> 1: stratification with minimum number 'stratvers' of observations per value of 'strat'}

\item{strat}{Name of one (!) stratification variable for undersampling (character);\cr
default = NA (no stratification)}

\item{seedl}{Should the seed for random numbers be set (TRUE / FALSE)?\cr
default = TRUE}

\item{nesvar}{Name of factor to be undersampled (character)}

\item{nesunder}{Data of factor to be undersampled (integer)}

\item{repin}{Number of repetitions (integer) for undersampling of 'nesvar'}

\item{minsplit}{Minimum number of elements in a node to be splitted;\cr
default = 20}

\item{minbucket}{Minimum number of elements in a node;\cr
default = 7}
}
\value{
\describe{  
\item{undba}{balanced accuracies on undersamples}
\item{imax}{indices of best trees on undersamples}
\item{undba3en}{balanced accuracies of ensembles of 3 best trees on undersamples}
\item{accF}{balanced accuracies on full sample}
\item{accE}{balanced accuracy on full sample of best ensemble of 3 trees from undersampling}
\item{maxt}{indices of best trees on full sample}
\item{treesb}{3 best trees of all undersamples of 'nesunder'; refer to an individual tree as \code{treesb[[k]]}, k = 1, ..., 3*repin}
}
}
\description{
Function for additional undersampling of the factor 'nesvar' with two unbalanced levels to avoid dominance of the level with higher frequency.
The factor 'nesvar' is allowed not be part of the input data frame 'datain'. The data of this factor is given in the vector 'nesunder'. 
The observations in 'nesunder' have to represent the same cases as in 'datain' in the same ordering.\cr
\code{\link{PrInDT}} is called 'repin' times with subsamples of the original data so that the level with the larger frequency in the vector 'nesunder' has 
approximately the same number of values as the level with the smaller frequency.\cr
Only the arguments 'nesvar', 'nesunder', and 'repin' relate to the additional undersampling, all the other arguments relate to the standard 
\code{\link{PrInDT}} procedure. \cr As in \code{\link{PrInDT}}, the aim is to optimally model the relationship between the two-class factor variable 'classname' and all other factor and  
numerical variables in the data frame 'datain' by means of 'N' repetitions of undersampling. The trees generated by \code{\link{PrInDT}} can be
restricted by excluding unacceptable trees which include split results specified in the character strings of the vector 'ctestv'.\cr
The probability threshold 'thres' for the prediction of the smaller class may be specified (default = 0.5).\cr
Undersampling may be stratified in two ways by the feature 'strat'.\cr
The results are evaluated on the full sample and on the subsamples of 'nesunder'.
The parameters 'conf.level', 'minsplit', and 'minbucket' can be used to control the size of the trees.\cr

\strong{Reference} \cr Weihs, C., Buschfeld, S. 2021b. NesPrInDT: Nested undersampling in PrInDT. 
arXiv:2103.14931
}
\details{
Standard output can be produced by means of \code{print(name)} or just \code{ name } as well as \code{plot(name)}  where 'name' is the output data 
frame of the function.\cr
The plot function will produce a series of more than one plot. If you use R, you might want to specify \code{windows(record=TRUE)} before 
\code{plot(name)} to save the whole series of plots. In R-Studio this functionality is provided automatically.
}
\examples{
# data input and preparation --> data frame with 
#   class variable, factors, and numericals (no character variables)!!
data <- PrInDT::data_speaker
data <- na.omit(data)
nesvar <- "SPEAKER"
N <- 49  # no. of repetitions in inner loop
plarge <- 0.06 # sampling percentage for larger class in nesunder-subsample
psmall <- 1 # sampling percentage for smaller class in nesunder-subsample
nesunder <- data$SPEAKER
data[,nesvar] <- list(NULL)
outNes <- NesPrInDT(data,"class",ctestv=NA,N,plarge,psmall,conf.level=0.95,nesvar=nesvar,
  nesunder=nesunder,repin=5)
outNes
plot(outNes)
hist(outNes$undba,main=" ",xlab = "balanced accuracies of 3 best trees of all undersamples")

}
